
(* Most of the game logic, and the main game loop, are in here. *)

structure Game =
struct

  type ptr = MLton.Pointer.t

  exception Nope

  exception Game of string

  (* XXX urg, bad choice *)
  structure U = Util
  open SDL
  structure Util = U

  fun abs z = if z < 0 then ~ z else z
  fun sgn x = if x < 0 then ~1 else 1

  fun messagebox s = 
      let
        (* val mb = _import "MessageBoxA" stdcall : int * string * string * int -> unit ; *)
      in
        (* mb (0, s ^ "\000", "message", 0) *)
        print (s ^ "\n")
      end

  val showclip = ref false

  (* Allowing resizing of the screen, so that we can edit at high
     resolutions. We can and should assume a constant 800x600 resolution
     for the player. *)
  val width = Params.asint 800 Args.width
  val height = Params.asint 600 Args.height

  val TILESW = width div Tile.TILEW
  val TILESH = height div Tile.TILEH

  val SNAILH = 32
  val SNAILW = 48

  open Types
  open Tile
  open Room

  val screen = makescreen (width, height)

  fun requireimage s =
      let val s = "graphics/" ^ s
      in
          case Image.load s of
              NONE => (print ("couldn't open " ^ s ^ "\n");
                       raise Nope)
            | SOME p => p
      end

  (* A crawler has a set of images, for each attachment, and for
     each orientation in that attachment. This takes a basename
     like "snail" and loads specialized graphics for each of the
     directions, or synthesizes them if they don't exist. The
     rules for synthesis are:

      1. There is no synthesis yet.
      *)
  fun require_crawl_images s =
      let
          val floorl = requireimage ("floor" ^ s ^ "l.png")
          val floorr = requireimage ("floor" ^ s ^ "r.png")
          val ceill = requireimage ("ceil" ^ s ^ "l.png")
          val ceilr = requireimage ("ceil" ^ s ^ "r.png")

          val leftu = requireimage ("left" ^ s ^ "u.png")
          val leftd = requireimage ("left" ^ s ^ "d.png")
          val rightu = requireimage ("right" ^ s ^ "u.png")
          val rightd = requireimage ("right" ^ s ^ "d.png")

      in
          { floorl = floorl,
            floorr = floorr,
            ceill = ceill,
            ceilr = ceilr,
            leftu = leftu,
            leftd = leftd,
            rightu = rightu,
            rightd = rightd }
          (* XXX should add falling versions of these. *)
      end

  structure Font = FontFn (val surf = requireimage "font.png"
                           val charmap =
                           " ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789" ^
                           "`-=[]\\;',./~!@#$%^&*()_+{}|:\"<>?" (* \" *)
                           (* CHECKMARK ESC HEART LCMARK1 LCMARK2 BAR_0 BAR_1 BAR_2 BAR_3 
                              BAR_4 BAR_5 BAR_6 BAR_7 BAR_8 BAR_9 BAR_10 BARSTART 
                              LRARROW LLARROW *)
                           val width = 9
                           val height = 16
                           val styles = 6
                           val overlap = 1
                           val dims = 3)

  val snail = require_crawl_images "snail"
  val snaill = #floorl snail
  val snailr = #floorr snail
  val right_snailu = #rightu snail
  val right_snaild = #rightd snail

  val snailr_fade = SDL.Util.alphadim (#floorr snail)
  val snaill_fade = SDL.Util.alphadim (#floorl snail)
  val right_snailu_fade = SDL.Util.alphadim right_snailu
  val right_snaild_fade = SDL.Util.alphadim right_snaild
  val error = requireimage "error_frame.png"

  val redhi = requireimage "redhighlight.png"
  val greenhi = requireimage "greenhighlight.png"
  val robobox = requireimage "robobox.png"

  val star = requireimage "redstar.png"
  val spider_up = requireimage "spider_up.png"
  val spider_down = requireimage "spider_down.png"

  val fireballr = requireimage "fireball.png"
  val fireballl = requireimage "fireball_left.png"

  (* snail flags *)
  structure BF :>
  sig
      include FLAG
      val ONGROUND : flag
  end =
  struct
      open Flag
      val ONGROUND = 0
      val NFLAGS = 1
      val flags = nflags NFLAGS
  end

  val snailx = ref (Room.STARTX * TILEW)
  val snaily = ref (Room.STARTY * TILEH)

  (* XXX Cosmetic, and currently not general. This should probably
     be relative to the orientation. It should not be set just
     because the left key was just pressed, since left means different
     things depending on the orientation and context. *)
  val snailface = ref FRIGHT
  (* snail dir is the snail's orientation, which is the direction
     that its feet are pointing. (In normal non-stuck gravity,
     DOWN.) *)
  val snaildir = ref DOWN
  val slime : (int * int) option ref = ref NONE
  val brightgreen = SDL.color (0w0, 0wxFF, 0w0, 0wxFF)
  val brightred = SDL.color (0wxFF, 0w0, 0w0, 0wxFF)
  val slimecolor = ref brightgreen

  (* position of top-left of scroll window, in pixels *)
  val scrollx = ref 0
  val scrolly = ref 0
  val scrolldx = ref 0
  val scrolldy = ref 0
  (* I think these became in units of 1 pixel per frame *)
  val snaildx = ref 0
  val snaildy = ref 0
  val snailstate = BF.flags

  (* For many keys, we keep track of their state (in the event loop below).
     This is so that you can hold both left and right momentarily without
     things getting cocked up. Note that keys are used both for the player
     movement (where they are canonicalized inside the movement code) and
     for control in the editor, and maybe menus some day... *)
  val KEYMASK_UP = 0w1 : Word32.word
  val KEYMASK_DOWN = 0w2 : Word32.word
  val KEYMASK_LEFT = 0w4 : Word32.word
  val KEYMASK_RIGHT = 0w8 : Word32.word
  val KEYMASK_SHIFT = 0w16 : Word32.word
  val KEYMASK_CTRL = 0w32 : Word32.word
  val keymask = ref (0w0 : Word32.word)
  fun holding_key mask = Word32.andb (!keymask, mask) <> 0w0

  datatype monster = STAR | SPIDER

  datatype object =
      Bullet   of { dx : int, dy : int, x : int, y : int }
    | Monster  of { which : monster, dir : dir, dx : int, dy : int, x : int, y : int }

  val objects = ref (nil : object list)
      
  val paused = ref false
  val advance = ref false

  fun onscreen (x, y) =
      SOME ((x + !scrollx) div TILEW,
            (y + !scrolly) div TILEH)

  datatype editmode = ZBG | ZFG | ZMASK | ZPROP | ZBRUSH

  val message = ref ""

  val room = ref Room.start

  (* This is the selectable tile area. *)

  (* Horizontal editor *)
  val TITLE_HEIGHT = Font.height + 2
  val OFFSET_OF_TILES_FROM_EDITY = TITLE_HEIGHT + TILEH + 2
  val EDITW = Tile.TILESET_WIDTH
  val EDITH = 16
  val EDITWIDTH = EDITW * TILEW + 4
  val EDITHEIGHT = 200
  val editmode = (* ref ZBRUSH *) ref ZBG
  val showmask = ref true
  val showprop = ref true
  val editx = ref 0
  val edity = ref (height - EDITHEIGHT)

  (* vertical editor, rhs *)
(*
  val TITLE_HEIGHT = Font.height + 2
  val OFFSET_OF_TILES_FROM_EDITY = TITLE_HEIGHT + TILEH + 2
  val EDITW = 32
  val EDITH = 16
  val EDITWIDTH = EDITW * TILEW + 4
  val EDITHEIGHT = height
  val editmode = (* ref ZBRUSH *) ref ZBG
  val showmask = ref true
  val showprop = ref true
  val editx = ref (width - EDITWIDTH)
  val edity = ref 0 (* (height - EDITHEIGHT) *)
*)

  (* XXX should be option, should be reset (or saved) when changing modes *)
  val edit_lastidx = ref 0

  val NUM_EDIT_TILESETS = Tile.TILESET_HEIGHT div EDITH + (case EDITH mod Tile.TILESET_HEIGHT of
                                                               0 => 0
                                                             | _ => 1 (* extra page for leftovers *))
  val edittileset = ref 0
  val editbgtile = ref (Tile.fromword 0w1)
  val editfgtile = ref (Tile.fromword 0w0)
  val editmask = ref Tile.MSOLID
  val editprop = ref Tile.PPLAIN
  val editbrush = ref (dirt_material, BRECT (3, 3))
  val editmenucolor = SDL.color (0wx44, 0wx14, 0wx14, 0wxFF)
  val editmodecolor = SDL.color (0wxFF, 0wx14, 0wx14, 0wxFF)
  val showedit = ref false
  local
      val thisone = requireimage "thisone.png"
      val brushes = requireimage "brush.png"

      val MASKS =
          Vector.fromList 
          [ MEMPTY, MSOLID,
            MRAMP LM, MRAMP MH, MRAMP HM, MRAMP ML,
            MDIAG NW, MDIAG NE, MDIAG SW, MDIAG SE,
            MCEIL LM, MCEIL MH, MCEIL HM, MCEIL ML,
            MLEFT LM, MLEFT MH, MLEFT HM, MLEFT ML,
            MRIGHT LM, MRIGHT MH, MRIGHT HM, MRIGHT ML

            (* XXX more *)
            ]

      val PROPS =
          Vector.fromList
          ([ PPLAIN, 
            PGOTO (UP, 0), PGOTO (DOWN, 0), PGOTO (LEFT, 0), PGOTO (RIGHT, 0),
            PGOTO (UP, 1), PGOTO (DOWN, 1), PGOTO (LEFT, 1), PGOTO (RIGHT, 1),
            PGOTO (UP, 2), PGOTO (DOWN, 2), PGOTO (LEFT, 2), PGOTO (RIGHT, 2),
            PGOTO (UP, 3), PGOTO (DOWN, 3), PGOTO (LEFT, 3), PGOTO (RIGHT, 3),
            PCURRENT UP, PCURRENT DOWN, PCURRENT LEFT, PCURRENT RIGHT,
            PLIGHT LWHITE, PLIGHT LYELLOW, PLIGHT LBLUE, PLIGHT LRED, PLIGHT LGREEN
            ]
           @ List.tabulate (14, PACTION))

      (* Has to have the same contents as the icon. *)
      val BRUSHW = 16 (* pixels *)
      val BRUSHH = 16
      val BRUSHES_SETW = 16 (* tiles *)
      val BRUSHES_SET =
          Vector.fromList
          [ SOME (eraser_material, BERASERECT (1,1)),
            SOME (eraser_material, BERASERECT (3,3)),
            SOME (eraser_material, BERASERECT (9,9)),
            (* NONE x 13 *)
            NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,
            SOME (dirt_material, BRECT (1, 1)),
            SOME (dirt_material, BRECT (3, 3)),
            SOME (dirt_material, BRECT (9, 9)),
            SOME (dirt_material, BSLOPEDRECT (4, 4)),
            SOME (dirt_material, BSLOPEDRECT (9, 9)),
            (* NONE x 11 *)
            
            NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, 
            SOME (block_material, BRECT (1, 1)),
            SOME (block_material, BRECT (3, 3)),
            SOME (block_material, BRECT (9, 9)),
            SOME (block_material, BSLOPEDRECT (4, 4)),
            SOME (block_material, BSLOPEDRECT (9, 9)), 
            (* NONE x 11 *)
            NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, 
            SOME (scribble_material, BRECT (1, 1)),
            SOME (scribble_material, BRECT (3, 3)),
            SOME (scribble_material, BRECT (9, 9)),
            SOME (scribble_material, BSLOPEDRECT (4, 4)),
            SOME (scribble_material, BSLOPEDRECT (9, 9)) ]

      fun edit_click (x, y) =
          let 

              (* XXX Should allow to click bg/fg/mask to switch to
                 editing that mode *)

              (* skip current bg/fg/mask display *)
              val y = y - OFFSET_OF_TILES_FROM_EDITY + TILEH (* XXX don't know why I need TILEH here, but works *)

              (* which tile are we clicking, then? *)
              val idx = ((y div TILEH) * EDITW + (x div TILEW))
              val () = print ("orig idx: " ^ Int.toString idx ^ "\n")
              val idx = idx + (!edittileset * EDITH) * Tile.TILESET_WIDTH
              val () = print ("scrolled idx: " ^ Int.toString idx ^ "\n")
          in
              if idx >= 0
              then
                  let in
                      edit_lastidx := idx;
                      print ("tile/mask " ^ Int.toString idx ^ "\n");
                      (case !editmode of
                           ZMASK =>
                               if idx >= 0 andalso idx < Vector.length MASKS
                               then editmask := Vector.sub(MASKS, idx)
                               else ()
                         | ZBG => editbgtile := Tile.fromword (Word32.fromInt idx)
                         | ZFG => editfgtile := Tile.fromword (Word32.fromInt idx)
                         | ZBRUSH =>
                               if idx >=0 andalso idx < Vector.length BRUSHES_SET
                               then case Vector.sub(BRUSHES_SET, idx) of
                                      NONE => ()
                                    | SOME b => editbrush := b
                               else ()
                         | ZPROP =>
                               if idx >=0 andalso idx < Vector.length PROPS
                               then editprop := Vector.sub(PROPS, idx)
                               else ())

                  end
              else ()
          end

      fun draw_edit_child (surf, x, y, _ (* width, fixed *), _ (* height, fixed *)) =
          let
              fun drawloop draw =
                let in
                  (Util.for 0 (EDITH - 1)
                   (fn yy =>
                    Util.for 0 (EDITW - 1)
                    (fn xx =>
                     draw (EDITW * yy + xx,
                           x + (TILEW * xx),
                           TILEH + 2 +
                           y + (TILEH * yy)))));

                  (* XXX do it analytically! PERF *)
                  (Util.for 0 (EDITH - 1)
                   (fn yy =>
                    Util.for 0 (EDITW - 1)
                    (fn xx =>
                     if (EDITW * yy + xx = !edit_lastidx)
                     then SDL.blitall (thisone, surf, 
                                       (* ugh Lololllll *)
                                       ~1 + x + (TILEW * xx),
                                       ~1 + TILEH + 2 +
                                       y + (TILEH * yy))
                     else ())))
                end

          in
              (* indicate the mode by coloring its background *)
              SDL.fillrect (surf, 
                            x + ~1 + 
                            ((TILEW + 2) *
                             (case !editmode of
                                     ZBG => 0
                                   | ZFG => 1
                                   | ZMASK => 2
                                   | ZPROP => 3
                                   | ZBRUSH => 4)), 
                            y - 1,
                            TILEW + 4,
                            TILEH + 4,
                            editmodecolor);

              (* the background, foreground, mask, prop *)
              Tile.draw     (!editbgtile, surf, x,                   y, 0, 0);
              Tile.draw     (!editfgtile, surf, x + 1 * (TILEW + 2), y, 0, 0);
              Tile.drawmask (!editmask,   surf, x + 2 * (TILEW + 2), y);
              Tile.drawprop (!editprop,   surf, x + 3 * (TILEW + 2), y);
              (* XXX draw brush? *)

              case !editmode of
                  ZMASK =>
                      drawloop (fn (idx, destx, desty) =>
                                (* might subscript *)
                                let val m = Vector.sub (MASKS, idx)
                                in
                                    Tile.drawmask (m, surf, destx, desty)
                                end handle Subscript => ())

                | ZPROP =>
                      drawloop (fn (idx, destx, desty) =>
                                (* might subscript *)
                                let val p = Vector.sub (PROPS, idx)
                                in
                                    Tile.drawprop (p, surf, destx, desty)
                                end handle Subscript => ())
 
                | ZBRUSH =>
                      drawloop (fn (idx, destx, desty) =>
                                (* This is different; we just use the
                                   organization in the graphic itself
                                   since these brushes don't need to
                                   be first class citizens. *)
                                let val x = idx mod BRUSHES_SETW
                                    val y = idx div BRUSHES_SETW
                                in
                                    SDL.blit (brushes, x * BRUSHW, y * BRUSHH,
                                              BRUSHW, BRUSHH, surf, destx, desty)
                                end)


                | _ =>
                      drawloop (fn (idx, destx, desty) =>
                                let 
                                    (* to account for scroll: a page is TILESET_WIDTH times
                                       the number of rows we show on a page. *)
                                    val idx = idx + (!edittileset * EDITH) * Tile.TILESET_WIDTH
                                    val t = Tile.fromword (Word32.fromInt idx)
                                in
                                    Tile.draw (t, surf, destx, desty, 0, 0)
                                end)
             end

      fun drawtitle (surf, x, y) =
          let val title = "editor. ^2z^< to change mode: ^3" ^
              (case !editmode of
                   ZBG => "background"
                 | ZFG => "foreground"
                 | ZMASK => "mask"
                 | ZPROP => "prop"
                 | ZBRUSH => "brush") ^
                   "^< ^2m^<ask ^3" ^
                   (if !showmask then "ON" else "OFF") ^
                   "^< ^2p^<rop ^3" ^
                   (if !showprop then "ON" else "OFF") ^
                   "^<"
          in
              Font.draw(surf, x, y + 1, title)
          end

  in
      val edit_window =
          Window.make { click = edit_click,
                        keydown = fn _ => false,
                        keyup = fn _ => false,
                        
                        border = 0,
                        border_color = editmenucolor,
                        resizable = false,
                        x = !editx, y = !edity,
                        w = EDITWIDTH, h = EDITHEIGHT,
                        title_height = TITLE_HEIGHT,

                        draw = draw_edit_child,
                        drawtitle = drawtitle }

      fun drawedit _ =
          if !showedit (* andalso holding_key KEY_SHIFT *)
          then Window.draw edit_window screen
          else ()

  end


  datatype drawwhat =
      (* Debugging: Draw the clip mask *)
      DRAWMASK
      (* Draw the properties *)
    | DRAWPROP
      (* Draw the indicated layer *)
    | DRAW of Room.layer

  (* draw a view of the world where the top left of
     the screen is at scrollx/scrolly *)
  fun drawzone (layer) =
    let
      (* tlx/tly is the (screen coordinate pixel) point 
         at which we draw the top-left tile. it is nonpositive *)
      val tlx = 0 - (!scrollx mod TILEW)
      val tly = 0 - (!scrolly mod TILEH)
        
      (* this is the tile number (world absolute) that we start
         by drawing *)
      val xstart = !scrollx div TILEW
      val ystart = !scrolly div TILEH
    in
      case layer of
        (* not really a layer, but useful when editing / debugging *)
        DRAWMASK =>
        (* nb, always doing one extra tile, because we might
           display only partial tiles if not aligned to scroll
           view *)
          Util.for 0 TILESW
          (fn x =>
           Util.for 0 TILESH
           (fn y =>
            case Room.maskat (!room, xstart + x, ystart + y) of
                MEMPTY => ()
              | m => Tile.drawmask(m, screen, 
                                   tlx + (x * TILEW), tly + (y * TILEH))
                    ))
      | DRAWPROP => 
        (* nb, always doing one extra tile, because we might
           display only partial tiles if not aligned to scroll
           view *)
          Util.for 0 TILESW
          (fn x =>
           Util.for 0 TILESH
           (fn y =>
            case Room.propat (!room, xstart + x, ystart + y) of
                PPLAIN => ()
              | p => Tile.drawprop(p, screen, 
                                   tlx + (x * TILEW), tly + (y * TILEH))
                    ))

      | DRAW layer => 
          Util.for 0 TILESW
          (fn x =>
           Util.for 0 TILESH
           (fn y =>
            Tile.draw (Room.tileat(!room, layer, xstart + x, ystart + y), screen, 
                       tlx + (x * TILEW),
                       tly + (y * TILEH),
                       xstart + x, ystart + y)
            ))
    end

  fun drawobjects () =
      let
          fun doobj (Bullet { x, y, dx, dy }) =
              blitall (if dx > 0 then fireballr
                       else fireballl, screen,
                       x - !scrollx, y - !scrolly)
            | doobj (Monster { which, dx, dy, x, y, dir }) =
              let val glyph = 
                  (case (dir, which, dy) of 
                       (_, STAR, _) => star
                     | (_, SPIDER, 0) => spider_down
                     | (_, SPIDER, _) => spider_up
                           (* more objects... *)
                           ) handle Match => error
              in
                  blitall (glyph,
                           screen,
                           x - !scrollx, y - !scrolly)
              end
      in
          app doobj (!objects)
      end

    fun drawsnail fade =
        let val img = 
            (case (!snaildir, !snailface, fade) of
                 (* FLEFT = cw of snaildir, FRIGHT = ccw of snaildir *)
                 (DOWN, FLEFT, false) => snaill
               | (DOWN, FLEFT, true) => snaill_fade
               | (DOWN, FRIGHT, false) => snailr
               | (DOWN, FRIGHT, true) => snailr_fade
               | (RIGHT, FLEFT, false) => right_snaild
               | (RIGHT, FLEFT, true)  => right_snaild_fade
               | (RIGHT, FRIGHT, false) => right_snailu
               | (RIGHT, FRIGHT, true)  => right_snailu_fade)
               (* XXX nonexhaustive *)
        in
            blitall (img, screen, !snailx - !scrollx, !snaily - !scrolly);
            (* draw the slime *)
            case !slime of
              NONE => ()
            | SOME (x, y) => drawcircle (screen, x - !scrollx, y - !scrolly,
                                         7, !slimecolor)
        end

  fun dtos LEFT = "left"
    | dtos RIGHT = "right"
    | dtos UP = "up"
    | dtos DOWN = "down"

  fun inttos (I_GO d) = "go " ^ dtos d
    | inttos I_JUMP = "jump"
    | inttos I_SUPERJUMP = "superjump"
    | inttos I_BOOST = "boost"

  (* returns a list of intentions from the current key state.
     These should be added to other intentions, like jumping,
     which are not contextual. *)
  fun contextual_intention keys =
      let
          (* The list of directions implied by the currently
             held keys, in screen orientation. *)
          val sd = nil
          val sd = if holding_key KEYMASK_UP then UP :: sd else sd
          val sd = if holding_key KEYMASK_DOWN then DOWN :: sd else sd
          val sd = if holding_key KEYMASK_LEFT then LEFT :: sd else sd
          val sd = if holding_key KEYMASK_RIGHT then RIGHT :: sd else sd
      in
          case !slime of
              (* XXX I think we should also canonize for freefall mode. If
                 I'm being pulled to the right because of current, then
                 I should be able to press 'up' to "move right" in my
                 reference frame *)
              NONE => map I_GO sd
            | SOME(x, y) =>
              let val slime_mask = Room.maskat (!room, x div TILEW, y div TILEH)
              in map I_GO (List.mapPartial (Slime.canonize (!snaildir) slime_mask) sd)
              end
      end

  fun intends il i = List.exists (fn x => x = i) il

  (* number of pixels per frame that we can walk at maximum *)
  val MAXWALK = 4
  val TERMINAL_VELOCITY = 4
  val JUMP_VELOCITY = 3

  (* make sure snail's in the scroll window! *)
  fun setscroll i =
    (* Don't need to do that if editing.
       XXX could at least prevent us from scrolling
       totally out of the level. *)
    if !showedit
    then
      let in
          if holding_key KEYMASK_UP
          then scrolly := !scrolly - 5
          else ();

          if holding_key KEYMASK_DOWN
          then scrolly := !scrolly + 5
          else ();

          if holding_key KEYMASK_LEFT 
          then scrollx := !scrollx - 5
          else ();

          if holding_key KEYMASK_RIGHT
          then scrollx := !scrollx + 5
          else ();

          ()
      end
    else
    let in
      (* stay in the center third of the screen. *)
      if !snailx - !scrollx < (width div 3)
      then ((* print "left edge\n"; *)
            scrollx := !snailx - (width div 3))
      else if (!scrollx + width) - !snailx < (width div 3)
           then ((* print "right edge\n"; *)
                 scrollx := !snailx - 2 * (width div 3))
           else ();

      if !snaily - !scrolly < (height div 3)
      then ((* print "top edge\n"; *)
            scrolly := !snaily - (height div 3))
      else if (!scrolly + height) - !snaily < (height div 3)
           then ((* print "snailtom edge\n"; *)
                 scrolly := !snaily - 2 * (height div 3))
           else ();

      (* But: if we can fit the whole game on-screen, then
         do so. *)
      if Room.ROOMH * TILEH < height
      then scrolly := (height - Room.ROOMH * TILEH) div ~2
      else ();

      if Room.ROOMW * TILEW < width
      then scrollx := (width - Room.ROOMW * TILEW) div ~2
      else ()
    end

  (* Move an entity, which is a solid object that has intentions *)
  fun moveent clip (i, x, y, dx, dy) =
    let
        (* val () = print (StringUtil.delimit ", " (map Int.toString [dx, dy, x, y]) ^ "\n") *)
            (*
        val () = 
            case i of
                nil => () 
              | _ => print (StringUtil.delimit " & " (map inttos i) ^ "\n")
                *)

        (* XXX these maximums should also be parameters *)
        (* first, react to the entity's intentions: *)
        val dx =
            if intends i (I_GO LEFT)
               andalso dx > ~ MAXWALK
            then dx - 1
            else dx
        val dx = 
            if intends i (I_GO RIGHT)
               andalso dx < MAXWALK
            then dx + 1
            else dx

        (*
        (* symmetric code to compute the dy from I_GO UP and I_GO DOWN *)
        (* while attached, only one of these two blocks should execute, because
           all I_GO actions should be orthogonal to the ground *)
        val dy =
            if intends i (I_GO UP)
               andalso dy > ~MAXWALK
            then dy - 1
            else dy
        val dy =
            if intends i (I_GO DOWN)
               andalso dy < MAXWALK
            then dy + 1
            else dy
        *)

        val (dx, i) = if intends i I_BOOST
                      then (dx * 2, List.filter (fn I_BOOST => false | _ => true) i)
                      else (dx, i)

        (* and if we don't intend to move at all, slow us down *)
        val dx = 
            if not (intends i (I_GO RIGHT)) andalso
               not (intends i (I_GO LEFT))
            then (if dx > 0 then dx - 1
                  else if dx < 0 then dx + 1 else 0)
            else dx

        (*
        (* symmetrically slow us down on a wall *)
        val dy =
            if not (intends i (I_GO UP)) andalso
               not (intends i (I_GO DOWN))
            then (if dy > 0 then dy - 1
                  else if dy < 0 then dy + 1 else 0)
            else dy
        *)

        (* unit vector into our feet *)
        val downwards = to_vector (!snaildir)
        (* unit vector out of our back *)
        val upwards = to_vector (dir_reverse (!snaildir))

        (* jumping -- only when not in air already!
           (either way the intention goes away) *)
        (*
        val (dy, i) = 
          if intends i (I_JUMP)
             (* something to push off of.. *)
             andalso Clip.clipped (!room) clip (x, y + 1)
          then (dy - JUMP_VELOCITY,
                List.filter (fn I_JUMP => false | _ => true) i)
          else (dy, List.filter (fn I_JUMP => false | _ => true) i)
        *)
        infix --
        fun (x1, y1) -- (x2, y2) = (x1 - x2, y1 - y2)
        infix ++
        fun (x1, y1) ++ (x2, y2) = (x1 + x2, y1 + y2)
        infix **
        fun a ** (x, y) = (a * x, a * y)

        val ((dx, dy), i) =
          if intends i (I_JUMP)
             andalso (print ("jumping: current orientation is " ^ dirtos (!snaildir) ^ "\n"); true)
             (* something to push off of.. *)
             andalso ( Clip.clipped (!room) clip ((x, y) ++ downwards)
                       orelse (print "can't jump here!\n"; false) )
          then ((dx, dy) ++ (JUMP_VELOCITY ** upwards),
                List.filter (fn I_JUMP => false | _ => true) i)
          else ((dx, dy), List.filter (fn I_JUMP => false | _ => true) i)

        (* XXX not fixing these for orientation --wjl *)
        (* superjumps are cheat *)
        val (dy, i) = 
          if intends i (I_SUPERJUMP)
          then (dy - (JUMP_VELOCITY * 2),
                List.filter (fn I_SUPERJUMP => false | _ => true) i)
          else (dy, i)


        (* Get gravity from currents.
           This should only apply if we are in freefall mode; otherwise
           it should come from our attachment. *)
        val (dx, dy) =
          case !slime of
            SOME _ => (dx, dy) ++ downwards
          | NONE =>
            let val (ox, oy) = Clip.center clip
                val cx = x + ox
                val cy = y + oy
            in
                (* XXX could average/vote in nearby squares... *)
                case propat (!room, cx div TILEW, cy div TILEH) of
                    PCURRENT UP => (dx, dy - 1)
                  | PCURRENT LEFT => (dx - 2, dy + 1)
                  | PCURRENT RIGHT => (dx + 2, dy - 1)
                  | PCURRENT DOWN => (dx, dy + 2)
                  | _ => (dx, dy + 1)
            end
            
        (* water resistance *)
        val dx = if dx > TERMINAL_VELOCITY
                 then dx - 1
                 else if dx < ~ TERMINAL_VELOCITY
                      then dx + 1 
                      else dx

        val dy = if dy > TERMINAL_VELOCITY
                 then dy - 1
                 else if dy < ~ TERMINAL_VELOCITY
                      then dy + 1
                      else dy

        (* XXX world effects... *)

        val (x, y, dx, dy) =
          let
              (* val xi = if dx < 0 then ~1 else 1 *)
              (* val yi = if dy < 0 then ~1 else 1 *)

              (* jcreed *)
              (* intended invariant: dv * (vix, viy) + dh * (hix, hiy) = (dx, dy) *)
              val (dh, dv) = case !snaildir of 
                                 DOWN => (abs dx, abs dy)
                               | UP => (abs dx, abs dy)
                               | LEFT => (abs dy, abs dx)
                               | RIGHT => (abs dy, abs dx)
              val (hix, hiy, vix, viy) = case !snaildir of
                                             DOWN => (sgn dx, 0, 0, sgn dy)
                                           | UP => (sgn dx, 0, 0, sgn dy)
                                           | LEFT => (0, sgn dy, sgn dx, 0)
                                           | RIGHT => (0, sgn dy, sgn dx, 0)   


              val xr = ref x
              val yr = ref y
              val stopx = ref false
              val stopy = ref false

              (* Allow vertical movement if walking up a gradient *)
              (*
              fun climb () =
                (yr := !yr - 1;
                 true)
              *)

              fun nudgeh _ = 
                  let val nextx = !xr + hix
                      val nexty = !yr + hiy
                      val next = (nextx, nexty)
                  in
                      if !stopx 
                      then (* SOME (!xr + 1, !yr + 1) *) NONE (* () *)
                      else if (* can move right:

                                 x .
                              *)
                              Clip.clipped (!room) clip next
                           then 
                               (* can move up a little if
                                  we are on a ramp:

                                  . .
                                  x #

                                  *)
                              (* XXX jcreed rotatable nudgeclimbing *)
                               if not (Clip.clipped (!room) clip (!xr, !yr - 1)
                                       orelse
                                       Clip.clipped (!room) clip (nextx, !yr - 1))
                                  (* andalso climb() *)
                               then SOME (nextx, !yr - 1) (* xr := next *)
                               else (stopx := true; (* SOME (!xr - 1, !yr - 1) *) NONE) (* stopx := true *)
                           else SOME (nextx, !yr) (* xr := next *)
                  end

              (* XXX *)
              fun flat_topped attachtile = attachtile = MSOLID

              (* XXX only dealing with downward orientation currently *)
              fun try_attach () =
                  case !slime of
                    SOME _ => ()
                  | NONE =>
                      let val sx = !xr + (3 * TILEW) div 2
                          val sy = !yr + (2 * TILEH)
                          val attachtile = Room.maskat (!room,
                                                        sx div TILEW,
                                                        sy div TILEH)
                      in
                          if flat_topped attachtile
                          (* should keep a local version XXX *)
                          then slime := SOME (sx, sy)
                          else ()
                      end

              fun nudgev _ = 
                  let val nexty = !yr + viy
                  in
                      if !stopy 
                      then ()
                      else if Clip.clipped (!room) clip (!xr, nexty)
                      then (stopy := true;
                            try_attach ())
                      else (yr := nexty;
                            (* detach *)
                            if viy = ~1 (* XXX jcreed *)
                            then ((* print "Detaching in nudgev\n"; *)
                                  slime := NONE)
                            else ())
                  end


              fun safeDetach (x,y) = true

              fun slimeh _ =
                  (* XXX needs to sometimes change direction *)
                  case !slime of
                    NONE => NONE
                  | SOME (sx, sy) =>
                  let
                      (* val next = !xr + xi *)

                      fun clipAt (x, y) =
                          Tile.clipmask (Room.maskat (!room, x div TILEW,
                                                             y div TILEH))
                                        (x mod TILEW, y mod TILEH)

                      fun try3 [p1, p2, p3] =
                          if      clipAt p1 then SOME p1
                          else if clipAt p2 then SOME p2
                          else if clipAt p3 then SOME p3
                          else NONE (* slime fell off the wall :( *)
                        | try3 _ = raise Fail "try3 called with non-triple"

                      fun offset s =
                          let
                              val (dx, dy) = to_vector (!snaildir)
                          in
                              (* (sx + dx * s + xi, sy + dy * s) *)
                              (* XXX jcreed *)
                              (sx + dx * s + hix, sy + dy * s + hiy)
                              (* hix and hiy are the spiffy new intention
                                 vector components *)
                          end

                      fun proposeCandidate () = try3 (map offset [~1, 0, 1])

                      fun sameTile (x, y) = sx div TILEW = x div TILEW
                                    andalso sy div TILEH = y div TILEH

                      (*
                      fun getAngle (x, y) = raise Match

                      fun resolveTransition p =
                          let val a1 = getAngle (sx, sy)
                              val a2 = getAngle (sx + xi, sy)
                          in
                              raise Fail "you have to put a function here"
                          end
                      *)

                      fun goodAngle (x, y) = true

                      fun reorient (x, y) d =
                        let val oldtile = maskat (!room, sx div TILEW, sy div TILEH)
                            val newtile = maskat (!room, x div TILEW, y div TILEH)
                        in
                            case (oldtile, newtile, d) of
                                (MRAMP MH, MRIGHT MH, DOWN) => (print "reorient right\n"; RIGHT)
                              | (MRIGHT MH, MRAMP MH, RIGHT) => (print "reorient down\n"; DOWN)
                              | _ => d
                        end

              (* progressSlime : pos * dir -> dir -> (pos * dir) option
                 XXX needs an orientation as well?
                 progressSlime d (x, y) = SOME (d', (x', y')) iff we moved successfully,
               = NONE iff we can't -- corner too pointy,
               or we're stuck
                  progressSlime returns an orientation for the snail because it
                  needs to take care of the flip, due to canonization. *)
                    val blue = SDL.color (0w0, 0w0, 0wxFF, 0wxFF)
                    fun plot (x, y) = print ("slimex suggests " ^ Int.toString x
                                            ^ ", " ^ Int.toString y ^ "\n")
                                      (*(drawcircle (screen, x - !scrollx,
                                                   y - !scrolly, 3, blue);
                                       flip screen)*)
                  in
                      case proposeCandidate () of
                          SOME p => ((* plot p; *)
                                    if sameTile p orelse goodAngle p
                                    then SOME (p, reorient p (!snaildir))
                                    else NONE
                                    )
                        | NONE => NONE
                  end

              fun bottomRightPixel () =
                if !snaildir = DOWN orelse !snaildir = UP
                then (* horiz *) (!xr + 3*TILEW - 1, !yr + 2*TILEH - 1)
                else (* vert *)  (!xr + 2*TILEW - 1, !yr + 3*TILEH - 1)

              fun wedged () =
                  (* back/front is left/right *)
                  let val (brx, bry) = bottomRightPixel ()
                      val (right_x, right_y) = (brx + 1, bry)
                      val (lower_x, lower_y) = (brx, bry + 1)
                  in
                      Room.maskat (!room, right_x div TILEW, right_y div TILEH)
                        = MRIGHT MH
                      andalso
                      Room.maskat (!room, lower_x div TILEW, lower_y div TILEH)
                        = MRAMP MH
                  end
                  (* orelse ... wedged on the back ... *)

              fun rotateTo d =
                let fun tryRotate (x, y, clip) =
                        if Clip.clipped (!room) clip (x, y)
                        then NONE
                        else SOME (x, y)
                in
                    tryRotate
                    (case (!snaildir, d) of
                        (DOWN, RIGHT) => (!xr + TILEW, !yr - TILEH, Clip.vert)
                      | (RIGHT, DOWN) => (!xr - TILEW, !yr - TILEH, Clip.horiz))
                        (* XXX this bottom one doesn't seem to flip correctly? *)
                      (* XXX more cases *)
                end

              fun slimeCanonical (sx, sy) (nx, ny) DOWN = sx = nx + (3 * TILEW) div 2
                | slimeCanonical (sx, sy) (nx, ny) UP   = sx = nx + (3 * TILEW) div 2
                | slimeCanonical (sx, sy) (nx, ny) RIGHT = sy = ny + (3 * TILEW) div 2
                | slimeCanonical (sx, sy) (nx, ny) LEFT  = sy = ny + (3 * TILEW) div 2
                (* XXX UP and LEFT cases correct? *)

              fun hyperh i =
                  if wedged () then
                      (* CONCAVE MAGIC:
                        - let the slime move into the corner while snail stays
                        - if it crosses the inflection point, rotate the snail
                        - if it moves back out, we exit magic mode
                      *)
                     (slimecolor := brightred ;
                      case (slimeh i, nudgeh i) of
                        (SOME ((sx,sy), d), NONE) =>
                            (* let slime move without snail *)
                            if d <> !snaildir
                            (* XXX rotateTo specialized to concave case *)
                            then case rotateTo d of
                                    SOME (snx, sny) => (* perform flip *)
                                                       (print("performing flip "
                                                             ^ dirtos d ^ "\n");
                                                        xr := snx; yr := sny;
                                                        slime := SOME (sx, sy);
                                                        snaildir := d)
                                  | NONE => (print "failed\n"; ()) (* cannot go on *)
                            else (* dir unchanged *)
                                 slime := SOME (sx, sy)
                      | (NONE, NONE) =>
                            (* detach, don't move -- impossible? *)
                            if safeDetach (!xr, !yr)
                            then (print "Weird!\n"; slime := NONE)
                            else ()
                      | (NONE, SOME (x, y)) =>
                            (* just like below *)
                            if safeDetach (x,y) 
                            then (xr := x; yr := y;
                                  (* print "Detaching in hyperh 1\n"; *)
                                  slime := NONE) 
                            else ()
                      | (SOME ((sx, sy), d), SOME (nx, ny)) =>
                             (* do what the slime says *)
                            (if true (* safeFlip d *) (* XXX XXX WHAT *)
                             then (slime := SOME (sx, sy)
                                   (* snaildir := d *)
                                   (* XXX do flip?  (NB: only in convex case) *)
                                   )
                             else ()
                             ;
                             (* not handling convex yet *)
                             if d <> !snaildir then print "ACK!\n" else ()
                             ;
                             (* only move snail if slime is where it belongs *)
                             if slimeCanonical (sx, sy) (nx, ny) d
                             then (xr := nx; yr := ny)
                             else ()
                            )
                     )
                  else
                     (slimecolor := brightgreen ;
                      case (slimeh i, nudgeh i) of 
                          (_, NONE) => () (* FUCKIT principle *)
                        | (NONE, SOME (x, y)) =>
                            if safeDetach (x,y) 
                            then (xr := x; yr := y;
                                  (* print "Detaching in hyperh 2\n"; *)
                                  slime := NONE) 
                            else ()
                        | (SOME ((sx,sy),d), SOME(nx, ny)) =>
                            (xr := nx; yr := ny;
                             slime := SOME (sx, sy);
                             if !snaildir <> d
                                then (print "why is this here what?\n";
                                      snaildir := d)
                                else ())
                     )

              fun hyperv i = nudgev i

          in
              (* try moving one pixel at a time
                 in the desired direction, until
                 it is not possible to move any 
                 more. *)

              (* XXX this calculation should probably
                 be symmetric. Here we do all of our
                 X movement before considering the Y
                 direction... *)

              (*
              Util.for 0 (abs dx - 1) nudgeh;
              Util.for 0 (abs dy - 1) nudgev;
              *)
              Util.for 0 (dh - 1) hyperh;
              Util.for 0 (dv - 1) hyperv;

              (!xr, !yr, 
               if !stopx then 0 else dx, 
               if !stopy then 0 else dy)
          end
    in
        (i, x, y, dx, dy)
    end 

  fun movesnail { nexttick, intention } =
      (* Don't move snail when editing. *)
      if !showedit
      then intention
      else
      let
          val () = if List.exists (fn (I_GO _) => true | _ => false) intention
                   then raise Game "snail should never have long-lived GO intentions"
                   else ()

          (* instantaneous intention is derived from the intention queue
             and the keys being held *)
          val intention_now = contextual_intention (!keymask) @ intention

          (* NB: clipping shape depends on orientation *)
          val clip = case !snaildir of
                        DOWN => Clip.horiz
                      | UP => Clip.horiz
                      | LEFT => Clip.vert
                      | RIGHT => Clip.vert
          val (intent, x, y, dx, dy) = 
              moveent clip (intention_now, !snailx, !snaily, !snaildx, !snaildy)

          (* Only intention we save is jumping. *)
          val intent = List.filter (fn I_JUMP => true | _ => false) intent
      in
          snailx := x;
          snaily := y;
          snaildx := dx;
          snaildy := dy;
          intent
      end

  (* draws the actual clip mask calculated pixel-by-pixel. slow. *)
  fun drawdebug () =
    if !showclip
    then 
      let in
        (* print "DEBUG..\n"; *)
        Util.for 0 (width - 1)
        (fn x => 
         let in
           Util.for 0 (height - 1)
           (fn y =>
            let
              val xx = !scrollx + x
              val yy = !scrolly + y
              val tx = xx div TILEW
              val ty = yy div TILEH
            in
              (* tile mask, absolute *)
              case Room.maskat (!room, tx, ty) of
                MEMPTY => ()
              | m => 
                  if Tile.clipmask m (xx mod TILEW, yy mod TILEH)
                  then SDL.drawpixel (screen, x, y, color (0w255, 0w0, 0w0, 0w255))
                  else SDL.drawpixel (screen, x, y, color (0w0, 0w255, 0w0, 0w255))

              (* (* clip mask in config space *)
              if Clip.clipped (!room) Clip.std (!scrollx + x, !scrolly + y)
              then SDL.drawpixel (screen, x, y, color (0w255, 0w0, 0w0, 0w255))
              else SDL.drawpixel (screen, x, y, color (0w0, 0w255, 0w0, 0w255))
              *)
            end)
(*          ; if x mod 16 = 0 then flip screen else () *)
         end)
        (* print "DONE.\n" *)
      end
    else ()

  fun moveobjects () =
      let
          fun doobj (Bullet { x, y, dx, dy }) =
              let 
                  (* XX only dx supported now *)
                  val xi = if dx < 0 then ~1 else 1
                  val xr = ref x
                  val stopx = ref false
                  fun nudgeh _ = 
                      let val next = !xr + xi
                      in
                          if !stopx orelse Clip.clipped (!room) Clip.bullet (next, y)
                          then stopx := true
                          else xr := next
                      end
              in
                  Util.for 0 (abs dx - 1) nudgeh;
                  if !stopx then NONE
                  else SOME (Bullet { x = !xr, y = y, dx = dx, dy = dy })
              end
            | doobj (Monster { which, dir, x, y, dx, dy }) =
              let
                val clip = 
                    case which of
                        STAR => Clip.star
                      | SPIDER => Clip.spider
                val (_, x, y, dx, dy) = moveent clip ([I_GO dir, I_JUMP], x, y, dx, dy)
                val dir = if dx = 0
                          then dir_reverse dir
                          else dir

              in
                (* never goes away *)
                SOME (Monster { which = which, dir = dir, x = x, y = y, dx = dx, dy = dy })
              end
      in
          objects := List.mapPartial doobj (!objects)
      end

  fun drawmaskzone () =
      case (!showedit, !showmask, !editmode) of
          (true, true, _) => drawzone DRAWMASK
        | (true, _, ZMASK) => drawzone DRAWMASK
        | _ => ()

  fun drawpropzone () =
      case (!showedit, !showprop, !editmode) of
          (true, true, _) => drawzone DRAWPROP
        | (true, _, ZPROP) => drawzone DRAWPROP
        | _ => ()

  fun drawlightmap () =
      case Room.lightmap (!room) of
          NONE => ()
        | SOME m => SDL.blit (m, !scrollx, !scrolly, width, height, screen, 0, 0)

  (* The idea is that this should give the snail's center *)
  fun snailpos () = 
      let 
        (* Gives pretty desirable desired rounding behavior *)
        (* XXX will be messed up by rotation *)
        val sx = (!snailx + SNAILW div 2) div TILEW
        val sy = (!snaily + SNAILH div 2) div TILEH 
      in
        (sx, sy)
      end

  fun checkaction () = 
      let 
        val (sx, sy) = snailpos ()
      in
        case Room.actionat (!room, sx, sy) of
          Say s => message := "\"^5" ^ s ^ "^<\""
        | _ => ()
      end

  fun consistent_dir (UP, dx, dy) = dy < 0
    | consistent_dir (DOWN, dx, dy) = dy > 0
    | consistent_dir (LEFT, dx, dy) = dx < 0
    | consistent_dir (RIGHT, dx, dy) = dx > 0

  fun say s =
      let in
          print (s ^ "\n");
          message := s
      end

  fun doornums 0 = "red"
    | doornums 1 = "green"
    | doornums 2 = "pink"
    | doornums 3 = "yellow"
    | doornums _ = "ILLEGAL"

  exception FoundDoor of int * int
  (* This finds the top-leftmost door that matches. That's crap because
     doors almost always have width or height or both. Slightly better would
     be to find the center of mass of all matching doors. Best would be to
     figure out the geometry of the source dir, the player's position as he
     entered it, and the geometry of the out door, and then put the player
     in the corresponding place. Since the doors are made up of tiles you
     have to do some battleship shit to figure out the geometry. *)
  fun find_door (desired_dir, desired_idx) =
      let in
          Util.for 0 (Room.ROOMH - 1)
          (fn y =>
           Util.for 0 (Room.ROOMW - 1)
           (fn x =>
            case Room.propat (!room, x, y) of
                PGOTO (d, i) =>
                    if d = desired_dir andalso
                       i = desired_idx
                    then raise FoundDoor (x, y)
                    else ()
              | _ => ()));
          NONE
      end handle FoundDoor p => SOME p

  (* XXX should also take pos, and preserve it in the new door. *)
  (* FIXME should check if the player would be stuck in the new position,
     to preserve invariant! *)
  fun warp (newroom, srcdir, dooridx) =
      let 
          val matching_dir = dir_reverse srcdir
          fun move_to_door () =
              case find_door (matching_dir, dooridx) of
                  NONE =>
                      let in
                          say "^4There's no matching door in this room! You should make one...";
                          snailx := TILEW * 20;
                          snaily := TILEH * 20
                      end
                | SOME (x, y) =>
                      let 
                          (* Center of destination tile. *)
                          val cx = TILEW * x + (TILEW div 2)
                          val cy = TILEH * y + (TILEH div 2)

                          (* Center of the tile right AFTER the door, using the
                             src direction. We want currents to be able to carry the player,
                             but they can't overlap doors. *)

                          val cx = cx + dir_dx srcdir * TILEW
                          val cy = cy + dir_dy srcdir * TILEH

                          (* But we want to put the center of this on the snail's center, not
                             the top-left corner. *)
                          (* XXX should check orientation?  or set it? *)
                          val (wx, wy) = Clip.center Clip.horiz
                          val sx = cx - wx
                          val sy = cy - wy
                      in
                          snailx := sx;
                          snaily := sy
                      end
      in
          say "^3LET'S DO THE ROOM WARP AGAIN!";
          room := newroom;
          (* find the corresponding door. *)
          move_to_door()
      end

  fun checkwarp () =
      let
        val (sx, sy) = snailpos ()
      in
        case Room.propat (!room, sx, sy) of
            PGOTO (dir, i) =>
              (* Have to have consistent velocity to enter. *)
              if consistent_dir (dir, !snaildx, !snaildy)
              then (case Room.dirfrom (!room) dir of
                      NONE => say "OMG there is no room in that direction!"
                    | SOME newroom => warp (newroom, dir, i))
              else say "go in the correct direction to enter the door"
          | _ => ()
      end

  fun checkmouseover () = ()

  exception QUITTER

  val initial_cutscene =
      map requireimage
          ["cutscene-1-1.png", "cutscene-1-2.png", "cutscene-1-3.png"]

  fun cutscene (music, musicfn, imgs) = 
    let fun loop [] = ()
          | loop (img :: imgs) = 
            let in
              clearsurface (screen, color (0w255, 0w99, 0w255, 0w255));
              SDL.blitall (img, screen, 0, 0);
              flip screen; 
              case pollevent () of 
                SOME (E_KeyDown _) => loop imgs
              | SOME (E_MouseDown _) => loop imgs
              | SOME E_Quit => raise QUITTER
              | _ => loop (img :: imgs)
            end
    in
      musicfn music;
      (* Start music *)
      loop imgs
      (* End music *)
    end

  fun loop { nexttick, intention } =
      if getticks () > nexttick
         andalso (not (!paused) orelse !advance)
      then 
          let
              (* Reset message, determine if mouseover/action resets it.
                 XXX Rob fix *)
              val () =               message := 
                                     "^1Nemo 911^<: ^3Anemonies As Weapons^<!"

              (* Before we do anything, check the action. It might warp us to
                 a new room. *)
              val () =               checkwarp ()
              val () =               checkaction ()

              val () =               clearsurface (screen, color (0w255, 0w33, 0w35, 0w99))
              val () =               setscroll intention
              val () =               drawzone (DRAW BACKGROUND)
              val () =               drawdebug ()
              val () =               drawobjects ()
              val () =               drawsnail true (* XXX shadow *)
              val intention = movesnail { nexttick = nexttick, intention = intention }

              val () =               moveobjects ()
              val () =               drawsnail false
              val () =               drawzone (DRAW FOREGROUND)
              val () =               drawlightmap ()
              val () =               drawmaskzone ()
              val () =               drawpropzone ()
              val () =               drawedit ()

              val () =               checkmouseover ()
              val () =               Font.draw(screen, 10, 10, !message)
          in
              flip screen;
              advance := false;
              Animate.increment ();
              key { nexttick = getticks() + 0w20, intention = intention }
          end
      else
          let in
              SDL.delay 1;
              key { nexttick = nexttick, intention = intention }
          end
              
  and click (x, y) =
    (case onscreen (x, y) of
         NONE => print "offscreen click\n"
       | SOME (tx, ty) => 
             let in
                 (* depend on the layer we're currently editing *)
                 case !editmode of
                     ZMASK => Room.setmask (!room, tx, ty) (!editmask)
                   (* XXX should set default mask here *)
                   | ZBG => Room.settile (!room, BACKGROUND, tx, ty) (!editbgtile)
                   | ZFG => Room.settile (!room, FOREGROUND, tx, ty) (!editfgtile)
                   | ZPROP => Room.setprop (!room, tx, ty) (!editprop)
                   | ZBRUSH => Room.setwithbrush (!room, #1 (!editbrush), #2 (!editbrush), tx, ty)
             end)

  and key ( cur as { nexttick, intention = i } ) = 
    case pollevent () of
       NONE => loop cur
     | SOME evt =>
       (* XXX should pass to window manager *)
       if !showedit andalso Window.handleevent edit_window evt
       then loop cur
       else case evt of
      E_KeyDown { sym = SDLK_ESCAPE } => () (* quit *)

    | E_MouseDown { x, y, ... } =>
           let in
               click (x, y);
               loop cur
           end
       
    (* You have to update in the parent directory to get SDL.mouse_left. *)
    | E_MouseMotion { state : mousestate, x : int, y : int, xrel, yrel, ... } =>
           (if SDL.mouse_left state
            then click (x, y)
            else ();
            loop cur)

    | E_KeyDown { sym = SDLK_LSHIFT } => 
         let in 
             keymask := Word32.orb(!keymask, KEYMASK_SHIFT);
             loop cur
         end

    | E_KeyDown { sym = SDLK_RSHIFT } => 
         let in 
             keymask := Word32.orb(!keymask, KEYMASK_SHIFT);
             loop cur
         end

    | E_KeyDown { sym = SDLK_LCTRL } => 
         let in 
             keymask := Word32.orb(!keymask, KEYMASK_CTRL);
             loop cur
         end

    | E_KeyDown { sym = SDLK_RCTRL } => 
         let in 
             keymask := Word32.orb(!keymask, KEYMASK_CTRL);
             loop cur
         end

(*
        let in
            (* fire bullet *)
            objects := Bullet { x = (case !snailface of
                                         FLEFT => !snailx - (surface_width fireballl)
                                       | FRIGHT => !snailx + surface_width snailr),
                                y = !snaily + 8,
                                dy = 0,
                                dx = (case !snailface of
                                          FLEFT => ~8
                                        | FRIGHT => 8) } :: !objects;
            loop cur
        end
*)

    | E_KeyDown { sym = SDLK_m } =>
        let in
            if !showedit
            then showmask := not (!showmask)
            else ();
            loop cur
        end

    | E_KeyDown { sym = SDLK_p } =>
        let in
            if !showedit
            then showprop := not (!showprop)
            else ();
            loop cur
        end

    (*
    | E_KeyDown { sym = SDLK_COMMA } => 
        let in
          objects := Monster { which = STAR,
                               x = (case !snailface of
                                      FLEFT => !snailx - (surface_width star)
                                    | FRIGHT => !snailx + surface_width snailr),
                               dir = (case !snailface of
                                      FLEFT => LEFT
                                    | FRIGHT => RIGHT),
                               y = !snaily - 4,
                               dy = ~4,
                               dx = (case !snailface of
                                       FLEFT => ~8
                                     | FRIGHT => 8) } :: !objects;
          loop cur
        end

    | E_KeyDown { sym = SDLK_n } => 
        let in
          objects := Monster { which = SPIDER,
                               x = (case !snailface of
                                      FLEFT => !snailx - (surface_width star)
                                    | FRIGHT => !snailx + surface_width snailr),
                               dir = (case !snailface of
                                      FLEFT => LEFT
                                    | FRIGHT => RIGHT),
                               y = !snaily - 4,
                               dy = ~4,
                               dx = (case !snailface of
                                       FLEFT => ~8
                                     | FRIGHT => 8) } :: !objects;
          loop cur
        end
    *)


    | E_KeyDown { sym = SDLK_RETURN } =>
        let in
          paused := not (!paused);
          loop cur
        end

    | E_KeyDown { sym = SDLK_PERIOD } => 
        let in
          advance := true;
          loop cur
        end

    | E_KeyDown { sym = SDLK_s } =>
        let in
          Room.save (!room);
          print "SAVED.\n";
          loop cur
        end

    | E_KeyDown { sym = SDLK_PAGEDOWN } =>
        let in
          edittileset := (!edittileset + 1) mod NUM_EDIT_TILESETS;
          loop cur
        end

    | E_KeyDown { sym = SDLK_PAGEUP } =>
        let in
          edittileset := (!edittileset - 1) mod NUM_EDIT_TILESETS;
          loop cur
        end

    | E_KeyDown { sym = SDLK_b } => 
        loop { nexttick = nexttick, intention = I_BOOST :: i }

    | E_KeyDown { sym = SDLK_UP } => 
         let in 
             keymask := Word32.orb(!keymask, KEYMASK_UP);
             loop cur
         end

    | E_KeyDown { sym = SDLK_DOWN } => 
         let in 
             keymask := Word32.orb(!keymask, KEYMASK_DOWN);
             loop cur
         end

    | E_KeyDown { sym = SDLK_LEFT } =>
        (* XXX for these there should also be an impulse event,
           so that when I tap the left key ever so quickly, I at least
           start moving left for one frame. *)
         let in 
             snailface := FLEFT; (* FIXME *)
             keymask := Word32.orb(!keymask, KEYMASK_LEFT);
             loop cur
         end

    | E_KeyDown { sym = SDLK_RIGHT } =>
         let in 
             snailface := FRIGHT; (* FIXME *)
             keymask := Word32.orb(!keymask, KEYMASK_RIGHT);
             loop cur
         end

    | E_KeyDown { sym = SDLK_SPACE } =>
          (* XXX in edit mode, scroll to snail *)
          if intends i (I_JUMP)
          then loop { nexttick = nexttick, intention = i }
          else loop { nexttick = nexttick, intention = (I_JUMP :: i) }


    | E_KeyDown { sym = SDLK_e } => 
          (showedit := not (!showedit);
           loop cur)

    | E_KeyDown { sym = SDLK_z } => 
          (editmode :=
           (case !editmode of
                ZMASK => ZPROP
              | ZPROP => ZBRUSH
              | ZBRUSH => ZBG
              | ZBG => ZFG
              | ZFG => ZMASK);
           loop cur)

    | E_KeyDown { sym = SDLK_c } => 
          (showclip := not (!showclip);
           loop cur)

    | E_KeyUp { sym = SDLK_UP } =>
         let in 
             keymask := Word32.andb(!keymask, Word32.notb KEYMASK_UP);
             loop cur
         end

    | E_KeyUp { sym = SDLK_DOWN } =>
         let in 
             keymask := Word32.andb(!keymask, Word32.notb KEYMASK_DOWN);
             loop cur
         end

    | E_KeyUp { sym = SDLK_LEFT } =>
         let in 
             keymask := Word32.andb(!keymask, Word32.notb KEYMASK_LEFT);
             loop cur
         end

    | E_KeyUp { sym = SDLK_RIGHT } =>
         let in 
             keymask := Word32.andb(!keymask, Word32.notb KEYMASK_RIGHT);
             loop cur
         end
    
    (* Was added by cmartens for character creation screen
    | E_KeyDown { sym = SDLK_0 } => 
              let
                val () = CharCreat.loop ()
              in
                loop { nexttick = nexttick, intention = i }
              end
    *)

    | E_KeyDown { sym } =>
          let in
              print ("unknown key " ^ SDL.sdlktos sym ^ "\n");
              loop { nexttick = nexttick, intention = i }
          end

(* XXX should do the same thing with joystick as with keys, above.
    | E_JoyDown { button = 3, ... } =>
          if intends i (I_JUMP)
          then loop { nexttick = nexttick, intention = i }
          else loop { nexttick = nexttick, intention = (I_JUMP :: i) }

    | E_JoyDown { button = 1, ... } =>
          (snailface := FRIGHT;
           if intends i (I_GO RIGHT)
           then loop { nexttick = nexttick, intention = i }
           else loop { nexttick = nexttick, intention = I_GO RIGHT :: i })

    | E_JoyUp { button = 1, ... } =>
            (loop { nexttick = nexttick, 
                    intention = List.filter (fn I_GO RIGHT => false | _ => true) i })

    | E_JoyUp { button = 0, ... } =>
            (loop { nexttick = nexttick, 
                    intention = List.filter (fn I_GO LEFT => false | _ => true) i })

    | E_JoyDown { which, button } =>  
            let in
                print ("unknown joydown " ^ Int.toString which ^ ":" ^
                       Int.toString button ^ "\n");
                loop cur
            end
    | E_JoyUp { which, button } =>  
            let in
                print ("unknown joyup " ^ Int.toString which ^ ":" ^
                       Int.toString button ^ "\n");
                loop cur
            end
*)

    | E_Quit => ()
    | _ => loop cur

  (* Main entry point *)

  val () = print "About to enter main loop.\n"

  val () =
      (cutscene (Music.theme, SDLMusic.loop, initial_cutscene); 
       loop { nexttick = 0w0, intention = nil })
    handle Game s => messagebox ("nemo 911: " ^ s)
        | SDL s => messagebox ("sdl error: " ^ s)
        | SDLMix.Error s => messagebox ("sdlmix error: " ^ s)
        | e => messagebox ("Uncaught exception: " ^ exnName e ^ " / " ^ exnMessage e)
end
