(* The world is populated by tiles. 

   There are foreground and background graphics (type tile), which
   come from tiles.png. These are 16x16. There are some conventions
   about the layout in that graphic, so that the brushes can generate
   nice corners and randomness by knowing about a group of tiles that
   work together (called a material). Some tiles are magic, in that
   they animate as frames progress, or are even drawn programatically.

   In addition to tiles, we have the clip mask, which is the solid
   geometry of the world.
   
   We also have properties, which is stuff like currents, transitions
   between rooms, and actions.
   *)

signature TILE =
sig

  exception Tile of string

  (* the mask tells us where things can walk *)
  datatype slope = LM | MH | HM | ML
  (* XXX probably dead to us *)
  datatype diag = NW | NE | SW | SE
  datatype mask = 
      MEMPTY 
    | MSOLID 
    | MRAMP of slope
    | MCEIL of slope
    | MDIAG of diag (* corner that is filled in *)
    | MLEFT of slope
    | MRIGHT of slope

  datatype brush =
      BRECT of int * int (* your basic rectangle (w x h) *)
    | BSLOPEDRECT of int * int (* (min slope) (w x h) *)
    | BDIAMOND of int * int (* (max slope) (w x h) *)
    | BERASERECT of int * int (* rectangular eraser *)
    | BOCTAGON of int * int (* (med slope) (w x h) *)
    | BHILLLH of int * int (* (l * h) *)   
    | BHILLHL of int * int (* (l * h) *)   
    | BPASTE

  datatype lightcolor = LWHITE | LYELLOW | LBLUE | LRED | LGREEN
  type attenuation = { constant: real, linear:real, quadratic: real }
  type lightdata = { red: Word8.word, green: Word8.word, blue: Word8.word, intensity: real, attenuation: attenuation }
  (* Intensity of light emission, in each color channel. (0, 0, 0) is
     equivalent to no light being present. *)
  val lightcolor : lightcolor -> lightdata

  datatype prop =
      PPLAIN
      (* there are up to four doors of each dir, colorized *)
    | PGOTO of Types.dir * int
    | PCURRENT of Types.dir
    | PACTION of int
    | PLIGHT of lightcolor

  type tile

  type material

  val TILEW : int
  val TILEH : int

  (* in tiles. *)
  val TILESET_WIDTH : int
  val TILESET_HEIGHT : int

  val select_tile : tile
  val eraser_material : material
  val dirt_material : material
  val block_material : material
  val scribble_material : material

  (*val closest_material : tile -> material*)

  val brushsize : brush -> int * int
  (* is this pixel of the mask clipped? *)
  val clipmask : mask -> int * int -> bool

  val masktostring : mask -> string

  (* drawat timestep tile surf pixel-x pixel-y  world-x world-y
     at a particular time step,
     draw the tile to the given surface at the
     given pixel coordinates.
     most tiles ignore world-x and world-y; they are used for
     effects like starfields, randomization, parallax, etc.
     *)
  val drawat : int * tile * SDL.surface * int * int * int * int -> unit
  (* or right now *)
  val draw : tile * SDL.surface * int * int * int * int -> unit

  (* These are for the editor, not in-game *)
  val drawmask : mask * SDL.surface * int * int -> unit
  val drawprop : prop * SDL.surface * int * int -> unit
  val drawselection : SDL.surface * int * int -> unit

  (* Needed for other parts of the code *) 
  val masktotilem : (material * mask) -> tile
  val masktotile : (tile * mask) -> tile
  val brushtomasks : brush -> (mask * int * int) list

  (* for serialization *)
  val toword   : tile -> Word32.word
  val fromword : Word32.word -> tile

  (* Tile to use when viewing out of the room's boundaries. 
     Would be useful to have this be configurable in the room itself. *)
  val OUTER_SPACE_BACKGROUND : tile
  val OUTER_SPACE_FOREGROUND : tile

  (* Plain, though GOTO could be reasonable? *)
  val OUTER_SPACE_PROP : prop

end
