
(* Most of the game logic, and the main game loop, are in here. *)

structure Game =
struct

  structure AnemoneSet = SplaySetFn (struct 
                                      type ord_key = string 
                                      val compare = String.compare 
                                     end)

  type ptr = MLton.Pointer.t

  exception Nope

  exception Game of string

  (* XXX urg, bad choice *)
  structure U = Util
  open SDL
  structure Util = U

  fun abs z = if z < 0 then ~ z else z
  fun sgn x = if x < 0 then ~1 else 1

  fun messagebox s = 
      let
        (* val mb = _import "MessageBoxA" stdcall : int * string * string * int -> unit ; *)
      in
        (* mb (0, s ^ "\000", "message", 0) *)
        print (s ^ "\n")
      end

  val showclip = ref false

  (* Allowing resizing of the screen, so that we can edit at high
     resolutions. We can and should assume a constant 800x600 resolution
     for the player. *)
  val width = Params.asint 800 Args.width
  val height = Params.asint 600 Args.height

  val TILESW = width div Tile.TILEW
  val TILESH = height div Tile.TILEH

  val SNAILH = 32
  val SNAILW = 48

  open Types
  open Tile
  open Room

  val screen = makescreen (width, height)

  (* A crawler has a set of images, for each attachment, and for
     each orientation in that attachment. This takes a basename
     like "snail" and loads specialized graphics for each of the
     directions, or synthesizes them if they don't exist. The
     rules for synthesis are:

      1. There is no synthesis yet.
      *)
  fun require_crawl_images s =
      let
          val floorl = Images.requireimage ("floor" ^ s ^ "l.png")
          val floorr = Images.requireimage ("floor" ^ s ^ "r.png")
          val ceill = Images.requireimage ("ceil" ^ s ^ "l.png")
          val ceilr = Images.requireimage ("ceil" ^ s ^ "r.png")

          val leftu = Images.requireimage ("left" ^ s ^ "u.png")
          val leftd = Images.requireimage ("left" ^ s ^ "d.png")
          val rightu = Images.requireimage ("right" ^ s ^ "u.png")
          val rightd = Images.requireimage ("right" ^ s ^ "d.png")

      in
          { floorl = floorl,
            floorr = floorr,
            ceill = ceill,
            ceilr = ceilr,
            leftu = leftu,
            leftd = leftd,
            rightu = rightu,
            rightd = rightd }
          (* XXX should add falling versions of these. *)
      end


  val snail = require_crawl_images "snail"
  val snaill = #floorl snail
  val snailr = #floorr snail
  val right_snailu = #rightu snail
  val right_snaild = #rightd snail
  val left_snailu = #leftu snail
  val left_snaild = #leftd snail
  val ceil_snaill = #ceill snail
  val ceil_snailr = #ceilr snail

  val snailr_fade = SDL.Util.alphadim (#floorr snail)
  val snaill_fade = SDL.Util.alphadim (#floorl snail)
  val error = Images.requireimage "error_frame.png"
  val right_snailu_fade = SDL.Util.alphadim right_snailu
  val right_snaild_fade = SDL.Util.alphadim right_snaild
  val left_snailu_fade = SDL.Util.alphadim (#leftu snail)
  val left_snaild_fade = SDL.Util.alphadim (#leftd snail)
  val ceil_snaill_fade = SDL.Util.alphadim (#ceill snail)
  val ceil_snailr_fade = SDL.Util.alphadim (#ceilr snail)

  val redhi = Images.requireimage "redhighlight.png"
  val greenhi = Images.requireimage "greenhighlight.png"
  val robobox = Images.requireimage "robobox.png"

  val star = Images.requireimage "redstar.png"
  val spider_up = Images.requireimage "spider_up.png"
  val spider_down = Images.requireimage "spider_down.png"

  val fireballr = Images.requireimage "fireball.png"
  val fireballl = Images.requireimage "fireball_left.png"

  val rocksprite = Images.requireimage "coral.png"

  (* snail flags *)
  structure BF :>
  sig
      include FLAG
      val ONGROUND : flag
  end =
  struct
      open Flag
      val ONGROUND = 0
      val NFLAGS = 1
      val flags = nflags NFLAGS
  end

  val snailx = ref (Room.STARTX * TILEW)
  val snaily = ref (Room.STARTY * TILEH)

  (* XXX Cosmetic, and currently not general. This should probably
     be relative to the orientation. It should not be set just
     because the left key was just pressed, since left means different
     things depending on the orientation and context. *)
  val snailface = ref FRIGHT
  (* snail dir is the snail's orientation, which is the direction
     that its feet are pointing. (In normal non-stuck gravity,
     DOWN.) *)
  val snaildir = ref DOWN

  val snailnormal = ref (3.0 *  Math.pi / 2.0)
  val attached = ref false

  val slime : (int * int) option ref = ref NONE
  val brightgreen = SDL.color (0w0, 0wxFF, 0w0, 0wxFF)
  val darkgreen = SDL.color (0w0, 0wx77, 0w0, 0wxFF)
  val brightred = SDL.color (0wxFF, 0w0, 0w0, 0wxFF)
  val slimecolor = ref brightgreen

  val lowerx = ref 0
  val lowery = ref 0

  (* position of top-left of scroll window, in pixels *)
  val scrollx = ref 0
  val scrolly = ref 0
  val scrolldx = ref 0
  val scrolldy = ref 0

  (* currentposition of the mouse. Used for showing overlay *)
  val mousex = ref 0
  val mousey = ref 0

  (* I think these became in units of 1 pixel per frame *)
  val snaildx = ref 0
  val snaildy = ref 0
  val snailstate = BF.flags

  (* For many keys, we keep track of their state (in the event loop below).
     This is so that you can hold both left and right momentarily without
     things getting cocked up. Note that keys are used both for the player
     movement (where they are canonicalized inside the movement code) and
     for control in the editor, and maybe menus some day... *)
  val KEYMASK_UP = 0w1 : Word32.word
  val KEYMASK_DOWN = 0w2 : Word32.word
  val KEYMASK_LEFT = 0w4 : Word32.word
  val KEYMASK_RIGHT = 0w8 : Word32.word
  val KEYMASK_SHIFT = 0w16 : Word32.word
  val KEYMASK_CTRL = 0w32 : Word32.word
  val keymask = ref (0w0 : Word32.word)
  fun holding_key mask = Word32.andb (!keymask, mask) <> 0w0

  (* XXX This should go away (also bullets) but the code might be useful for
  * reference if we make e.g. crabs. --chris *)
  datatype monster = STAR | SPIDER

  datatype object =
      Bullet   of { dx : int, dy : int, x : int, y : int }
    | Monster  of { which : monster, dir : dir, dx : int, dy : int, x : int, y : int }

  val objects = ref (nil : object list)
  val rocks = [ref (Rock { x = 240, y = 238, dx = 0, dy = 0 })] (* XXX nil : rock ref list *)
      
  val paused = ref false
  val advance = ref false

  fun onscreen (x, y) =
      SOME ((x + !scrollx) div TILEW,
            (y + !scrolly) div TILEH)

  datatype editmode = ZBG | ZFG | ZMASK | ZPROP | ZBRUSH

  val message = ref ""

  val room = ref Room.start

  
  val anemonestotal = ref 8
  val anemonesfree = ref AnemoneSet.empty 

  (* This is the selectable tile area. *)

  (* Horizontal editor *)
  val TITLE_HEIGHT = Font.Normal.height + 2
  val OFFSET_OF_TILES_FROM_EDITY = TITLE_HEIGHT + TILEH + 2
  val EDITW = Tile.TILESET_WIDTH
  val EDITH = 16
  val EDITWIDTH = EDITW * TILEW + 4
  val EDITHEIGHT = EDITH * TILEH + 4
  val editmode = (* ref ZBRUSH *) ref ZBG

  val showmask = ref true
  val showprop = ref true
  val editx = ref 0
  val edity = ref (height - (EDITHEIGHT + TITLE_HEIGHT))


  (* vertical editor, rhs *)
(*
  val TITLE_HEIGHT = Font.height + 2
  val OFFSET_OF_TILES_FROM_EDITY = TITLE_HEIGHT + TILEH + 2
  val EDITW = 32
  val EDITH = 16
  val EDITWIDTH = EDITW * TILEW + 4
  val EDITHEIGHT = height
  val editmode = (* ref ZBRUSH *) ref ZBG
  val showmask = ref true
  val showprop = ref true
  val editx = ref (width - EDITWIDTH)
  val edity = ref 0 (* (height - EDITHEIGHT) *)
*)

  (* XXX should be option, should be reset (or saved) when changing modes *)
  
  val edit_lastidx = ref 0

  val NUM_EDIT_TILESETS = Tile.TILESET_HEIGHT div EDITH + (case EDITH mod Tile.TILESET_HEIGHT of
                                                               0 => 0
                                                             | _ => 1 (* extra page for leftovers *))

  val edittileset = ref 0
  val editbgtile = ref (Tile.fromword 0w1)
  val editfgtile = ref (Tile.fromword 0w0)
  val editmask = ref Tile.MSOLID
  val editprop = ref Tile.PPLAIN
  val editbrush = ref (dirt_material, BRECT (3, 3))
  val editmenucolor = SDL.color (0wx44, 0wx14, 0wx14, 0wxFF)
  val editmodecolor = SDL.color (0wxFF, 0wx14, 0wx14, 0wxFF)
  (* for selecting *)
  val editselectx = ref 0
  val editselecty = ref 0
  val editselectx0 = ref 0
  val editselecty0 = ref 0
  val editselectmode = ref false
  val editclipboard = Vector.fromList (List.tabulate (4,fn _ => 
      (ref []:(Tile.mask * int * int) list ref,
       ref []:(Tile.prop * Tile.tile * Tile.tile * Tile.mask * int * int) list list ref)))
  val editshowoverlay = ref true
  val editcopymaskarea = ref (#1 (Vector.sub (editclipboard,0)))
  val editcopyarea = ref (#2 (Vector.sub (editclipboard,0)))

  val showedit = ref false
  local
      val thisone = Images.requireimage "thisone.png"
      val brushes = Images.requireimage "brush.png"

      val MASKS =
          Vector.fromList 
          [ MEMPTY, MSOLID,
            MRAMP LM, MRAMP MH, MRAMP HM, MRAMP ML,
            MDIAG NW, MDIAG NE, MDIAG SW, MDIAG SE,
            MCEIL LM, MCEIL MH, MCEIL HM, MCEIL ML,
            MLEFT LM, MLEFT MH, MLEFT HM, MLEFT ML,
            MRIGHT LM, MRIGHT MH, MRIGHT HM, MRIGHT ML

            (* XXX more *)
            ]

      val PROPS =
          Vector.fromList
          ([ PPLAIN, 
            PGOTO (UP, 0), PGOTO (DOWN, 0), PGOTO (LEFT, 0), PGOTO (RIGHT, 0),
            PGOTO (UP, 1), PGOTO (DOWN, 1), PGOTO (LEFT, 1), PGOTO (RIGHT, 1),
            PGOTO (UP, 2), PGOTO (DOWN, 2), PGOTO (LEFT, 2), PGOTO (RIGHT, 2),
            PGOTO (UP, 3), PGOTO (DOWN, 3), PGOTO (LEFT, 3), PGOTO (RIGHT, 3),
            PCURRENT UP, PCURRENT DOWN, PCURRENT LEFT, PCURRENT RIGHT,
            PLIGHT LWHITE, PLIGHT LYELLOW, PLIGHT LBLUE, PLIGHT LRED, PLIGHT LGREEN
            ]
           @ List.tabulate (14, PACTION))

      (* Has to have the same contents as the icon. *)
      val BRUSHW = 16 (* pixels *)
      val BRUSHH = 16
      val BRUSHES_SETW = 16 (* tiles *)
      val BRUSHES_SET =
          Vector.fromList
          [ SOME (eraser_material, BERASERECT (1,1)),
            SOME (eraser_material, BERASERECT (3,3)),
            SOME (eraser_material, BERASERECT (9,9)),
            SOME (dirt_material,BPASTEMASKS editcopymaskarea),
            SOME (block_material,BPASTEMASKS editcopymaskarea),
            SOME (scribble_material,BPASTEMASKS editcopymaskarea),
            SOME (eraser_material, BPASTE),
            (* NONE x 9 *)
            NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,
            SOME (dirt_material, BRECT (1, 1)),
            SOME (dirt_material, BRECT (3, 3)),
            SOME (dirt_material, BRECT (9, 9)),
            SOME (dirt_material, BSLOPEDRECT (4, 4)),
            SOME (dirt_material, BSLOPEDRECT (9, 9)),
            (* NONE x 11 *)
            
            NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, 
            SOME (block_material, BRECT (1, 1)),
            SOME (block_material, BRECT (3, 3)),
            SOME (block_material, BRECT (9, 9)),
            SOME (block_material, BSLOPEDRECT (4, 4)),
            SOME (block_material, BSLOPEDRECT (9, 9)), 
            (* NONE x 11 *)
            NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, 
            SOME (scribble_material, BRECT (1, 1)),
            SOME (scribble_material, BRECT (3, 3)),
            SOME (scribble_material, BRECT (9, 9)),
            SOME (scribble_material, BSLOPEDRECT (4, 4)),
            SOME (scribble_material, BSLOPEDRECT (9, 9)) ]

      fun edit_click (x, y) =
          let 

              (* XXX Should allow to click bg/fg/mask to switch to
                 editing that mode *)

              (* skip current bg/fg/mask display *)
              val y = y - OFFSET_OF_TILES_FROM_EDITY + TILEH (* XXX don't know why I need TILEH here, but works *)

              (* which tile are we clicking, then? *)
              val idx = ((y div TILEH) * EDITW + (x div TILEW))
(*              val () = print ("orig idx: " ^ Int.toString idx ^ "\n") *)
              val idx = idx + (!edittileset * EDITH) * Tile.TILESET_WIDTH
(*              val () = print ("scrolled idx: " ^ Int.toString idx ^ "\n") *)
          in
              if idx >= 0
              then
                  let in
                      edit_lastidx := idx;
(*                      print ("tile/mask " ^ Int.toString idx ^ "\n"); *)
                      (case !editmode of
                           ZMASK =>
                               if idx >= 0 andalso idx < Vector.length MASKS
                               then editmask := Vector.sub(MASKS, idx)
                               else ()
                         | ZBG => editbgtile := Tile.fromword (Word32.fromInt idx)
                         | ZFG => editfgtile := Tile.fromword (Word32.fromInt idx)
                         | ZBRUSH =>
                               if idx >=0 andalso idx < Vector.length BRUSHES_SET
                               then case Vector.sub(BRUSHES_SET, idx) of
                                      NONE => ()
                                    | SOME b => editbrush := b
                               else ()
                         | ZPROP =>
                               if idx >=0 andalso idx < Vector.length PROPS
                               then editprop := Vector.sub(PROPS, idx)
                               else ())

                  end
              else ()
          end

      fun draw_edit_child (surf, x, y, _ (* width, fixed *), _ (* height, fixed *)) =
          let
              fun drawloop draw =
                let in
                  (Util.for 0 (EDITH - 1)
                   (fn yy =>
                    Util.for 0 (EDITW - 1)
                    (fn xx =>
                     draw (EDITW * yy + xx,
                           x + (TILEW * xx),
                           TILEH + 2 +
                           y + (TILEH * yy)))));

                  (* XXX do it analytically! PERF *)
                  (Util.for 0 (EDITH - 1)
                   (fn yy =>
                    Util.for 0 (EDITW - 1)
                    (fn xx =>
                     if (EDITW * yy + xx = !edit_lastidx)
                     then SDL.blitall (thisone, surf, 
                                       (* ugh Lololllll *)
                                       ~1 + x + (TILEW * xx),
                                       ~1 + TILEH + 2 +
                                       y + (TILEH * yy))
                     else ())))
                end

          in
              (* indicate the mode by coloring its background *)
              SDL.fillrect (surf, 
                            x + ~1 + 
                            ((TILEW + 2) *
                             (case !editmode of
                                     ZBG => 0
                                   | ZFG => 1
                                   | ZMASK => 2
                                   | ZPROP => 3
                                   | ZBRUSH => 4)), 
                            y - 1,
                            TILEW + 4,
                            TILEH + 4,
                            editmodecolor);

              (* the background, foreground, mask, prop *)
              Tile.draw     (!editbgtile, surf, x,                   y, 0, 0);
              Tile.draw     (!editfgtile, surf, x + 1 * (TILEW + 2), y, 0, 0);
              Tile.drawmask (!editmask,   surf, x + 2 * (TILEW + 2), y);
              Tile.drawprop (!editprop,   surf, x + 3 * (TILEW + 2), y);
              (* XXX draw brush? *)

              case !editmode of
                  ZMASK =>
                      drawloop (fn (idx, destx, desty) =>
                                (* might subscript *)
                                let val m = Vector.sub (MASKS, idx)
                                in
                                    Tile.drawmask (m, surf, destx, desty)
                                end handle Subscript => ())

                | ZPROP =>
                      drawloop (fn (idx, destx, desty) =>
                                (* might subscript *)
                                let val p = Vector.sub (PROPS, idx)
                                in
                                    Tile.drawprop (p, surf, destx, desty)
                                end handle Subscript => ())
 
                | ZBRUSH =>
                      drawloop (fn (idx, destx, desty) =>
                                (* This is different; we just use the
                                   organization in the graphic itself
                                   since these brushes don't need to
                                   be first class citizens. *)
                                let val x = idx mod BRUSHES_SETW
                                    val y = idx div BRUSHES_SETW
                                in
                                    SDL.blit (brushes, x * BRUSHW, y * BRUSHH,
                                              BRUSHW, BRUSHH, surf, destx, desty)
                                end)


                | _ =>
                      drawloop (fn (idx, destx, desty) =>
                                let 
                                    (* to account for scroll: a page is TILESET_WIDTH times
                                       the number of rows we show on a page. *)
                                    val idx = idx + (!edittileset * EDITH) * Tile.TILESET_WIDTH
                                    val t = Tile.fromword (Word32.fromInt idx)
                                in
                                    Tile.draw (t, surf, destx, desty, 0, 0)
                                end)
             end

      fun drawtitle (surf, x, y) =
          let val title = "editor. ^2z^< to change mode: ^3" ^
              (case !editmode of
                   ZBG => "background"
                 | ZFG => "foreground"
                 | ZMASK => "mask"
                 | ZPROP => "prop"
                 | ZBRUSH => "brush") ^
                   "^< ^2m^<ask ^3" ^
                   (if !showmask then "ON" else "OFF") ^
                   "^< ^2p^<rop ^3" ^
                   (if !showprop then "ON" else "OFF") ^
                   "^<"
          in
              Font.Normal.draw(surf, x, y + 1, title)
          end

  in
      val edit_window =
          Window.make { click = edit_click,
                        keydown = fn _ => false,
                        keyup = fn _ => false,
                        
                        border = 0,
                        border_color = editmenucolor,
                        resizable = false,
                        x = !editx, y = !edity,
                        w = EDITWIDTH, h = EDITHEIGHT,
                        title_height = TITLE_HEIGHT,

                        draw = draw_edit_child,
                        drawtitle = drawtitle }

      fun drawedit _ =
          if !showedit (* andalso holding_key KEY_SHIFT *)
          then Window.draw edit_window screen
          else ()

  end


  datatype drawwhat =
      (* Debugging: Draw the clip mask *)
      DRAWMASK
      (* Draw the properties *)
    | DRAWPROP
      (* Draw the indicated layer *)
    | DRAW of Room.layer

  (* draw a view of the world where the top left of
     the screen is at scrollx/scrolly *)
  fun drawzone (layer) =
    let
      (* tlx/tly is the (screen coordinate pixel) point 
         at which we draw the top-left tile. it is nonpositive *)
      val tlx = 0 - (!scrollx mod TILEW)
      val tly = 0 - (!scrolly mod TILEH)
        
      (* this is the tile number (world absolute) that we start
         by drawing *)
      val xstart = !scrollx div TILEW
      val ystart = !scrolly div TILEH
    in
      case layer of
        (* not really a layer, but useful when editing / debugging *)
        DRAWMASK =>
        (* nb, always doing one extra tile, because we might
           display only partial tiles if not aligned to scroll
           view *)
          Util.for 0 TILESW
          (fn x =>
           Util.for 0 TILESH
           (fn y =>
            case Room.maskat (!room, xstart + x, ystart + y) of
                MEMPTY => ()
              | m => Tile.drawmask(m, screen, 
                                   tlx + (x * TILEW), tly + (y * TILEH))
                    ))
      | DRAWPROP => 
        (* nb, always doing one extra tile, because we might
           display only partial tiles if not aligned to scroll
           view *)
          Util.for 0 TILESW
          (fn x =>
           Util.for 0 TILESH
           (fn y =>
            case Room.propat (!room, xstart + x, ystart + y) of
                PPLAIN => ()
              | p => Tile.drawprop(p, screen, 
                                   tlx + (x * TILEW), tly + (y * TILEH))
                    ))

      | DRAW layer => 
          Util.for 0 TILESW
          (fn x =>
           Util.for 0 TILESH
           (fn y =>
            Tile.draw (Room.tileat(!room, layer, xstart + x, ystart + y), screen, 
                       tlx + (x * TILEW),
                       tly + (y * TILEH),
                       xstart + x, ystart + y)
            ))
    end

  fun drawselection () = 
    if !showedit 
    then 
      let
        val (x0,x1,y0,y1) = 
          (!editselectx0,!editselectx,!editselecty0,!editselecty)
        val (x0,x1) = if (x0 > x1) then (x1 div TILEW,x0 div TILEW) else (x0 div TILEW,x1 div TILEW)

        val (y0,y1) = if (y0 > y1) then (y1 div TILEH,y0 div TILEH) else (y0 div TILEH,y1 div TILEH)
      in
        Util.for x0 x1
        (fn x => 
          Util.for y0 y1
          (fn y =>
            Tile.drawselection (screen, (x * TILEW) - !scrollx,(y * TILEH) -
            !scrolly)))
      end
    else ()

  fun areacopy () =
    if !showedit
    then
      let
        val (x0,x1,y0,y1) = 
          (!editselectx0,!editselectx,!editselecty0,!editselecty)
        val (x0,x1) = if (x0 > x1) then (x1 div TILEW,x0 div TILEW) else (x0 div TILEW,x1 div TILEW)

        val (y0,y1) = if (y0 > y1) then (y1 div TILEH,y0 div TILEH) else (y0 div TILEH,y1 div TILEH)
        val _ = (!editcopyarea) := (Room.areacopy (!room,x0,y0,x1,y1))
        val _ = (!editcopymaskarea) := (List.map (fn (_,_,_,m,x,y) => (m,x,y))
          (List.concat (!(!editcopyarea))))
      in
        ()
      end
    else ()

  fun snailify () =
    if !showedit 
    then
      let
        val (x0,x1,y0,y1) = 
          (!editselectx0,!editselectx,!editselecty0,!editselecty)
        val (x0,x1) = if (x0 > x1) then (x1 div TILEW,x0 div TILEW) else (x0 div TILEW,x1 div TILEW)
        val (y0,y1) = if (y0 > y1) then (y1 div TILEH,y0 div TILEH) else (y0 div TILEH,y1 div TILEH)
        val masks = Tile.brushtomasks (Tile.BHILLLH ((x1-x0) + 1,(y1-y0) + 1))
        val _ = (!editcopymaskarea) := masks
      in
        ()
      end
    else ()

  fun drawbrushoverlay () = 
    if !editshowoverlay andalso !editmode = ZBRUSH
    then
      let
        val (x0,y0) = ((!mousex + !scrollx) div TILEW, (!mousey + !scrolly) div TILEH)
        val masks = Tile.brushtomasks (#2 (!editbrush))
      in
        List.app (fn (m,x,y) => Tile.drawmask (m,screen,(x + x0) * TILEW - !scrollx,(y + y0) * TILEH - !scrolly)) masks
      end
    else ()


  fun drawobjects () =
      let
          fun doobj (Bullet { x, y, dx, dy }) =
              blitall (if dx > 0 then fireballr
                       else fireballl, screen,
                       x - !scrollx, y - !scrolly)
            | doobj (Monster { which, dx, dy, x, y, dir }) =
              let val glyph = 
                  (case (dir, which, dy) of 
                       (_, STAR, _) => star
                     | (_, SPIDER, 0) => spider_down
                     | (_, SPIDER, _) => spider_up
                           (* more objects... *)
                           ) handle Match => error
              in
                  blitall (glyph,
                           screen,
                           x - !scrollx, y - !scrolly)
              end
      in
          app doobj (!objects)
      end

  fun drawrocks () =
      let
          fun dorock (ref (Rock { x, y, dx, dy })) =
              blitall (rocksprite, screen, x - !scrollx, y - !scrolly)
      in
          List.app dorock rocks
      end

    val mecolor = SDL.color (0w0, 0wxFF, 0wxFF, 0wxFF)
    fun drawpixel_clip x = drawpixel x handle SDL _ => ()
    fun drawsnail fade =
        let val img = 
            (case (!snaildir, !snailface, fade) of
                 (* FLEFT = cw of snaildir, FRIGHT = ccw of snaildir *)
                 (DOWN, FLEFT, false) => snaill
               | (DOWN, FLEFT, true) => snaill_fade
               | (DOWN, FRIGHT, false) => snailr
               | (DOWN, FRIGHT, true) => snailr_fade
               | (UP, FLEFT, false) => ceil_snailr
               | (UP, FLEFT, true)  => ceil_snailr_fade
               | (UP, FRIGHT, false) => ceil_snaill
               | (UP, FRIGHT, true)  => ceil_snaill_fade
               | (RIGHT, FLEFT, false) => right_snaild
               | (RIGHT, FLEFT, true)  => right_snaild_fade
               | (RIGHT, FRIGHT, false) => right_snailu
               | (RIGHT, FRIGHT, true)  => right_snailu_fade
               | (LEFT, FLEFT, false) => left_snailu
               | (LEFT, FLEFT, true)  => left_snailu_fade
               | (LEFT, FRIGHT, false) => left_snaild
               | (LEFT, FRIGHT, true)  => left_snaild_fade)
        in
            blitall (img, screen, !snailx - !scrollx, !snaily - !scrolly);
            (* locate me *)
(*            drawcircle (screen, !snailx - !scrollx, !snaily - !scrolly, 3,
                        mecolor);
            (* and lower left tile *)
            drawcircle (screen, !lowerx - !scrollx, !lowery - !scrolly, 3,
                        mecolor);

            (* draw normal vector *)
            drawcircle (screen, 
                        !snailx - !scrollx + SNAILW div 2 
                            + Real.round(20.0 * Math.cos(!snailnormal)), 
                        !snaily - !scrolly + SNAILH div 2
                             + Real.round(20.0 * Math.sin(!snailnormal)), 
                                          5, !slimecolor);
*)
            (* draw the slime *)
            case !slime of
              NONE => ()
            | SOME (x, y) => (drawcircle (screen, x - !scrollx, y - !scrolly,
                                          7, !slimecolor);
                              drawpixel_clip (screen, x - !scrollx, y - !scrolly, darkgreen))
        end

  fun dtos LEFT = "left"
    | dtos RIGHT = "right"
    | dtos UP = "up"
    | dtos DOWN = "down"

  fun inttos (I_GO d) = "go " ^ dtos d
    | inttos I_JUMP = "jump"
    | inttos I_SUPERJUMP = "superjump"
    | inttos I_BOOST = "boost"

  (* returns a list of intentions from the current key state.
     These should be added to other intentions, like jumping,
     which are not contextual. *)
  fun contextual_intention keys =
      let
          (* The list of directions implied by the currently
             held keys, in screen orientation. *)
          val sd = nil
          val sd = if holding_key KEYMASK_UP then UP :: sd else sd
          val sd = if holding_key KEYMASK_DOWN then DOWN :: sd else sd
          val sd = if holding_key KEYMASK_LEFT then LEFT :: sd else sd
          val sd = if holding_key KEYMASK_RIGHT then RIGHT :: sd else sd
      in
          case !slime of
              (* XXX I think we should also canonize for freefall mode. If
                 I'm being pulled to the right because of current, then
                 I should be able to press 'up' to "move right" in my
                 reference frame *)
              NONE => map I_GO sd
            | SOME(x, y) =>
              let val slime_mask = Room.maskat (!room, x div TILEW, y div TILEH)
              in map I_GO (List.mapPartial (Slime.canonize (!snaildir) slime_mask) sd)
              end
      end

  fun intends il i = List.exists (fn x => x = i) il

  (* number of pixels per frame that we can walk at maximum *)
  val MAXWALK = 4
  val TERMINAL_VELOCITY = 4
  val JUMP_VELOCITY = 3

  (* make sure snail's in the scroll window! *)
  fun setscroll i =
    (* Don't need to do that if editing.
       XXX could at least prevent us from scrolling
       totally out of the level. *)
    if !showedit
    then
      let in
          if holding_key KEYMASK_UP
          then scrolly := !scrolly - 5
          else ();

          if holding_key KEYMASK_DOWN
          then scrolly := !scrolly + 5
          else ();

          if holding_key KEYMASK_LEFT 
          then scrollx := !scrollx - 5
          else ();

          if holding_key KEYMASK_RIGHT
          then scrollx := !scrollx + 5
          else ();

          ()
      end
    else
    let in
      (* stay in the center third of the screen. *)
      if !snailx - !scrollx < (width div 3)
      then ((* print "left edge\n"; *)
            scrollx := !snailx - (width div 3))
      else if (!scrollx + width) - !snailx < (width div 3)
           then ((* print "right edge\n"; *)
                 scrollx := !snailx - 2 * (width div 3))
           else ();

      if !snaily - !scrolly < (height div 3)
      then ((* print "top edge\n"; *)
            scrolly := !snaily - (height div 3))
      else if (!scrolly + height) - !snaily < (height div 3)
           then ((* print "snailtom edge\n"; *)
                 scrolly := !snaily - 2 * (height div 3))
           else ();

      (* But: if we can fit the whole game on-screen, then
         do so. *)
      if Room.ROOMH * TILEH < height
      then scrolly := (height - Room.ROOMH * TILEH) div ~2
      else ();

      if Room.ROOMW * TILEW < width
      then scrollx := (width - Room.ROOMW * TILEW) div ~2
      else ()
    end

  datatype clip_orientation = HORIZ | VERT

  (* Move the snail. The code assumes that the entity is oriented downward,
     so that "gravity" and sticking all happen in that direction. This function
     is called where "down" is actually a different direction (left, right, up)
     and the inputs and outputs are rotated to do the impedance matching. So,
     this function should not read any refs, including the world. Instead, that
     stuff should be mediated by argument functions, so that those functions
     can do the requisite rotations. Additionally, this function should not modify
     any state, so that the outputs can be suitably interpreted. *)
  fun movesnail_oriented
      {
       (* The list of intentions. The only intentions we can
          carry out are JUMP, GO LEFT, and GO RIGHT. Other intentions
          don't make sense because when you're oriented down, you can't
          move up or down. *)
        i : intention list,
        (* The pixel coordinates of the top left of the snail.
           Remember this only works when we're in the downward orientation,
           but the caller can interpret the resulting changes to x as a
           change in ~y, if he wants. *)
        x : int, y : int,
        (* The current velocity in those directions. *)
        dx : int, dy : int,
        (* Slime, if we have it, rotated into the local reference frame. *)
        slime : (int * int) option,
        (* Is the config (which can only be HORIZ or VERT) blocked, given
           its upper left corner? *)
        is_stopped : clip_orientation -> int * int * int * int -> bool,

        (* Get the mask at the tile positions, assuming local coordinate
           system. The mask should be rotated! *)
        mask_at : int * int -> Tile.mask,

        (* From a pixel coordinate, give the tile coordinate. *)
        tile_from_coord : int * int -> int * int,

        (* From a pixel coordinate, give the offset within the tile
           at that coord *)
        tile_offset_from_coord : int * int -> int * int

       } =
    let
        val slimer = ref slime

        (* Won't need this! snaildir is down *)
        val snaildir = 12345
        val slime = "NO"
        exception clip

        (* XXX these maximums should also be parameters *)
        (* first, react to the entity's intentions. We only allow left
           and right movement in this orientation. *)
        val dx =
            if intends i (I_GO LEFT)
               andalso dx > ~ MAXWALK
            then dx - 1
            else dx
        val dx = 
            if intends i (I_GO RIGHT)
               andalso dx < MAXWALK
            then dx + 1
            else dx

        (* and if we don't intend to move at all, slow us down *)
        val dx = 
            if not (intends i (I_GO RIGHT)) andalso
               not (intends i (I_GO LEFT))
            then (if dx > 0 then dx - 1
                  else if dx < 0 then dx + 1 else 0)
            else dx

        (* jumping -- only when not in air already!
           (either way the intention goes away) *)
        val (dy, i) = 
          if intends i (I_JUMP)
             (* something to push off of.. *)
             andalso is_stopped HORIZ (x, y, 0, 1)
             andalso (slimer := NONE; true)
          then (dy - JUMP_VELOCITY,
                List.filter (fn I_JUMP => false | _ => true) i)
          else (dy, List.filter (fn I_JUMP => false | _ => true) i)


        (* Apply constant downward gravity.
           Gravity is always (locally) downward. In freefall mode,
           we only change orientation because we're in a current, which
           then causes us to be pushed that way. *)
        val dy = dy + 1

        (* water resistance *)
        val dx = if dx > TERMINAL_VELOCITY
                 then dx - 1
                 else if dx < ~ TERMINAL_VELOCITY
                      then dx + 1 
                      else dx

        val dy = if dy > TERMINAL_VELOCITY
                 then dy - 1
                 else if dy < ~ TERMINAL_VELOCITY
                      then dy + 1
                      else dy


        val xi = if dx < 0 then ~1 else 1
        val yi = if dy < 0 then ~1 else 1

        val xr = ref x
        val yr = ref y
        val stopx = ref false
        val stopy = ref false

        (* Allow vertical movement if walking up a gradient *)
        (*
        fun climb () =
          (yr := !yr - 1;
           true)
        *)

        fun nudgex _ = 
            let val nextx = !xr + xi
            in
                if !stopx
                then NONE
                else if (* can I move to the right without climbing?
                           x .
                        *)
                        is_stopped HORIZ (!xr, !yr, xi, 0)
                     then 
                         (* can't move immediately right, but might
                            be able to go up a ramp.

                            . .
                            x #

                            *)
                         if not (is_stopped HORIZ (!xr, !yr, 0, ~1)
                                 orelse
                                 is_stopped HORIZ (!xr, !yr, xi, ~1))
                            (* climb. *)
                         then SOME (nextx, !yr - 1)
                             (* stopped and stuck. *)
                         else (stopx := true; NONE)
                     else SOME (nextx, !yr)
            end

        (* Only allow attachment on tiles that have completely flat tops. *)
        fun flat_topped MSOLID = true
          | flat_topped (MCEIL _) = true
          | flat_topped (MRIGHT HM) = true
          | flat_topped (MLEFT HM) = true
          | flat_topped _ = false

        fun try_attach () =
            case !slimer of
              SOME _ => ()
            | NONE =>
                let val sx = !xr + (3 * TILEW) div 2
                    val sy = !yr + (2 * TILEH)
                    val (tx, ty) = tile_from_coord (sx, sy)

                    val attachtile = mask_at (tx, ty)
                in
                    if flat_topped attachtile
                    then slimer := SOME (sx, sy)
                    else ()
                end

        fun nudgey _ = 
            let val nexty = !yr + yi
            in
                if !stopy 
                then ()
                else if is_stopped HORIZ (!xr, !yr, 0, yi)
                then (stopy := true;
                      try_attach ())
                else (yr := nexty)
                      (***********
                      (* detach *)
                      (* XXX Detach if there's upward force. But
                         we should detach only if there's jump,
                         or the slime requires us to. *)
                      if yi = ~1
                      then ((* print "Detaching in nudgey\n"; *)
                            slimer := NONE)
                      else ())
                      ************)
            end

        (* XXX This should be more like rotateto.
           Note that we'll need to know the detached orientation
           (which comes from the current/gravity field), so that
           we can check that it's safe to be there. *)
        fun safeDetach (x, y) = true

        fun slimex _ =
            (* XXX needs to sometimes change direction *)
            case !slimer of
              NONE => NONE
            | SOME (sx, sy) =>
            let
                (* val next = !xr + xi *)

                fun clipAt (x, y) =
                    let 
                        val (tx, ty) = tile_from_coord (x, y)
                        (* val (mx, my) = tile_offset_from_coord (x, y) *)
                            (* This is ok certified tom7 23 Aug 2010 *)
                        val (mx, my) = (x mod TILEW, y mod TILEH)
                        val ma = mask_at (tx, ty)
                    in
                        Tile.clipmask ma
                        (mx, my)
                    end

                fun try3 (p1, p2, p3) =
                    if      clipAt p1 then SOME p1
                    else if clipAt p2 then SOME p2
                    else if clipAt p3 then SOME p3
                    else NONE (* slime fell off the wall :( *)

                (* Get pixels to the right of us *)
                val tryx = sx + xi
                fun proposeCandidate () = try3 ((tryx, sy - 1),
                                                (tryx, sy),
                                                (tryx, sy + 1))

                fun sameTile (x, y) = 
                    tile_from_coord (x, y) =
                    tile_from_coord (sx, sy)

                (*
                fun getAngle (x, y) = raise Match

                fun resolveTransition p =
                    let val a1 = getAngle (sx, sy)
                        val a2 = getAngle (sx + xi, sy)
                    in
                        raise Fail "you have to put a function here"
                    end
                *)

                fun goodAngle (x, y) = true

                fun reorient (x, y) =
                  let 
                      val (otx, oty) = tile_from_coord (sx, sy)
                      val (tx, ty) = tile_from_coord (x, y)
                      val oldtile = maskat (!room, otx, oty)
                      val newtile = maskat (!room, tx, ty)
                  in
                      case (oldtile, newtile) of
                          (MRAMP MH, MRIGHT MH) => RIGHT (* (print "reorient right\n"; RIGHT) *)
                        | (MRAMP HM, MLEFT MH) => LEFT(* (print "reorient right\n"; LEFT) *)
                        | _ => DOWN
                  end

            in
                (* returns an orientation for the snail because it
                   needs to take care of the flip, due to canonization. *)
                case proposeCandidate () of
                    SOME p => (if sameTile p orelse ((*print "new tile\n";*) goodAngle p)
                               then SOME (p, reorient p)
                               else NONE)
                  | NONE => NONE
            end

        (* NOTE: These assume the shape of the snail. *)
        fun bottomRightPixel () = (!xr + 3 * TILEW - 1, !yr + 2 * TILEH - 1)
        fun bottomLeftPixel () = (!xr, !yr + 2 * TILEH - 1)

        fun wedged () =
            (* back/front is left/right *)
            let val (brx, bry) = bottomRightPixel ()
                val (right_x, right_y) = (brx + 1, bry)
                val (lower_x, lower_y) = (brx, bry + 1)
                val (rtx, rty) = tile_from_coord (right_x, right_y)
                val (ltx, lty) = tile_from_coord (lower_x, lower_y)
            in
                mask_at (rtx, rty) = MRIGHT MH andalso
                mask_at (ltx, lty) = MRAMP MH
            end
          orelse
            (* test XXX orelse: Wedged at the left. *)
            let val (blx, bly) = bottomLeftPixel ()
                val (left_x, left_y) = (blx - 1, bly)
                val (lower_x, lower_y) = (blx + 2, bly + 3)
            in
                lowerx := lower_x;
                lowery := lower_y;
(*                print ("left tile: "
                     ^ Tile.masktostring (mask_at (left_x div TILEW, left_y div TILEH))
                     ^ "\n");
                print ("lower tile: "
                     ^ Tile.masktostring (mask_at (lower_x div TILEW, lower_y div TILEH))
                     ^ "\n"); *)
                mask_at (left_x div TILEW, left_y div TILEH) = MLEFT MH andalso
                mask_at (lower_x div TILEW, lower_y div TILEH) = MRAMP HM
            end

        fun rotate_wedged new_dir =
          let 
              val (newx, newy) = 
                  case new_dir of
                      RIGHT => (!xr + TILEW, !yr - TILEH)
                    | LEFT => (!xr, !yr - TILEH)
                    | _ => raise Game "impossible input to rotate_wedged"
                      (* XXX: need to do LEFT too. *)
          in
              (* We know the new orientation is VERT, because we started
                 as HORIZ. *)
              (*
              print ("oldx = " ^ Int.toString (!xr)
                 ^ ", oldy = " ^ Int.toString (!yr) ^ "\n");
              print ("newx = " ^ Int.toString newx
                 ^ ", newy = " ^ Int.toString newy ^ "\n");
              *)
              (* if is_clipped VERT (newx, newy) *)
              if is_stopped VERT (newx, newy, 0, 0)
              then NONE
              else SOME (newx, newy)
          end

        (* Is the slime canonical (given its coordinates (sx, sy)),
           meaning under my belly, when I'm at (nx, ny)? *)
        fun slimeCanonical (sx, sy) (nx, ny) = sx = nx + (3 * TILEW) div 2

        (* We leave early if a rotation happens. *)
        exception Rotated of { new_dir : Types.dir,
                               slime : (int * int) option,
                               new_loc : int * int }

        fun hyperx i =
            if wedged () then
                (* CONCAVE MAGIC:
                  - let the slime move into the corner while snail stays
                  - if it crosses the inflection point, rotate the snail
                  - if it moves back out, we exit magic mode
                *)
               (slimecolor := brightred ;
                case (slimex i, nudgex i) of
                  (SOME ((sx, sy), d), NONE) =>
                      (* let slime move without snail *)
                      if d <> DOWN
                      then case rotate_wedged d of
                              SOME (snx, sny) => (* perform flip *)
                                  (print("performing flip "
                                         ^ dirtos d ^ "\n");
                                   raise Rotated { new_dir = d,
                                                   new_loc = (snx, sny),
                                                   slime = SOME (sx, sy) })
                            | NONE => (print "failed to flip\n"; ()) (* cannot go on *)
                      else (* dir unchanged *)
                           slimer := SOME (sx, sy)
                | (NONE, NONE) =>
                      (* detach, don't move -- impossible? *)
                      if safeDetach (!xr, !yr)
                      then (print "Weird!\n"; slimer := NONE)
                      else ()
                | (NONE, SOME (x, y)) =>
                      (* just like below *)
                      if safeDetach (x,y) 
                      then (xr := x; yr := y;
                            print "Detaching in hyperx 1\n";
                            slimer := NONE) 
                      else ()
                | (SOME ((sx, sy), d), SOME (nx, ny)) =>
                       (* do what the slime says *)
                      (if true (* safeFlip d *) (* XXX XXX WHAT *)
                       then (slimer := SOME (sx, sy)
                             (* snaildir := d *)
                             (* XXX do flip?  (NB: only in convex case) *)
                             )
                       else ()
                       ;
                       (* This should not happen: How can we be wedged,
                          but able to move, and rotating if we're in the
                          concave case!? *)
                       if d <> DOWN then print "ACK!\n" else ()
                       ;
                       (* only move snail if slime is where it belongs *)
                       if slimeCanonical (sx, sy) (nx, ny)
                       then (xr := nx; yr := ny)
                       else ()
                      )
               )
            else
               (slimecolor := brightgreen ;
                case (slimex i, nudgex i) of 
                    (_, NONE) => () (* FUCKIT principle *)
                  | (NONE, SOME (x, y)) =>
                      if safeDetach (x,y) 
                      then (xr := x; yr := y;
                            print "Detaching in hyperx 2\n";
                            slimer := NONE) 
                      else ()
                  | (SOME ((sx, sy), d), SOME (nx, ny)) =>
                      (xr := nx; yr := ny;
                       slimer := SOME (sx, sy);
                       if d <> DOWN
                          then (* Convex rotation. *)
                              (* TODO: Need to compute the new position,
                                 and then we need to check to see if it's
                                 clipped. If it's not, then we raise Rotated
                                 and we're done. If it's clipped, then
                                 we stop here. *)
                              (* snaildir := d *)
                              let in
                                  print "Sorry, can't do this rotation yet.\n";
                                  ()
                              end
                          else ())
               )

        fun hypery i = nudgey i


    in
        (* try moving one pixel at a time in the desired direction, until it
           is not possible to move any more. Rotations stop us
           immediately by raising the exception. *)

        (Util.for 0 (abs dx - 1) hyperx;
         Util.for 0 (abs dy - 1) hypery;

         (i, !xr, !yr, 
          if !stopx then 0 else dx,
          if !stopy then 0 else dy,
          !slimer,
          DOWN))
        handle Rotated { new_dir, slime, new_loc = (xx, yy) } =>
         (i, xx, yy,
          (* you lose all velocity when you rotate. *)
          0, 0,
          slime,
          new_dir)
    end 

  (* We just implement a single clockwise rotation, since all of the rotations
     are an iteration of that. *)
  fun cw_dir UP = RIGHT
    | cw_dir RIGHT = DOWN
    | cw_dir DOWN = LEFT
    | cw_dir LEFT = UP
  fun cw_intention (I_GO d) = I_GO (cw_dir d)
    | cw_intention i = i
  fun cw_xy (x, y) = (~y, x)
  fun cw_mask MSOLID = MSOLID
    | cw_mask MEMPTY = MEMPTY
    | cw_mask (MCEIL r) =
      MRIGHT (case r of
                  LM => HM
                | MH => ML
                | HM => LM
                | ML => MH)
    | cw_mask (MRAMP r) = MLEFT r
    | cw_mask (MRIGHT r) =
      MRAMP (case r of
                 HM => MH
               | ML => LM
               | LM => ML
               | MH => HM)
    | cw_mask (MLEFT r) =
      MCEIL (case r of
                 MH => LM
               | LM => MH
               | ML => HM
               | HM => ML)
    | cw_mask (MDIAG _) = raise Game "mdiag is dead to us"

  fun cw_orientation HORIZ = VERT
    | cw_orientation VERT = HORIZ

  (* snail is rotated with its volume *)
  fun cw_snail (orientation, x, y) =
      let
          (* put in rotated coordinates *)
          val (rot_x, rot_y) = cw_xy (x, y)
          val rot_orientation = cw_orientation orientation
          val rot_x = 
              case rot_orientation of
                  HORIZ => rot_x - Clip.clip_width Clip.horiz + 1
                | VERT => rot_x - Clip.clip_width Clip.vert + 1
      in
          (rot_orientation, rot_x, rot_y)
      end

  fun movesnail { nexttick, intention } =
      (* Don't move snail when editing. *)
      if !showedit
      then intention
      else
      let
          val () = if List.exists (fn (I_GO _) => true | _ => false) intention
                   then raise Game "snail should never have long-lived GO intentions"
                   else ()

          (* instantaneous intention is derived from the intention queue
             and the keys being held *)
          val intention_now = contextual_intention (!keymask) @ intention

          fun tile_from_coord (pixelx, pixely) =
              (pixelx div TILEW, pixely div TILEH)

          fun tile_offset_from_coord (pixelx, pixely) =
              (pixelx mod TILEW, pixely mod TILEH)


          fun mask_at (x, y) = 
              let 
                  (* Get actual mask there *)
                  val mask = Room.maskat (!room, x, y)
              in
                  (* Uncomment to enable STARWALKER MODE! *)
                  (* Room.settile (!room, BACKGROUND, x, y) (Tile.fromword 0w32);
                     Room.settile (!room, FOREGROUND, x, y) (Tile.fromword 0w32); *)
                  (* rotate the mask *)
                  mask
              end



          fun clipAt (x, y) =
              let 
                  val (tx, ty) = tile_from_coord (x, y)
                  (* val (mx, my) = tile_offset_from_coord (x, y) *)
                  (* This is ok certified tom7 23 Aug 2010 *)
                  val (mx, my) = (x mod TILEW, y mod TILEH)
                  val ma = mask_at (tx, ty)
              in
                  Tile.clipmask ma
                                (mx, my)
              end


           (*  Is there a mask pixel on this circle point?  *)
          fun sense (x,y) r th =
              let 
                  val x0 = x + Real.fromInt(SNAILW) / 2.0
                  val y0 = y + Real.fromInt(SNAILH) / 2.0
                  val x1 = r * (Math.cos (th)) + x0
                  val y1 = y0 + r * (Math.sin (th)) 
                  val (x2,y2) = (Real.round x1,Real.round y1)
              in
(*                  drawcircle (screen, 
                        x2 - !scrollx,
                        y2 - !scrolly,
                                 1, brightred); *)
                  clipAt (x2,y2)
              end

          val pi =  Math.pi

          fun clearwalk (x,y) dir =
              let val r = 16.0
                  val sensors = [dir - 0.75,
                                 dir - 0.35,
                                 dir - 0.2,
                                 dir,
                                 dir + 0.2,
                                 dir + 0.35, 
                                 dir + 0.75
                                ]
                  val blocked = map (sense (x,y) r) sensors
                  val oneclear = 
                      (List.exists (fn x => not x) blocked)
                  val oneblocked = 
                      (List.exists (fn x =>  x) blocked)
              in 
(*                  print ( "blocked: "); *)
(*                  map (fn x => print ((Bool.toString x) ^ " ")) blocked; *)
(*                  print "\n"; *)
                  oneclear (* andalso oneblocked*)
              end


          fun checknofloor (x,y, upnrm) =
              let val nrm = upnrm + Math.pi
                  val r = 14.0
                  val sensors = [nrm - 0.5,
                                 nrm - 0.4,
                                  nrm - 0.3,
                                  nrm - 0.2,
                                  nrm - 0.1,
                                  nrm,
                                  nrm + 0.1,
                                  nrm + 0.2, 
                                  nrm  + 0.3,
                                  nrm + 0.4,
                                 nrm + 0.5
                                ]
                  val blocked = map (sense (x,y) r) sensors
                  val allclear = 
                      (List.all (fn x => not x) blocked)
                  
              in
(*                  print ( "checknofloor blocked: "); *)
(*                  map (fn x => print ((Bool.toString x) ^ " ")) blocked; *)
(*                  print "\n"; *)
                  if allclear then attached := false else ();
                  ()
              end


          fun adjustsnail (x,y, upnrm) =
              let val nrm = upnrm + Math.pi
                  val r = 11.0
                  val ns = 10
                  val dt = Math.pi / Real.fromInt(ns)
                  val ccwsensors = [
                                     nrm - 10.0 * dt,
                                     nrm - 9.0 * dt,
                                     nrm - 8.0 * dt,
                                     nrm - 7.0 * dt,
                                     nrm - 6.0 * dt,
                                     nrm - 5.0 * dt,
                                     nrm - 4.0 * dt,
                                     nrm - 3.0 * dt,
                                     nrm - 2.0 * dt,
                                     nrm - 1.0 * dt]
                  val cwsensors = [
                                     nrm + 10.0 * dt,
                                     nrm + 9.0 * dt,
                                     nrm + 8.0 * dt,
                                     nrm + 7.0 * dt,
                                     nrm + 6.0 * dt,
                                     nrm + 5.0 * dt,
                                     nrm + 4.0 * dt,
                                     nrm + 3.0 * dt,
                                     nrm + 2.0 * dt,
                                     nrm + 1.0 * dt]
                  fun isahit a =
                      sense (x,y) r a
                  val ccwhits = List.length (List.filter isahit ccwsensors)
                  val cwhits = List.length (List.filter isahit cwsensors)
(*                  val anglestep = Real.fromInt(cwhits - ccwhits) * dt *)
                  val cd = cwhits - ccwhits
                  val anglestep =  0.04 * (Real.fromInt cd )

                  val newnrm = upnrm + anglestep
                  val unsink = if (cwhits + ccwhits >= ns) then 1.0 else 0.0
                  val newx = x + Math.cos(newnrm) * unsink
                  val newy = y + Math.sin(newnrm) * unsink
              in
(*                  print ( "adjustsnail. newnrm =   " ^ Real.toString(newnrm));
                  print ( "\n old =   " ^ Real.toString(upnrm));
                  print ( "\nccwhits =   " ^ Int.toString(ccwhits));
                  print ( "\ncwhits =   " ^ Int.toString(cwhits));
                  print "\n"; *)
                  (newx,newy,newnrm)
              end



          val i = intention_now

(* dx is the linear velocity when we are attached   *)
(* when we are free falling, it is the dx you think it is XXX not yet *)                  
          val dx = !snaildx

          val dx =
              if intends i (I_GO LEFT)
                 andalso dx > ~ MAXWALK
              then dx - 1
              else dx
          val dx = 
              if intends i (I_GO RIGHT)
                 andalso dx < MAXWALK
              then dx + 1
              else dx

          (* and if we don't intend to move at all, slow us down *)
          val dx = 
              if not (intends i (I_GO RIGHT)) andalso
                 not (intends i (I_GO LEFT))
              then (if dx > 0 then dx - 1
                    else if dx < 0 then dx + 1 else 0)
              else dx

          val dy = !snaildy

          val (dx,dy) = if !attached 
                        then (dx,dy)
                        else (!snaildx, dy)




          val dy = dy + 1

          val dy = if dy > TERMINAL_VELOCITY
                   then dy - 1
                   else if dy < ~ TERMINAL_VELOCITY
                        then dy + 1
                        else dy


      


          val x = ref (Real.fromInt(!snailx))
          val y = ref (Real.fromInt(!snaily))
          val nrm = !snailnormal

          val (newx, newy, nrm) = if !attached then adjustsnail (!x,!y,nrm) else (!x,!y,nrm)
          val () = (x:= newx; y:= newy)

          val () = checknofloor(!x,!y,nrm) 
          val fwd = dx >= 0
          val walkdir = if fwd 
                        then nrm + Math.pi / 2.0 
                        else nrm - Math.pi / 2.0


          val  i = 
              if intends i (I_JUMP)
              then (attached := false;
(*                    print "jumping\n"; *)
                    x:= (!x + 10.0*Math.cos(nrm));
                    y:= (!y + 10.0*Math.sin(nrm));
                    List.filter (fn I_JUMP => false | _ => true) i)
              else  List.filter (fn I_JUMP => false | _ => true) i






          fun nudgex _ = 
              let 
              in
                  if clearwalk (!x,!y) walkdir
                  then let 
                          val xstep = Math.cos(walkdir)
                          val ystep = Math.sin(walkdir)
                      in
(*                          print ( "walking in direction: " ^ (Real.toString walkdir) ^ "\n"); *)
                          x := (!x + xstep);
                          y := (!y + ystep);
                          ()
                      end
                  else ()
              end



          fun nudgefall _ = 
              let 
                  val xstepwalk =  0.0  (* Math.cos(walkdir) *)
                  val ystepwalk =  0.0 (* Math.sin(walkdir) *)
                  val (curx,cury) =
                      (case Room.propat (!room, 
                                        (!snailx + SNAILW div 2) div TILEW, 
                                        (!snaily + SNAILH div 2) div TILEH) of
                          PCURRENT UP =>
                          (0.0,~1.0)
                        | PCURRENT DOWN =>
                          (0.0,1.0)
                        | PCURRENT LEFT =>
                          (~1.0,0.0)
                        | PCURRENT RIGHT =>
                          (1.0,0.0)
                        | _ => 
                          (0.0, 0.0) )
                  val fallx = 1.0 * curx (* TODO add currents here *)
                  val fally = 1.7 * cury + 1.0 (* add one for gravity *)
                  val falldir = Math.atan2(fally, fallx)
              in
                  if clearwalk (!x,!y) falldir
                  then let 
                      in
(*                          print ( "falling in direction: " ^ (Real.toString falldir) ^ "\n");
                          print ( "fallx= " ^ (Real.toString fallx) ^ "\n");
*) 
                          x := (!x + fallx);
                          y := (!y + fally);
                          ()
                      end
                  else attached := true
              end


          fun updatesnaildir _ =
              let val th = Real.realMod(Real.realMod ((!snailnormal /( Math.pi * 2.0)) + 1.0))
                  val d = if th >= 0.125 andalso th <= 0.375 
                          then UP 
                          else if th >= 0.375 andalso th <= 0.625
                          then RIGHT
                          else if th >= 0.625 andalso th <= 0.875
                          then DOWN
                          else LEFT
              in
                  snaildir := d
              end
              

          val intent = List.filter (fn I_JUMP => true | _ => false) i

      in



          if (not (!attached)) 
          then Util.for 0 (abs dy - 1) nudgefall
          else  Util.for 0 (abs dx - 1) nudgex;


          snailx := Real.round(!x);
          snaily := Real.round(!y);
          snaildx := dx;
          snaildy := dy;
          snailnormal := nrm;
          updatesnaildir();
          slime := NONE;

          intent
      end
(*
          (* Do the requisite number of rotations to make snaildir into down. *)
          fun rot cw =
              case !snaildir of
                  DOWN => (fn x => x)
                | RIGHT => cw
                | UP => cw o cw
                | LEFT => cw o cw o cw

          (* Perform the inverse rotation. (unrot cw) o (rot cw) is the identity,
             if cw is a clockwise rotation. *)
          fun unrot cw =
              case !snaildir of
                  DOWN => (fn x => x)
                | RIGHT => cw o cw o cw
                | UP => cw o cw
                | LEFT => cw

          (* XXX have to implement currents somewhere, which flip you if they can *)

          val rot_intention_now = map (rot cw_intention) intention_now
          val orientation = case !snaildir of
              UP => HORIZ
            | DOWN => HORIZ
            | LEFT => VERT
            | RIGHT => VERT
          val (rot_orientation, rot_x, rot_y) = rot cw_snail (orientation, !snailx, !snaily)
          val () = if rot_orientation <> HORIZ
                   then raise Game "Tom is a bad programmer"
                   else ()

          val (rot_dx, rot_dy) = rot cw_xy (!snaildx, !snaildy)
          val rot_slime = Option.map (rot cw_xy) (!slime)

          fun rot_is_stopped rot_orientation (rot_x, rot_y, rot_dx, rot_dy) =
              let
                  fun couldpushrock (rock as (ref (Rock { x, y, dx = rdx, dy = rdy })), dx, dy) =
                    let 
                      val result =
                        if (Clip.clipped (!room) (List.filter (fn r => not (r = rock)) rocks) Clip.rock (x + dx, y + dy)) then
                          (0, 0)
                        else 
                          (dx, dy)
                      val () =
                        (print ("Trypushrock: " ^ (Int.toString (#1 result)) ^ ", " ^ (Int.toString (#2 result)) ^ "\n"))
                    in result end

                  (* first, put local coordinates back in world coordinates. *)
                  (* also need to unrotate the orientation. *)
                  val (orientation, x, y) = unrot cw_snail (rot_orientation, rot_x, rot_y)
                  val (dx, dy) = unrot cw_xy (rot_dx, rot_dy)
                  val clip = case orientation of
                      HORIZ => Clip.horiz
                    | VERT => Clip.vert
              in
                  case (Clip.cliplist (!room) rocks clip (x + dx, y + dy)) of
                      nil => false
                    | [Clip.ClipRock (rock as (ref (Rock _)))] =>
                        let val pushresult = couldpushrock (rock, dx, dy)
                          val () = print ("Trying to push rock by (" ^
                            (Int.toString dx) ^ ", " ^ (Int.toString dy) ^
                            "), got (" ^ (Int.toString (#1 pushresult)) ^
                            ", " ^ (Int.toString (#2 pushresult)) ^ ")\n")
                        in
                          (case pushresult of
                              (0, 0) => true
                            | (n, 0) => not (n = dx) (* XXX no partial push *)
                            | (0, n) => not (n = dy) (* XXX no partial push *)
                            | (m, n) => true) (* XXX this would be a poor idea
                              because of my shitty implementation. See
                              eject_rock, which only works orthogonally, and
                              could push a rock into the geometry if we're not
                              careful. *)
                        end
                    | _ => true
              end
          
          fun rot_mask_at (rot_x, rot_y) = 
              let 
                  (* first, put local coordinates back in world coordinates. *)
                  val (x, y) = unrot cw_xy (rot_x, rot_y)
                  (* Get actual mask there *)
                  val mask = Room.maskat (!room, x, y)
              in
                  (* Uncomment to enable STARWALKER MODE! *)
                  (* Room.settile (!room, BACKGROUND, x, y) (Tile.fromword 0w32);
                     Room.settile (!room, FOREGROUND, x, y) (Tile.fromword 0w32); *)
                  (* rotate the mask *)
                  rot cw_mask mask
              end

          fun rot_tile_from_coord (rot_pixelx, rot_pixely) =
              let val (pixelx, pixely) = unrot cw_xy (rot_pixelx, rot_pixely)
              in rot cw_xy (pixelx div TILEW, pixely div TILEH)
              end

          fun rot_tile_offset_from_coord (rot_pixelx, rot_pixely) =
              let val (pixelx, pixely) = unrot cw_xy (rot_pixelx, rot_pixely)
              in rot cw_xy (pixelx mod TILEW, pixely mod TILEH)
              end
      
          (* XXX this is some bullshit for wjl. actually it might be fine to do this here,
             if we rotate the prop...? *)
          val (cx, cy) = (Clip.center Clip.horiz)
          val rot_dy =
              if !snaildir = DOWN andalso
                 Room.propat (!room, (!snailx + cx) div TILEW, (!snaily + cy) div TILEH) = PCURRENT UP
              then rot_dy - 2
              else rot_dy

          (* results in local coordinates! *)
          val (rot_intent, rot_x, rot_y, rot_dx, rot_dy, rot_new_slime, rot_new_orientation) =
              movesnail_oriented { i = rot_intention_now,
                                   x = rot_x,
                                   y = rot_y,
                                   dx = rot_dx,
                                   dy = rot_dy,
                                   slime = rot_slime,
                                   tile_from_coord = rot_tile_from_coord,
                                   tile_offset_from_coord = rot_tile_offset_from_coord,
                                   is_stopped = rot_is_stopped,
                                   mask_at = rot_mask_at }

          (* Only intention we save is jumping, so we don't even need to rotate back.
             (but we would just call unrot cw_intention) *)
          val intent = List.filter (fn I_JUMP => true | _ => false) rot_intent
              
          val (new_orientation, x, y) = unrot cw_snail (rot_orientation, rot_x, rot_y)
          val (dx, dy) = unrot cw_xy (rot_dx, rot_dy)
          val new_slime = Option.map (unrot cw_xy) rot_new_slime
          val new_orientation = unrot cw_dir rot_new_orientation

      in
          snailx := x;
          snaily := y;
          snaildx := dx;
          snaildy := dy;
          slime := new_slime;
          if !snaildir <> new_orientation
          then (snaildir := new_orientation;
                print ("setting orientation: " ^ dirtos new_orientation ^ "\n"))
          else ();
          intent
      end
*)

  (* draws the actual clip mask calculated pixel-by-pixel. slow. *)
  fun drawdebug () =
    if !showclip
    then 
      let in
        (* print "DEBUG..\n"; *)
        Util.for 0 (width - 1)
        (fn x => 
         let in
           Util.for 0 (height - 1)
           (fn y =>
            let
              val xx = !scrollx + x
              val yy = !scrolly + y
              val tx = xx div TILEW
              val ty = yy div TILEH
            in
              (* tile mask, absolute *)
              case Room.maskat (!room, tx, ty) of
                MEMPTY => ()
              | m => 
                  if Tile.clipmask m (xx mod TILEW, yy mod TILEH)
                  then SDL.drawpixel (screen, x, y, color (0w255, 0w0, 0w0, 0w255))
                  else SDL.drawpixel (screen, x, y, color (0w0, 0w255, 0w0, 0w255))

              (* (* clip mask in config space *)
              if Clip.clipped (!room) Clip.std (!scrollx + x, !scrolly + y)
              then SDL.drawpixel (screen, x, y, color (0w255, 0w0, 0w0, 0w255))
              else SDL.drawpixel (screen, x, y, color (0w0, 0w255, 0w0, 0w255))
              *)
            end)
(*          ; if x mod 16 = 0 then flip screen else () *)
         end)
        (* print "DONE.\n" *)
      end
    else ()

    (*
  fun freeanemone n =
    let
      val freeanemone n
    in
    end
    *)


  fun moveobjects () =
      let
          fun doobj (Bullet { x, y, dx, dy }) =
              let 
                  (* XX only dx supported now *)
                  val xi = if dx < 0 then ~1 else 1
                  val xr = ref x
                  val stopx = ref false
                  fun nudgeh _ = 
                      let val next = !xr + xi
                      in
                          if !stopx orelse Clip.clipped (!room) rocks Clip.bullet (next, y)
                          then stopx := true
                          else xr := next
                      end
              in
                  Util.for 0 (abs dx - 1) nudgeh;
                  if !stopx then NONE
                  else SOME (Bullet { x = !xr, y = y, dx = dx, dy = dy })
              end
            | doobj (Monster { which, dir, x, y, dx, dy }) =
              let
                val clip = 
                    case which of
                        STAR => Clip.star
                      | SPIDER => Clip.spider
                (* XXX, sorry, crabs cannot move any more. :) *)
                (* val (_, x, y, dx, dy) = moveent clip ([I_GO dir, I_JUMP], x, y, dx, dy) *)

                val dir = if dx = 0
                          then dir_reverse dir
                          else dir

              in
                (* never goes away *)
                SOME (Monster { which = which, dir = dir, x = x, y = y, dx = dx, dy = dy })
              end
      in
          objects := List.mapPartial doobj (!objects)
      end

  fun moverock clip otherrocks (foo as { x, y, dx, dy }) =
    (* XXX: We take it as an article of faith that dx is always 0 for a rock.  *)
    let
      val dy = if dy < TERMINAL_VELOCITY then dy + 1 else dy
      fun can_movey min max =
        if min = max then min else
        if Clip.clipped (!room) otherrocks clip (x, y + min + 1) then min
        else can_movey (min+1) max
    in
      { x = x + dx, y = y + (can_movey 0 dy), dx = dx, dy = dy }
    end

  fun ejectrock (rock as (ref (Rock { x, y, dx = rdx, dy = rdy})), { width, height })
                (snailx, snaily, { width = snailwidth, height = snailheight }) = 
    let
      (* d is the distance we'd have to go; dx,dy is the direction we'd have to
         go in. We will choose the one with smallest d. Since we're colliding
         they'd all better be positive. *)
      val poss = [
                  { d = snailx+snailwidth - x, dx = 1, dy = 0 },
                  { d = x+width - snailx, dx = ~1, dy = 0 },
                  { d = y+height - snaily, dx = 0, dy = ~1 },
                  { d = snaily+snailheight - y, dx = 0, dy = 1 }]
      val best = ref (hd poss)
      fun findbest [] = ()
        | findbest (x::xs) = (if (#d x) < (#d (!best)) then best := x else ();
                              findbest xs)
      val () = findbest poss
      val { d, dx, dy } = !best
      val () = rock := (Rock { x = x + (d * dx), y = y + (d * dy), dx = rdx, dy = rdy })
    in
      ()
    end

  fun moverocks () =
      let
          fun dorock (rock as (ref (Rock r))) =
            (rock := Rock (moverock Clip.rock
                          (List.filter (fn rx => not (rx = rock)) rocks) 
                          r))
          val clip = case !snaildir of
              UP => Clip.horiz
            | DOWN => Clip.horiz
            | LEFT => Clip.vert
            | RIGHT => Clip.vert
          val () = (case (Clip.cliplist (!room) rocks clip (!snailx, !snaily)) of
              nil => ()
            | [Clip.ClipRock (rock as (ref ( Rock _)))] =>
                ejectrock (rock, Clip.clip_rect Clip.rock) (!snailx, !snaily, Clip.clip_rect clip)
            | s => ()   ) (* print("unhandled\n")) *)
      in
          List.app dorock rocks
      end

  fun drawmaskzone () =
      case (!showedit, !showmask, !editmode) of
          (true, true, _) => drawzone DRAWMASK
        | (true, _, ZMASK) => drawzone DRAWMASK
        | _ => ()

  fun drawpropzone () =
      case (!showedit, !showprop, !editmode) of
          (true, true, _) => drawzone DRAWPROP
        | (true, _, ZPROP) => drawzone DRAWPROP
        | _ => ()

  fun drawlightmap () =
      case Room.lightmap (!room) of
          NONE => ()
        | SOME m => SDL.blit (m, !scrollx, !scrolly, width, height, screen, 0, 0)

  (* The idea is that this should give the snail's center *)
  fun snailpos () = 
      let 
        (* Gives pretty desirable desired rounding behavior *)
        (* XXX will be messed up by rotation *)
        val sx = (!snailx + SNAILW div 2) div TILEW
        val sy = (!snaily + SNAILH div 2) div TILEH 
      in
        (sx, sy)
      end

  fun consistent_dir (UP, dx, dy) = dy < 0
    | consistent_dir (DOWN, dx, dy) = dy > 0
    | consistent_dir (LEFT, dx, dy) = dx < 0
    | consistent_dir (RIGHT, dx, dy) = dx > 0

  fun say s =
      let in
          print (s ^ "\n");
          message := s
      end

  fun doornums 0 = "red"
    | doornums 1 = "green"
    | doornums 2 = "pink"
    | doornums 3 = "yellow"
    | doornums _ = "ILLEGAL"

  exception FoundDoor of int * int
  (* This finds the top-leftmost door that matches. That's crap because
     doors almost always have width or height or both. Slightly better would
     be to find the center of mass of all matching doors. Best would be to
     figure out the geometry of the source dir, the player's position as he
     entered it, and the geometry of the out door, and then put the player
     in the corresponding place. Since the doors are made up of tiles you
     have to do some battleship shit to figure out the geometry. *)
  fun find_door (desired_dir, desired_idx) =
      let in
          Util.for 0 (Room.ROOMH - 1)
          (fn y =>
           Util.for 0 (Room.ROOMW - 1)
           (fn x =>
            case Room.propat (!room, x, y) of
                PGOTO (d, i) =>
                    if d = desired_dir andalso
                       i = desired_idx
                    then raise FoundDoor (x, y)
                    else ()
              | _ => ()));
          NONE
      end handle FoundDoor p => SOME p

  val lastwarp = ref (Room.start, RIGHT, 0) (* XXX is the start a door? *)

  (* XXX should also take pos, and preserve it in the new door. *)
  (* FIXME should check if the player would be stuck in the new position,
     to preserve invariant! *)
  fun warp (newroom, srcdir, dooridx) =
      let 
          val () = lastwarp := (newroom, srcdir, dooridx)
          val matching_dir = dir_reverse srcdir
          fun move_to_door () =
              case find_door (matching_dir, dooridx) of
                  NONE =>
                      let in
                          say "^4There's no matching door in this room! You should make one...";
                          snailx := TILEW * 20;
                          snaily := TILEH * 20
                      end
                | SOME (x, y) =>
                      let 
                          (* Center of destination tile. *)
                          val cx = TILEW * x + (TILEW div 2)
                          val cy = TILEH * y + (TILEH div 2)

                          (* Center of the tile right AFTER the door, using the
                             src direction. We want currents to be able to carry the player,
                             but they can't overlap doors. *)

                          val cx = cx + dir_dx srcdir * TILEW
                          val cy = cy + dir_dy srcdir * TILEH

                          (* But we want to put the center of this on the snail's center, not
                             the top-left corner. *)
                          (* XXX should check orientation?  or set it? *)
                          val (wx, wy) = Clip.center Clip.horiz
                          val sx = cx - wx
                          val sy = cy - wy
                      in
                          snailx := sx;
                          snaily := sy;
                          attached := false
                      end
          val newsong = Room.getsong( newroom)
      in
          say "^3LET'S DO THE ROOM WARP AGAIN!";
          if (Room.getsong (!room) <> newsong )
          then           Music.playsong (newsong)
          else ();
          room := newroom;

          (* find the corresponding door. *)
          move_to_door()
      end

  fun checkwarp () =
      let
        val (sx, sy) = snailpos ()
      in
        case Room.propat (!room, sx, sy) of
            PGOTO (dir, i) =>
              (* Have to have consistent velocity to enter. *)
              if consistent_dir (dir, !snaildx, !snaildy)
              then (case Room.dirfrom (!room) dir of
                      NONE => say "OMG there is no room in that direction!"
                    | SOME newroom => warp (newroom, dir, i))
              else say "go in the correct direction to enter the door"
          | _ => ()
      end

  fun checkmouseover () = ()

  exception QUITTER

  val win_cutscene = 
      map (fn (x, y) => (Images.requireimage x, y))
      [
        ("win-1-1.png", "YOU BOUNCED CRAB")
      ]

  (* XXX not used *)
  val credits_cutscene =
      [
        (Images.requireimage "credits.png", "We did it !!")
      ]

  val initial_cutscene =
      map (fn (x, y) => (Images.requireimage x, y))
          [
           ("cutscene-0-0.png", "\nPresented by CORNER CASE"),
           ("cutscene-1-1.png", "What a beautiful day"), 
           ("cutscene-1-2.png", "I have received mail"), 
           ("cutscene-1-2.png", "I will visit\nmy trusted friend Roshelle" ^
                                "\nat her place of residence") ]

  val murder_cutscene =
      map (fn (x, y) => (Images.requireimage x, y))
          [
           ("cutscene-1-3.png", "OH NO! THIS IS MURRRRDEERRRR!"),
           ("cutscene-1-3.png", "I must avenge Roshelle" ^ 
                                "\nAnd free all the anemones")]


  fun cutscene (music, musicfn, scenes) = 
    let 
      fun advance (img, text, n, remaining_scenes) = 
          if n < size text
          then loop (img, text, size text, remaining_scenes)
          else (case remaining_scenes of 
                  [] => ()
                | (img, text) :: remaining_scenes => 
                  loop (img, text, 0, remaining_scenes))

      and loop (img, text, n, remaining_scenes) =
            let
              val n = Int.min (n + 1, size text)
            in
              clearsurface (screen, color (0w255, 0w0, 0w0, 0w0)); 
              SDL.blitall (img, screen, 48, 48);
              Font.Huge.drawcenter 
                  (screen, 400, 440, String.extract(text, 0, SOME n));
              flip screen; 
              
              case pollevent () of 
                SOME (E_KeyDown _) => advance (img, text, n, remaining_scenes)
              | SOME (E_MouseDown _) => advance (img, text, n, remaining_scenes)
              | SOME E_Quit => raise QUITTER
              | _ => loop (img, text, n + 1, remaining_scenes)
            end
    in
      musicfn music;
      case scenes of 
        [] => ()
      | (img, text) :: remaining_scenes => 
        loop (img, text, 0, remaining_scenes)
    end

  fun cuttext (music, musicfn, texts) =
    let  
      fun advance (text, n, remaining_scenes) =
        if n < size text 
        then loop (text, size text, remaining_scenes)
        else (case remaining_scenes of
                  [] => ()
                | text :: remaining_scenes =>
                    loop (text, 0, remaining_scenes))
      and loop (text, n, remaining_scenes) =
        let  
          val n = Int.min (n + 1, size text)
        in   
          clearsurface (screen, color (0w255, 0w0, 0w0, 0w0));
          Font.Huge.drawcenter
            (screen, 400, 300, String.extract(text, 0, SOME n)); 
          flip screen;
          case pollevent () of 
            SOME (E_KeyDown _) => advance (text, n, remaining_scenes)
          | SOME (E_MouseDown _) => advance (text, n, remaining_scenes)
          | SOME E_Quit => raise QUITTER
          | _ => loop (text, n + 1, remaining_scenes)
        end  
    in
      musicfn music;
      case texts of
        [] => ()
      | text :: remaining_texts =>
          loop (text, 0, remaining_texts)
    end

  val credits = [ "Programming, art, music, &c.:", 
                  "8",
                  "chrisamaphone",
                  "dr tom7 phd",
                  "dwrensha",
                  "gwillen",
                  "jcreed",
                  "not-so-young neil",
                  "portable rob",
                  "rob classic",
                  "wjl",
                  "thank you !!",
                  "(can you beat 3:20?)" ]

  fun win () =
  let in
      cutscene (Music.happy, SDLMusic.loop, win_cutscene);
     (* cutscene (Music.theme, SDLMusic.loop, credits_cutscene); *)
      cuttext (Music.theme, SDLMusic.loop, credits);
      raise Game "GAME WON"
  end

  fun checkaction () = 
      let 
        val (sx, sy) = snailpos ()
      in
        case Room.actionat (!room, sx, sy) of
          Say s => message := "\"^5" ^ s ^ "^<\""
        | Free n => 
            let 
              val alreadyfreed = (AnemoneSet.member (!anemonesfree,n))
              val _ = anemonesfree := (AnemoneSet.add (!anemonesfree,n))
              val anemonesleft = !anemonestotal - (AnemoneSet.numItems (!anemonesfree))
              val namelist = String.concatWith "," (AnemoneSet.listItems (!anemonesfree))
            in
              if anemonesleft = 0
              then win ()
              else 
                if alreadyfreed
                then (message := ("\"^5" ^ "You freed "^n^
                            "! You have freed "^namelist^
                            ".\nYou still need to free the other "^
                            Int.toString anemonesleft^"^<\""))
                else (message := ("\"^5" ^ "You freed " ^n^
                            "! You have freed "^namelist^
                            "!\nYou still need to free the other "^
                            Int.toString anemonesleft^"^<\""))
            end
        | PlayMusic songname => Music.playsong songname
        | _ => ()
      end

  fun loop { nexttick, intention } =
      if getticks () > nexttick
         andalso (not (!paused) orelse !advance)
      then 
          let
              (* Reset message, determine if mouseover/action resets it.
                 XXX Rob fix *)
              val () =               message := 
                                     "^1Nemo 911^<:\n^3Anemonies\nAs\nWeapons^<!"

              (* Before we do anything, check the action. It might warp us to
                 a new room. *)
              val () =               checkwarp ()
              val () =               checkaction ()

              val () =               clearsurface (screen, color (0w255, 0w33, 0w35, 0w99))
              val () =               setscroll intention
              val () =               drawzone (DRAW BACKGROUND)
              val () =               drawdebug ()
              val () =               drawobjects ()
              val () =               drawrocks ()
              val () =               drawsnail true (* XXX shadow *)
              val intention = movesnail { nexttick = nexttick, intention = intention }

              val () =               moveobjects ()
              val () =               moverocks ()  (* Only handles momentum and gravity; rock pushing is in snail movement *)
              val () =               drawsnail false
              val () =               drawzone (DRAW FOREGROUND)
              val () =               drawselection ()
              val () =               drawlightmap ()
              val () =               drawmaskzone ()
              val () =               drawbrushoverlay ()
              val () =               drawpropzone ()
              val () =               drawedit ()

              val () =               checkmouseover ()
              val _  =               Font.Normal.drawlines(screen, 30, 30, !message)
                  
              val speedrun = 
                  let 
                      fun twodigits n = 
                          if n < 10
                             then "0" ^ Int.toString n
                          else Int.toString n
                      val rest = Word32.toInt (SDL.getticks ())
                      (* val ds = (real ms) / 1000.0 *)
                      val m = rest div (60 * 1000)
                      val rest = rest mod (60 * 1000)
                      val s = rest div 1000
                      val ms = rest mod 1000
                      val ms = ms div 10
                  in
(*                      Real.fmt (StringCvt.FIX (SOME 1)) ds *)
                      "^0" ^ twodigits m ^ "^1:^0" ^ twodigits s ^ "^1:^0" ^ twodigits ms
                  end
              val _  =               Font.Huge.draw(screen, 
                                                      10,
                                                      height - Font.Huge.height - 8,
                                                      speedrun)
          in
              flip screen;
              advance := false;
              Animate.increment ();
              key { nexttick = getticks() + 0w20, intention = intention }
          end
      else
          let in
              SDL.delay 1;
              key { nexttick = nexttick, intention = intention }
          end
              
  and click (x, y) =
    (case onscreen (x, y) of
         NONE => print "offscreen click\n"
       | SOME (tx, ty) => 
             let in
                 (* depend on the layer we're currently editing *)
                 case !editmode of
                     ZMASK => Room.setmask (!room, tx, ty) (!editmask)
                   (* XXX should set default mask here *)
                   | ZBG => Room.settile (!room, BACKGROUND, tx, ty) (!editbgtile)
                   | ZFG => Room.settile (!room, FOREGROUND, tx, ty) (!editfgtile)
                   | ZPROP => Room.setprop (!room, tx, ty) (!editprop)
                   | ZBRUSH => 
                       (case #2 (!editbrush) of
                            BPASTE => Room.areapaste (!room,!(!editcopyarea),tx,ty)
                          | _ => Room.replacewithbrush (!room, #1 (!editbrush),
                            #2 (!editbrush), tx, ty))
             end)

  and key ( cur as { nexttick, intention = i } ) = 
    case pollevent () of
       NONE => loop cur
     | SOME evt =>
       (* XXX should pass to window manager *)
       if !showedit andalso Window.handleevent edit_window evt
       then loop cur
       else case evt of
      E_KeyDown { sym = SDLK_ESCAPE } => () (* quit *)

    (* left mouse *)
    | E_MouseDown { button = 1, x, y, ... } =>
           let in
               click (x, y);
               loop cur
           end
    (* right mouse *)
    | E_MouseDown { button = 3, x, y, ... } =>
           let val _ = (editselectx0 := !scrollx + x; editselecty0 := !scrolly + y)
              val _ = editselectmode := true
           (*
             val _ = (print "Right mouse down\n")
            val _ = print ("Button number"^Int.toString button)
            *)
           in
               loop cur
           end
    (* right mouse *)
    | E_MouseUp {button = 0,x, y, ...} =>
        if !editselectmode
        then
          let 
            val _ = (editselectx := !scrollx + x; editselecty := !scrolly + y)
            val _ = editselectmode := false
            (*
            val _ = print ("Right mouse up\n")
            val _ = print ("Button number"^Int.toString button)
            val _ = 
              print ("selected x:"^Int.toString (!editselectx)^",y:"^Int.toString
              (!editselecty)^"\n")
             *)
          in
            loop cur
          end
        else loop cur
       
    (* You have to update in the parent directory to get SDL.mouse_left. *)
    | E_MouseMotion { state : mousestate, x : int, y : int, xrel, yrel, ... } =>
           (if SDL.mouse_left state
            then click (x, y)
            else if !editselectmode 
                 then (editselectx := !scrollx + x; editselecty := !scrolly + y) 
                 else ();
            mousex := x;
            mousey := y;
            loop cur)

    | E_KeyDown { sym = SDLK_LSHIFT } => 
         let in 
             keymask := Word32.orb(!keymask, KEYMASK_SHIFT);
             loop cur
         end

    | E_KeyDown { sym = SDLK_RSHIFT } => 
         let in 
             keymask := Word32.orb(!keymask, KEYMASK_SHIFT);
             loop cur
         end

    | E_KeyDown { sym = SDLK_LCTRL } => 
         let in 
             keymask := Word32.orb(!keymask, KEYMASK_CTRL);
             loop cur
         end

    | E_KeyDown { sym = SDLK_RCTRL } => 
         let in 
             keymask := Word32.orb(!keymask, KEYMASK_CTRL);
             loop cur
         end

(*
        let in
            (* fire bullet *)
            objects := Bullet { x = (case !snailface of
                                         FLEFT => !snailx - (surface_width fireballl)
                                       | FRIGHT => !snailx + surface_width snailr),
                                y = !snaily + 8,
                                dy = 0,
                                dx = (case !snailface of
                                          FLEFT => ~8
                                        | FRIGHT => 8) } :: !objects;
            loop cur
        end
*)

    | E_KeyDown { sym = SDLK_m } =>
        let in
            if !showedit
            then showmask := not (!showmask)
            else ();
            loop cur
        end

    | E_KeyDown { sym = SDLK_p } =>
        let in
            if !showedit
            then showprop := not (!showprop)
            else ();
            loop cur
        end

    
    | E_KeyDown { sym = SDLK_u } => 
      let in
          print "CHEATER\n";
          snailnormal := (Math.pi * 3.0 / 2.0) ;
          loop cur
      end
(*
    | E_KeyDown { sym = SDLK_n } => 
        let in
          objects := Monster { which = SPIDER,
                               x = (case !snailface of
                                      FLEFT => !snailx - (surface_width star)
                                    | FRIGHT => !snailx + surface_width snailr),
                               dir = (case !snailface of
                                      FLEFT => LEFT
                                    | FRIGHT => RIGHT),
                               y = !snaily - 4,
                               dy = ~4,
                               dx = (case !snailface of
                                       FLEFT => ~8
                                     | FRIGHT => 8) } :: !objects;
          loop cur
        end
    *)


    | E_KeyDown { sym = SDLK_RETURN } =>
        let in
          paused := not (!paused);
          loop cur
        end

    | E_KeyDown { sym = SDLK_PERIOD } => 
        let in
          advance := true;
          loop cur
        end
    | E_KeyDown { sym = SDLK_y } =>
        if !showedit
        then let val _ = areacopy () in loop cur end
        else loop cur
    (* XXX: Make insides a function *)
    | E_KeyDown { sym = SDLK_1 } =>
        if !showedit
        then let 
                val copyareas = (Vector.sub (editclipboard,0))
                val _ = editcopymaskarea := (#1 (copyareas))
                val _ = editcopyarea := (#2 (copyareas))
             in 
               loop cur
             end
        else loop cur 
    | E_KeyDown { sym = SDLK_2 } =>
        if !showedit
        then let 
                val copyareas = (Vector.sub (editclipboard,1))
                val _ = editcopymaskarea := (#1 (copyareas))
                val _ = editcopyarea := (#2 (copyareas))
             in 
               loop cur
             end
        else loop cur 
    | E_KeyDown { sym = SDLK_3 } =>
        if !showedit
        then let 
                val copyareas = (Vector.sub (editclipboard,2))
                val _ = editcopymaskarea := (#1 (copyareas))
                val _ = editcopyarea := (#2 (copyareas))
             in 
               loop cur
             end
        else loop cur 
    | E_KeyDown { sym = SDLK_4 } =>
        if !showedit
        then let 
                val copyareas = (Vector.sub (editclipboard,3))
                val _ = editcopymaskarea := (#1 (copyareas))
                val _ = editcopyarea := (#2 (copyareas))
             in 
               loop cur
             end
        else loop cur 
    | E_KeyDown { sym = SDLK_5 } =>
        if !showedit
        then let
            val _ = (!editcopymaskarea) := mirrormasksyaxis (!(!editcopymaskarea))
          in
            loop cur
          end
        else loop cur
    | E_KeyDown { sym = SDLK_6 } =>
        if !showedit
        then let
            val _ = (!editcopymaskarea) := mirrormasksxaxis (!(!editcopymaskarea))
          in
            loop cur
          end
        else loop cur
    | E_KeyDown { sym = SDLK_7 } =>
        if !showedit
        then let
            val _ = (!editcopymaskarea) := mirrormasksdiag (!(!editcopymaskarea))
          in
            loop cur
          end
        else loop cur

    | E_KeyDown { sym = SDLK_r } => 
        (if (! showedit) then
          (snailify ();
           loop cur)
        else (* restart *)
          let in
            warp (!lastwarp);
            loop cur
          end)

    | E_KeyDown { sym = SDLK_h } =>
        let in
          if (not (! showedit)) then
            let in
              snailx := (Room.STARTX * TILEW);
              snaily := (Room.STARTY * TILEH);
              snailface := FRIGHT;
              snaildir := DOWN;
              snailnormal := 3.0 *  Math.pi / 2.0;
              slime := NONE;
              lowerx := 0;
              lowery := 0;
              scrollx := 0;
              scrolly := 0;
              scrolldx := 0;
              scrolldy := 0;
              snaildx := 0;
              snaildy := 0;
              warp (Room.start, RIGHT, 0)
            end
          else ();
          loop cur
        end

           (*
    | E_KeyDown { sym = SDLK_t } => 
          (snailify Tile.block_material;
           loop cur)
           *)

    | E_KeyDown { sym = SDLK_s } =>
        let in
          Room.save (!room);
          print "SAVED.\n";
          loop cur
        end

    | E_KeyDown { sym = SDLK_PAGEDOWN } =>
        let in
          edittileset := (!edittileset + 1) mod NUM_EDIT_TILESETS;
          loop cur
        end

    | E_KeyDown { sym = SDLK_PAGEUP } =>
        let in
          edittileset := (!edittileset - 1) mod NUM_EDIT_TILESETS;
          loop cur
        end

    | E_KeyDown { sym = SDLK_b } => 
        loop { nexttick = nexttick, intention = I_BOOST :: i }

    | E_KeyDown { sym = SDLK_UP } => 
         let in 
             snailnormal := Math.pi * 3.0 / 2.0;
             keymask := Word32.orb(!keymask, KEYMASK_UP);
             loop cur
         end

    | E_KeyDown { sym = SDLK_DOWN } => 
         let in 
             keymask := Word32.orb(!keymask, KEYMASK_DOWN);
             loop cur
         end

    | E_KeyDown { sym = SDLK_LEFT } =>
        (* XXX for these there should also be an impulse event,
           so that when I tap the left key ever so quickly, I at least
           start moving left for one frame. *)
         let in 
             snailface := FLEFT; (* FIXME *)
             keymask := Word32.orb(!keymask, KEYMASK_LEFT);
             loop cur
         end

    | E_KeyDown { sym = SDLK_RIGHT } =>
         let in 
             snailface := FRIGHT; (* FIXME *)
             keymask := Word32.orb(!keymask, KEYMASK_RIGHT);
             loop cur
         end

    | E_KeyDown { sym = SDLK_SPACE } =>
          (* XXX in edit mode, scroll to snail *)
          if intends i (I_JUMP)
          then loop { nexttick = nexttick, intention = i }
          else loop { nexttick = nexttick, intention = (I_JUMP :: i) }


    | E_KeyDown { sym = SDLK_e } => 
          (showedit := not (!showedit);
           loop cur)

    | E_KeyDown { sym = SDLK_z } => 
          (editmode :=
           (case !editmode of
                ZMASK => ZPROP
              | ZPROP => ZBRUSH
              | ZBRUSH => ZBG
              | ZBG => ZFG
              | ZFG => ZMASK);
           loop cur)

    | E_KeyDown { sym = SDLK_c } => 
          (showclip := not (!showclip);
           loop cur)

    | E_KeyDown { sym = SDLK_m } => 
          (showclip := not (!showclip);
           loop cur)


    | E_KeyUp { sym = SDLK_UP } =>
         let in 
             keymask := Word32.andb(!keymask, Word32.notb KEYMASK_UP);
             loop cur
         end

    | E_KeyUp { sym = SDLK_DOWN } =>
         let in 
             keymask := Word32.andb(!keymask, Word32.notb KEYMASK_DOWN);
             loop cur
         end

    | E_KeyUp { sym = SDLK_LEFT } =>
         let in 
             keymask := Word32.andb(!keymask, Word32.notb KEYMASK_LEFT);
             loop cur
         end

    | E_KeyUp { sym = SDLK_RIGHT } =>
         let in 
             keymask := Word32.andb(!keymask, Word32.notb KEYMASK_RIGHT);
             loop cur
         end
    
    (* Was added by cmartens for character creation screen
    | E_KeyDown { sym = SDLK_0 } => 
              let
                val () = CharCreat.loop ()
              in
                loop { nexttick = nexttick, intention = i }
              end
    *)
    | E_KeyDown { sym = SDLK_w } => win ()

    | E_KeyDown { sym } =>
          let in
              print ("unknown key " ^ SDL.sdlktos sym ^ "\n");
              loop { nexttick = nexttick, intention = i }
          end

(* XXX should do the same thing with joystick as with keys, above.
    | E_JoyDown { button = 3, ... } =>
          if intends i (I_JUMP)
          then loop { nexttick = nexttick, intention = i }
          else loop { nexttick = nexttick, intention = (I_JUMP :: i) }

    | E_JoyDown { button = 1, ... } =>
          (snailface := FRIGHT;
           if intends i (I_GO RIGHT)
           then loop { nexttick = nexttick, intention = i }
           else loop { nexttick = nexttick, intention = I_GO RIGHT :: i })

    | E_JoyUp { button = 1, ... } =>
            (loop { nexttick = nexttick, 
                    intention = List.filter (fn I_GO RIGHT => false | _ => true) i })

    | E_JoyUp { button = 0, ... } =>
            (loop { nexttick = nexttick, 
                    intention = List.filter (fn I_GO LEFT => false | _ => true) i })

    | E_JoyDown { which, button } =>  
            let in
                print ("unknown joydown " ^ Int.toString which ^ ":" ^
                       Int.toString button ^ "\n");
                loop cur
            end
    | E_JoyUp { which, button } =>  
            let in
                print ("unknown joyup " ^ Int.toString which ^ ":" ^
                       Int.toString button ^ "\n");
                loop cur
            end
*)

    | E_Quit => ()
    | _ => loop cur

  (* Main entry point *)

  val () = print "About to enter main loop.\n"

  val () =
      (cutscene (Music.hypersnail, SDLMusic.loop, initial_cutscene); 
       cutscene (Music.march_of_the_starfish, SDLMusic.loop, murder_cutscene);
       SDLMusic.loop Music.theme;
       loop { nexttick = 0w0, intention = nil })
    handle Game s => messagebox ("nemo 911: " ^ s)
        | SDL s => messagebox ("sdl error: " ^ s)
        | SDLMix.Error s => messagebox ("sdlmix error: " ^ s)
        | e => messagebox ("Uncaught exception: " ^ exnName e ^ " / " ^ exnMessage e)
end
