structure Lightmap =
struct

  val num_light_samples = 4
  val invLightSamples = 1.0 / Real.fromInt num_light_samples

  fun clamp min max v = if v > max then max else if v < min then min else v
  fun clamp' min max v = if v > max then max else if v < min then min else v

  val maxDarkness = 0w200

  (* the light equation *)
  (* ((int * int) * Tile.lightdata * real) list -> Word8.word * Word8.word * Word8.word * Word8.word *)
  (* lighting: 0x00 (lightest) to 0xAA (darkest) *)
  fun light_eqn lights =
    let fun getAttenuation {constant,linear,quadratic} dist = 1.0 / (constant + linear*dist + quadratic*dist*dist)
        fun mapRGB f (r,g,b) = (f r, f g, f b)
        fun mapRGBA f (r,g,b,a) = (f r, f g, f b, f a)

        fun scale_color color = Real.fromInt (Word8.toInt color) * invLightSamples
        fun scale_rgb (_,{red,green,blue,intensity, attenuation},dist) =
            let val atten = getAttenuation attenuation dist
                val c = mapRGBA scale_color (red,green,blue, 0w255)
                val c = mapRGBA (fn c => c * atten) c
            in c end

        fun add_rgba ((r,g,b,a),(r',g',b',a')) = (r+r',g+g',b+b',a+a')

        fun realToWord8 r =
            if Real.isFinite r then
                let val i = clamp 0 255 (Real.trunc r)
                in Word8.fromInt i end
            else 0w255

        val (rr,rg,rb,ra) = List.foldl add_rgba (0.0, 0.0, 0.0, 0.0) (List.map scale_rgb lights)
        fun scaleByAlpha x = x * (ra / 255.0)
        val (rr, rg, rb) = mapRGB scaleByAlpha (rr,rg,rb)
        val (r,g,b,alpha) = mapRGBA realToWord8 (rr,rg,rb,ra)

        val invAlpha = clamp' 0w00 maxDarkness (0wxff - alpha)
    in
        (r, g, b, invAlpha)
    end

  fun mk_rgba (clipmask,lights) (px,py) =
    if Array2.sub (clipmask, px, py)
    then
        let val lights' = Raytrace.raytrace (clipmask, lights, (px,py))
        in
            case lights' of nil => (0w0,0w0,0w0,maxDarkness)
                          | _ => light_eqn lights'
        end
    (* clipped pixel, zero opacity *)
    else (0w0, 0w0, 0w0, 0w0)

  fun lightmap room =
    let
        open Tile
        val () = print ("rendering lightmap for room " ^ Room.getname room ^ "\n")
        val lights = Room.prop_filter (fn PLIGHT _ => true | _ => false) room
        fun maskat (x,y) =
            let val tx = Int.quot (x, TILEW)
                val ty = Int.quot (y, TILEH)
                val ox = Int.rem (x, TILEW)
                val oy = Int.rem (y, TILEH)
            in
            not (Tile.clipmask (Room.maskat (room, tx, ty)) (ox, oy))
            end
        val clipmask = Array2.tabulate Array2.ColMajor (Room.ROOMW*TILEW, Room.ROOMH*TILEH, maskat)
        val mt = MersenneTwister.init32 0w42
        fun jitterSpot (x,y) _ =
            let val x' = x + MersenneTwister.random_nat mt (TILEW+1)
                val y' = y + MersenneTwister.random_nat mt (TILEH+1)
            in (x',y') end
        fun tile_to_pixel (v,dim) = v * dim
        fun get_rgba ( rgba) = rgba
        fun scale_tile ((x,y),PLIGHT color) =
          let val (x,y) = (x * TILEW, y * TILEH)
              val xy = (x + TILEW div 2, y + TILEH div 2)
              val rgba = lightcolor color

              val spots = List.tabulate (num_light_samples - 1, jitterSpot (x,y))
          (* XXX xy is the middle of the light source,
             maybe use multiple points in the tile *)
          in List.map (fn xy => (xy,rgba)) (xy::spots) end
        val lights' = List.concat (List.map scale_tile lights)
        val b = Room.render (room, mk_rgba (clipmask,lights'))
    in () end

  val () =
    case ! Args.render
     of "all" => Vector.app lightmap Room.rooms
      | name =>
        (lightmap (Room.getbyname name)
         handle ex => print ("no room " ^ name ^ " or got exception " ^ exnName ex ^ "\n"))

end
