(* A room is a fixed-sized (96x96) grid that contains foreground,
   background, clipping mask, and property tiles. (And maybe other
   stuff like entity spawns and lights.) Rooms are four-connected as
   described in world.txt. This file loads all of the rooms and keeps
   track of their connections. *)

signature ROOM =
sig

  datatype layer = FOREGROUND | BACKGROUND

  (* All rooms are loaded at startup, don't worry. *)
  type room
  val rooms : room vector
  (* All rooms are the same size (96x96). This is that
     width and height in tiles. *)
  val ROOMW : int
  val ROOMH : int

  val start : room
  (* The character should start at a position (given by tiles) in the room *)
  val STARTX : int
  val STARTY : int 

               
  val getsong : room -> string

  (* The world specifies how the rooms are arranged.
     Each room has unique neighbors on a 4-connected grid.
     (Or none in that direction). *)
  val rightfrom : room -> room option
  val leftfrom : room -> room option
  val upfrom : room -> room option
  val downfrom : room -> room option
  val dirfrom : room -> Types.dir -> room option

  val propat : room * int * int -> Tile.prop
  val prop_filter : (Tile.prop -> bool) -> room -> ((int * int) * Tile.prop) list
  val actionat : room * int * int -> Types.action
  val setprop : room * int * int -> Tile.prop -> unit

  val maskat : room * int * int -> Tile.mask
  val setmask : room * int * int -> Tile.mask -> unit

  val tileat : room * layer * int * int -> Tile.tile
  val settile : room * layer * int * int -> Tile.tile -> unit

  val areacopy : room * int * int * int * int -> (Tile.prop * Tile.tile * Tile.tile * Tile.mask * int *
  int) list list
  val areapaste : room * (Tile.prop * Tile.tile * Tile.tile * Tile.mask * int * int) list list * int * int -> unit
  val areasnailify : room * Tile.material * int * int * int * int -> unit

  val lightmap : room -> SDL.surface option

  val setwithbrush : room * Tile.material * Tile.brush * int * int -> unit
  val replacewithbrush : room * Tile.material * Tile.brush * int * int -> unit

  val getname : room -> string
  val getbyname : string -> room

  (* Save the current room to its file. *)
  val save : room -> unit

  (* Render the room to "lightmap.<room name>.png". *)
  val render : room * (int * int -> Word8.word * Word8.word * Word8.word * Word8.word) -> bool

end
