
structure Room :> ROOM =
struct
  open Types
  open Tile

  (* Names all the rooms that exist, defines how
     they connect, in a tabular layout *)
  val WORLD_FILE = ! Args.world

  datatype layer = FOREGROUND | BACKGROUND

  exception Room of string

  val NUM_ACTIONS = 36
  val ROOMW = 96
  val ROOMH = 96

  datatype room = R of { name : string,
                         mask : mask array,
                         fore : Tile.tile array,
                         back : Tile.tile array,
                         prop : Tile.prop array,
                         acts : Types.action vector,
                         song : string,
                         lightmap : SDL.surface option ref,
                         left : string option,
                         right : string option,
                         up : string option,
                         down : string option }

  (* XXX: Would be good to check at load-time that all of the
     doors are reasonable (maximum of one rectangle per color
     per direction; none that shoot you out of the world;
     matched pairs; right sizes; etc. *)

  fun empty_room (name, l, r, u, d) = 
      R { name = name,
          mask = Array.array(ROOMW * ROOMH, MEMPTY),
          fore = Array.array(ROOMW * ROOMH, Tile.fromword 0w0),
          back = Array.array(ROOMW * ROOMH, Tile.fromword 0w0),
          prop = Array.array(ROOMW * ROOMH, PPLAIN),
          lightmap = ref NONE,
          acts = Vector.fromList [],
          song = "theme",
          left = l,
          right = r,
          up = u,
          down = d }

  fun inlevel (x, y) = 
      x < ROOMW andalso x >= 0 andalso
      y < ROOMH andalso y >= 0

  fun maskat (R { mask, ... }, x, y) =
    if inlevel (x, y)
    then Array.sub(mask, y * ROOMW + x) 
    else MSOLID

  fun tileat (R { back, fore, ... }, layer, x, y) =
    if inlevel (x, y)
    then Array.sub((case layer of
                        BACKGROUND => back
                      | FOREGROUND => fore), 
                   y * ROOMW + x)
    else case layer of
        BACKGROUND => Tile.OUTER_SPACE_BACKGROUND
      | FOREGROUND => Tile.OUTER_SPACE_FOREGROUND

  fun lightmap (R { lightmap = ref l, ... }) = l

  fun propat (R { prop, ... }, x, y) =
    if inlevel (x, y)
    then Array.sub(prop, y * ROOMW + x)
    else Tile.OUTER_SPACE_PROP

  fun prop_filter pred room =
    Util.ford 0 (ROOMH - 1) nil
     (fn (y,accy) =>
       Util.ford 0 (ROOMW - 1) accy
       (fn (x,accx) =>
         let val prop = propat (room,x,y)
         in
             if pred prop
             then ((x,y),prop)::accx
             else accx
         end))

  fun actionat (room as R { acts, ... }, x, y) =
    case propat (room, x, y) of 
      PACTION a => (Vector.sub(acts, a)
                    handle Subscript => Nothing)
    | _ => Types.Nothing

  fun setmask (R { mask, ... }, x, y) m = 
      if inlevel (x, y)
      then
          Array.update(mask, y * ROOMW + x, m)
      else () (* ignore updates outside of open rectangle *)

  fun settile (R { back, fore, ... }, layer, x, y) t = 
      if inlevel (x, y)
      then Array.update((case layer of
                             BACKGROUND => back
                           | FOREGROUND => fore), 
                        y * ROOMW + x, t)
      else () (* ignore updates outside of open rectangle *)

  fun setprop (R { prop, ... }, x, y) p =
    if inlevel (x, y)
    then Array.update(prop, y * ROOMW + x, p)
    else () (* ignore updates *)


  fun areacopy (rm,x0,y0,x1,y1) =
    let
      fun row y = List.tabulate ((x1 - x0) + 1,fn x => 
        (propat (rm,x + x0,y + y0),
         tileat (rm,FOREGROUND,x + x0,y + y0),
         tileat (rm,BACKGROUND,x + x0,y + y0),
         maskat (rm,x + x0,y + y0),x,y))
      val rows = List.tabulate ((y1 - y0) + 1, fn y => row y)
    in
      rows
    end

  fun areapaste (rm,rows,x,y) = 
    List.app (fn r => 
      List.app (fn (p,tf,tb,m,x',y') => 
        (setmask (rm,x + x',y + y') m; settile (rm,FOREGROUND,x + x',y + y') tf;
         settile (rm,BACKGROUND,x + x',y + y') tb; setprop (rm,x + x',y + y') p)) r) rows
        

  fun maskf MEMPTY = "e"
    | maskf MSOLID = "s"
    | maskf (MRAMP LM) = "lm"
    | maskf (MRAMP MH) = "mh"
    | maskf (MRAMP HM) = "hm"
    | maskf (MRAMP ML) = "ml"
    | maskf (MDIAG NW) = "nw"
    | maskf (MDIAG NE) = "ne"
    | maskf (MDIAG SW) = "sw"
    | maskf (MDIAG SE) = "se"
    | maskf (MCEIL LM) = "clm"
    | maskf (MCEIL MH) = "cmh"
    | maskf (MCEIL HM) = "chm"
    | maskf (MCEIL ML) = "cml"
    | maskf (MLEFT LM) = "llm"
    | maskf (MLEFT MH) = "lmh"
    | maskf (MLEFT HM) = "lhm"
    | maskf (MLEFT ML) = "lml"
    | maskf (MRIGHT LM) = "rlm"
    | maskf (MRIGHT MH) = "rmh"
    | maskf (MRIGHT HM) = "rhm"
    | maskf (MRIGHT ML) = "rml"

(*    | maskf _ = raise Room "maskf" *)

  fun fmask "s" = MSOLID
    | fmask "lm" = MRAMP LM
    | fmask "mh" = MRAMP MH
    | fmask "hm" = MRAMP HM
    | fmask "ml" = MRAMP ML
    | fmask "nw" = MDIAG NW
    | fmask "ne" = MDIAG NE
    | fmask "sw" = MDIAG SW
    | fmask "se" = MDIAG SE
    | fmask "clm" = MCEIL LM
    | fmask "cmh" = MCEIL MH
    | fmask "chm" = MCEIL HM
    | fmask "cml" = MCEIL ML
    | fmask "llm" = MLEFT LM
    | fmask "lmh" = MLEFT MH
    | fmask "lhm" = MLEFT HM
    | fmask "lml" = MLEFT ML
    | fmask "rlm" = MRIGHT LM
    | fmask "rmh" = MRIGHT MH
    | fmask "rhm" = MRIGHT HM
    | fmask "rml" = MRIGHT ML

    | fmask _ = MEMPTY

  fun clightcolor c =
      case c of
          #"w" => LWHITE
        | #"y" => LYELLOW
        | #"b" => LBLUE
        | #"r" => LRED
        | #"g" => LGREEN
        | _ => raise Room "clightcolor?"

  fun lightcolorc LWHITE = "w"
    | lightcolorc LYELLOW = "y"
    | lightcolorc LBLUE = "b"
    | lightcolorc LRED = "r"
    | lightcolorc LGREEN = "g"

  fun cdir c =
      case c of
          #">" => RIGHT
        | #"<" => LEFT
        | #"^" => UP
        | #"v" => DOWN
        | _ => raise Room "cdir?"

  fun dirc Types.UP = "^"
    | dirc Types.DOWN = "v"
    | dirc Types.LEFT = "<"
    | dirc Types.RIGHT = ">"

  fun requireint s =
      case Int.fromString s of
          NONE => raise Room ("required int: " ^ s)
        | SOME i => i

  fun fprop "p" = PPLAIN
    | fprop "" = raise Room "empty fprop"
    | fprop s =
      case explode s of
          [#"g", d, i] => PGOTO (cdir d, requireint (implode [i]))
        | [#"c", d] => PCURRENT (cdir d)
        | #"a" :: rest => PACTION (requireint (implode rest))
        | [#"l", l] => PLIGHT (clightcolor l)
        | _ => raise Room "bad fprop"

  fun propf PPLAIN = "p"
    | propf (PCURRENT d) = "c" ^ dirc d
    | propf (PGOTO (d, i)) = "g" ^ dirc d ^ Int.toString i
    | propf (PACTION x) = "a" ^ Int.toString x
    | propf (PLIGHT l) = "l" ^ lightcolorc l

  fun maskfile r = "room." ^ r ^ ".mask"
  fun bgfile r = "room." ^ r ^ ".bg"
  fun fgfile r = "room." ^ r ^ ".fg"
  fun propfile r = "room." ^ r ^ ".prop"
  fun actsfile r = "room." ^ r ^ ".acts"
  fun lightfile r = "lightmap." ^ r ^ ".png"

  fun save (room as R { name, ... }) =
      let 
          val fm = TextIO.openOut (maskfile name)
          val fb = TextIO.openOut (bgfile name)
          val ff = TextIO.openOut (fgfile name)
          val fp = TextIO.openOut (propfile name)
      in
          Util.for 0 (ROOMH - 1)
          (fn y =>
           (Util.for 0 (ROOMW - 1)
            (fn x =>
             let in
                 TextIO.output(fb, Word32.toString 
                               (Tile.toword (tileat (room, BACKGROUND, x, y))) ^ " ");
                 TextIO.output(ff, Word32.toString 
                               (Tile.toword (tileat (room, FOREGROUND, x, y))) ^ " ");
                 TextIO.output(fm, maskf (maskat (room, x, y)) ^ " ");
                 TextIO.output(fp, propf (propat (room, x, y)) ^ " ")
             end);
            TextIO.output(ff, "\n");
            TextIO.output(fb, "\n");
            TextIO.output(fm, "\n");
            TextIO.output(fp, "\n")
            ));
          app TextIO.closeOut [fm, ff, fb, fp]
      end

  exception Bad

  fun any_file room file parse set =
    let
       val f = StringUtil.readfile file
       val l = Vector.fromList (String.tokens (StringUtil.ischar #"\n") f)
       val c = Vector.map (fn s => Vector.fromList 
                           (String.tokens (StringUtil.ischar #" ") s)) l
    in
        Util.for 0 (ROOMH - 1)
        (fn y =>
         (Util.for 0 (ROOMW - 1)
          (fn x =>
           set (room, x, y) (parse (Vector.sub(Vector.sub(c, y), x)))
           )))
    end

  fun ftile s = 
      case Word32.fromString s of
          NONE => raise Bad
        | SOME w => Tile.fromword w

  fun loadlightmap (R { lightmap, ... }) name =
      let in
          case SDL.Image.load name of
              NONE => (print ("No lightmap for " ^ name ^ "\n");
                       lightmap := NONE)
            | SOME p => lightmap := SOME p
      end

  (* Initializes the (empty) room from a file. If we can't open a file, then
     fill it with some kind of default stuff. *)
  fun initialize_from_file room name =
    let
    in
        print "1";
        any_file room (maskfile name) fmask setmask;
        print "2";
        any_file room (bgfile name) ftile (fn (r, x, y) => settile (r, BACKGROUND, x, y));
        print "3";
        any_file room (fgfile name) ftile (fn (r, x, y) => settile (r, FOREGROUND, x, y));
        print "4";
        any_file room (propfile name) fprop setprop;
        print "5";
        loadlightmap room (lightfile name)
    end handle Io =>
        let in
            print ("error opening room " ^ name ^ "\n");
            print "No saved room, so using default.\n";
            print "XXX I set masks, but not tiles. Enjoy invisible floors\n";
            print "XXX this shit is bullshit fyi\n";
            Util.for 0 (ROOMW - 1) (fn x => setmask (room, x, ROOMH - 1) MSOLID);
            Util.for 4 (ROOMW - 1) (fn x => 
                                    let in
                                        setmask(room, x, ROOMH - 1 - (x div 2))
                                        (MRAMP (if x mod 2 = 0
                                                then LM
                                                else MH));
                                        if x > 12 andalso x mod 2 = 0
                                        then setmask (room, x, ROOMH - (x div 2)) MSOLID
                                        else ()
                                    end);

            Util.for 0 (Int.min (ROOMW, ROOMH) - 1) 
            (fn x => 
             if x mod 3 = 0 
             then setmask (room, x, ROOMH - 1 - x) MSOLID
             else ())
          end

  exception Parseact of string

  fun getsong (R {song, ...}) = song

  fun load_acts (room as R { name, mask, fore, back, prop, acts, song,
                             lightmap, left, right, up, down }) =
    let
      val f = StringUtil.readfile (actsfile name)
        before print ("loaded acts file " ^ (actsfile name) ^ "\n") (* XXX debug *)
      val lines = Vector.fromList (String.tokens (StringUtil.ischar #"\n") f)
      val words = Vector.map (fn s => Vector.fromList
                              (String.tokens (StringUtil.ischar #" ") s)) lines
      val newsong = ref song
      val acts = (* Turn each line of text into an action *)
        Vector.map
          (fn actwords =>
            case Vector.sub(actwords, 0) of
                 "say" => 
                  let
                    val rest = case (Vector.foldr op:: nil actwords) of
                      nil => raise Parseact "Impossible"
                    | (s::nil) => raise Parseact "say what?"
                    | (s::ws) => ws
                  in
                    Types.Say (String.concatWith " " rest)
                  end
                | "music" =>
                  let
                    val song1 = case (Vector.foldr op:: nil actwords) of
                      nil => raise Parseact "IMPOSSIBLE"
                    | (m :: s :: rest) => s
                    | _ => raise Parseact "bad"
                  in
                    newsong := song1;
                    Types.PlayMusic song1
                  end
                | "free" =>
                  let
                    val rest = case (Vector.foldr op:: nil actwords) of
                      nil => raise Parseact "Impossible"
                    | (s::nil) => raise Parseact "free who?"
                    | (s::ws) => ws
                  in
                    Types.Free (String.concatWith " " rest)
                  end
                | s => raise Parseact s)
          words
    in
      R { name = name,
          mask = mask,
          fore = fore,
          back = back,
          prop = prop,
          acts = acts,
          song = !newsong,
          lightmap = lightmap,
          left = left,
          right = right,
          up = up,
          down = down }
    end handle Io => (print "No actions defined\n"; room)


  fun loadroom (name, l, r, u, d) : room =
      let
          val room = empty_room (name, l, r, u, d)
          val room = load_acts room
          val R {acts, name, ... } = room
          val () = 
              if Vector.length acts > 0 then
                 print ("loaded " 
                        ^ Int.toString (Vector.length acts)
                        ^ " actions!\n") (* XXX debug *)
              else ()
      in
          initialize_from_file room name;
          room
      end

  (* Initialize the whole world. *)
  val (rooms, start, STARTX, STARTY) =
    let
        val world = Script.linesfromfile WORLD_FILE

        val (start_name, startx, starty, world) = 
            case world of 
              [] => (print "world.txt has no initialization line.";
                     raise Room "there is no initialization line")
            | init :: world =>
              let val initline = String.tokens StringUtil.whitespec init in
                case ListPair.zip (initline, map Int.fromString initline) of
                  [ (startroom, _), (_, SOME startx), (_, SOME starty) ]
                  => (startroom, startx, starty, world)
                | _ => (print "init line must be a room followed by an x-cord";
                        print "and a y-cord.";
                        raise Room "inproper formatting of initialization line")
              end

        val world = Vector.fromList (map (Vector.fromList o String.tokens StringUtil.whitespec) 
                                     world)    

        (* Abort if the world is not rectangular *)
        val (width, height) =
          case Vector.length world of
              0 => (print "world.txt has to have at least one room.";
                    raise Room "there are no rooms")
            | height =>
              let
                  val width = Vector.length (Vector.sub(world, 0))
                  fun rect row =
                      if row = Vector.length world
                      then ()
                      else (if Vector.length (Vector.sub(world, row)) = width
                            then rect (row + 1)
                            else (print ("row " ^ Int.toString row ^ " of world is ragged: expected "
                                         ^ Int.toString width);
                                  raise Room "ragged world"))
              in
                  rect 1;
                  (width, height)
              end

        fun worldat (row, col) = Vector.sub (Vector.sub (world, row), col)
        fun maybeat (row, col) = 
            if row < 0 orelse col < 0 orelse
               row >= height orelse col >= width
            then NONE
            else case worldat (row, col) of
                "-" => NONE
              | s => SOME s

        val start_room = ref NONE
        val rooms = ref nil
    in
        Util.for 0 (height - 1)
        (fn row =>
         Util.for 0 (width - 1)
         (fn col =>
          case worldat (row, col) of
              "-" => ()
            | name => 
              let val room = 
                      loadroom (name,
                                maybeat (row, col - 1),
                                maybeat (row, col + 1),
                                maybeat (row - 1, col),
                                maybeat (row + 1, col))
              in
                  rooms := room :: !rooms;
                  if name = start_name then start_room := SOME room else ()
              end));
        case !start_room of 
          NONE => (print ("start room \"" ^ start_name ^ "\" did not appear");
                   raise Room "start room?")
        | SOME room => (Vector.fromList (!rooms), room, startx, starty)
    end


  fun setmaskandtile (rm,mat,x,y) (m,xrel,yrel) = 
    let
      val m = case (maskat (rm,x+xrel,y+yrel),m) of 
                   (_,MEMPTY) => MEMPTY
                 | (MEMPTY,m) => m
                 | (m',m) => if m = m' then m else MSOLID
      val () = setmask (rm,x+xrel,y+yrel) m
      val t = Tile.masktotilem (mat,m)
      val () = settile (rm,BACKGROUND,x+xrel,y+yrel) t
    in
      ()
    end

  fun replacemaskandtile (rm,mat,x,y) (m,xrel,yrel) = 
    let
      val erase = if m = Tile.MEMPTY then true else false
      val () = setmask (rm,x+xrel,y+yrel) m
      val t = Tile.masktotilem (mat,m)
      val () = settile (rm,BACKGROUND,x+xrel,y+yrel) t
      val () = if erase then settile (rm,FOREGROUND,x+xrel,y+yrel) t else ()
    in
      ()
    end

  fun setwithbrush (rm,m,br,x,y) =
    List.app (setmaskandtile (rm,m,x,y)) (Tile.brushtomasks br)
  fun replacewithbrush (rm,m,br,x,y) = 
    List.app (replacemaskandtile (rm,m,x,y)) (Tile.brushtomasks br)

  (*val areasnailify : room * int * int * int * int -> unit*)
  fun areasnailify (rm,mat,x0,x1,y0,y1) =
    let
      val brush =  
        (case maskat (rm,x0,y0) of 
            MEMPTY => Tile.BHILLLH ((x1 - x0) + 1,(y1 - y0) + 1)
          | _ => Tile.BHILLHL ((x1 - x0) + 1,(y1 - y0) + 1))
    in
      replacewithbrush (rm,mat,brush,x0,y0)
    end

  fun getname (R { name, ... }) = name

  fun getbyname n =
      case Vector.find (fn (R { name, ... }) => name = n) rooms of
          NONE => raise Room ("room not found: " ^ n)
        | SOME r => r

  fun downfrom (R { down, ... }) = Option.map getbyname down
  fun upfrom (R { up, ... }) = Option.map getbyname up
  fun leftfrom (R { left, ... }) = Option.map getbyname left
  fun rightfrom (R { right, ... }) = Option.map getbyname right

  fun render (R{ name, mask, ... },f) =
    let
        val height = ROOMH * TILEH
        val width = ROOMW * TILEW

        fun cart_to_idx (x,y) = y * width + x
        fun set arr (x,y) (r,g,b,a) =
          let val offset = cart_to_idx (x,y)
          in Array.update (arr, 4*offset, r)
           ; Array.update (arr, 4*offset+1, g)
           ; Array.update (arr, 4*offset+2, b)
           ; Array.update (arr, 4*offset+3, a)
          end
        fun to_rgba arr (x,y) =
          let val offset = cart_to_idx (x,y)
              val rgba = f (x,y)
          in set arr (x,y) rgba
          end

        val iterWidth = width - 1
        val iterHeight = height - 1

        fun reportProgress n = if n mod 5 = 0 then
            print ("Progress (" ^ Int.toString n ^ "/" ^ Int.toString iterWidth ^ ")\r")
            else ()

        (* create and fill array *)
        val a = Array.tabulate (4 * height * width, fn _ => 0w0)
        val () =
            Util.for 0 iterWidth (fn x => (reportProgress x;
              Util.for 0 iterHeight (fn y =>
                to_rgba a (x,y))))
        val () = print "\n"
        val filename = "lightmap." ^ name ^ ".png"
    in PNGsave.save (filename, width, height, a)
    end

  fun dirfrom r dir =
      case dir of
          UP => upfrom r
        | DOWN => downfrom r
        | LEFT => leftfrom r
        | RIGHT => rightfrom r

  val () = print "Rooms loaded.\n"

end
