
(* Defines runtime representation of tiles, masks, and properties.
   The serialization is done in room.sml *)

structure Tile :> TILE =
struct

  exception Tile of string

  structure MT = MersenneTwister

  val mt = MT.init32 0wxdeadbeef

  fun whichrare (rand,rares,acc) =
    if rand - rares < 0 
    then acc 
    else whichrare ((rand - rares),rares - 1,acc + 1)

  fun randomrare rares =
    whichrare ((MT.random_nat mt (rares * rares + rares)) div 2,rares,0)

  (* norm is number of normal tiles *)
  fun randomtile (max,norm) =
    Word32.fromInt
      (if (MT.random_nat mt max) mod 5 = 0 andalso max <> norm 
      then norm + (randomrare (max - norm))
      else MT.random_nat mt norm)

      (* format is (tstart,solids,hdecor,vdecor,normal)
      * tstart = index of first normal tile
      * solids = number of soli
      * hdecor = number of horizontal decoration tiles
      * vdecor = number of vertical decoration tiles
      * normal = number of normal solid tiles (not rare)
      *)
  val select_tile = 0w160
  val eraser_material = (0w0, 1, 0, 0, 1)
  val dirt_material = (0w265, 8, 2, 2, 4)
  val block_material = (0w41, 5, 0, 0, 1)
  val scribble_material = (0w49, 2, 2, 2, 1)

  (* the mask tells us where things can walk *)
  datatype slope = LM | MH | HM | ML
  datatype diag = NW | NE | SW | SE
  datatype mask = 
      MEMPTY 
    | MSOLID 
    | MRAMP of slope
    | MCEIL of slope
    | MDIAG of diag (* corner that is filled in *)
    | MLEFT of slope
    | MRIGHT of slope

  datatype brush =
      BRECT of int * int (* your basic rectangle (w x h) *)
    | BSLOPEDRECT of int * int (* (min slope) (w x h) *)
    | BDIAMOND of int * int (* (max slope) (w x h) *)
    | BERASERECT of int * int (* rectangular eraser *)
    | BOCTAGON of int * int (* (med slope) (w x h) *)
    | BHILLLH of int * int 
    | BHILLHL of int * int 
    | BPASTEMASKS of (mask * int * int) list ref ref
    | BPASTE

  datatype lightcolor = LWHITE | LYELLOW | LBLUE | LRED | LGREEN
  type attenuation = { constant: real, linear:real, quadratic: real }
  type lightdata = { red: Word8.word, green: Word8.word, blue: Word8.word, intensity: real, attenuation: attenuation }
  datatype prop =
      PPLAIN
    | PGOTO of Types.dir * int
    | PCURRENT of Types.dir
    | PACTION of int (* Index into a per-room list of actions *)
    | PLIGHT of lightcolor

  val default_attenuation = {
    constant = 1.8,
    linear = 0.03,
    quadratic = 0.0
  }

  fun lightcolor LWHITE = { red = 0w255 : Word8.word, green = 0w255 : Word8.word, blue = 0w255 : Word8.word, intensity = 0.9, attenuation = default_attenuation }
    | lightcolor LYELLOW = { red = 0w255 : Word8.word, green = 0w241 : Word8.word, blue = 0w140 : Word8.word, intensity = 0.6, attenuation = default_attenuation }
    | lightcolor LBLUE = { red = 0w140 : Word8.word, green = 0w238 : Word8.word, blue = 0w255 : Word8.word, intensity = 0.25, attenuation = default_attenuation }
    | lightcolor LRED = { red = 0w255 : Word8.word, green = 0w140 : Word8.word, blue = 0w140 : Word8.word, intensity = 0.87, attenuation = default_attenuation }
    | lightcolor LGREEN = { red = 0w154 : Word8.word, green = 0w255 : Word8.word, blue = 0w140 : Word8.word, intensity = 1.0, attenuation = default_attenuation }

  (* need to implement this somehow.. *)
  type tile = Word32.word

  (* start tile * fill length * width * height *)
  type material = tile * int * int * int * int

  val TILEW = 16
  val TILEH = 16
  val TILEBITSW = 0w4
  val TILEBITSH = 0w4
  val MASKW = 0w15 : Word32.word
  val MASKH = 0w15 : Word32.word

  val OUTER_SPACE_BACKGROUND = 0w265 (* random dirt. *)
  val OUTER_SPACE_FOREGROUND = 0w0  (* space *)

  val OUTER_SPACE_PROP = PPLAIN
  val resizeable_brushsize = ref (0,0)

  fun areasize (xmax,ymax) [] = (xmax+1,ymax+1)
    | areasize (xmax,ymax) ((_,x,y)::masks) = 
      let
        val (x,y) = (if x > xmax then x else xmax, if y > ymax then y else ymax)
      in
        areasize (x,y) masks
      end

  fun brushsize (BRECT (w,h)) = (w,h)
    | brushsize (BSLOPEDRECT (w,h)) = (w,h)
    | brushsize (BERASERECT (w,h)) = (w,h)
    | brushsize (BDIAMOND (w,h)) = (w,h)
    | brushsize (BOCTAGON (w,h)) = (w,h)
    | brushsize (BHILLLH (l,h)) = (l,h)
    | brushsize (BHILLHL (l,h)) = (l,h)
    | brushsize (BPASTEMASKS masks) = areasize (0,0) (!(!masks))
    | brushsize _ = !resizeable_brushsize

  val quat = Int.quot
  infix 5 quat

  fun clipmask MEMPTY _ = false
    | clipmask MSOLID _ = true
    | clipmask (MRAMP LM) (x, y) = y >= (8 + (7 - x div 2))
    | clipmask (MRAMP MH) (x, y) = y >= (7 - x div 2)
    | clipmask (MRAMP HM) (x, y) = y >= x div 2
    | clipmask (MRAMP ML) (x, y) = y >= (8 + x div 2)
    | clipmask (MCEIL LM) (x, y) = y <= (8 + (7 - x div 2))
    | clipmask (MCEIL MH) (x, y) = y <= (7 - x div 2)
    | clipmask (MCEIL HM) (x, y) = y <= x div 2
    | clipmask (MCEIL ML) (x, y) = y <= (8 + x div 2)

    | clipmask (MLEFT HM) (x, y) = x <= (8 + (7 - y div 2))
    | clipmask (MLEFT ML) (x, y) = x <= (7 - y div 2)
    | clipmask (MLEFT LM) (x, y) = x <= y div 2
    | clipmask (MLEFT MH) (x, y) = x <= (8 + y div 2)


    | clipmask (MRIGHT LM) (x, y) = x >= (8 + (7 - y div 2))
    | clipmask (MRIGHT MH) (x, y) = x >= (7 - y div 2)
    | clipmask (MRIGHT HM) (x, y) = x >= y div 2
    | clipmask (MRIGHT ML) (x, y) = x >= (8 + y div 2)

    | clipmask (MDIAG SW) (x, y)  = y >= x
    | clipmask (MDIAG SE) (x, y)  = y >= (15 - x)
    | clipmask (MDIAG NE) (x, y)  = y < x
    | clipmask (MDIAG NW) (x, y)  = y < (15 - x)

  val TILESET_WIDTH = 32
  (* This changes. Don't rely on it being 32. *)
  val TILESET_HEIGHT = 32

  fun tilecoordsrel mask =
      (case mask of 
          MEMPTY => raise Tile "Impossible mask for relative coordinates"
        | MSOLID => (0,0)
        | (MRAMP LM) => (3,6)
        | (MRAMP MH) => (3,7)
        | (MRAMP HM) => (3,4)
        | (MRAMP ML) => (3,5)
        | (MCEIL LM) => (2,4)
        | (MCEIL MH) => (2,5)
        | (MCEIL HM) => (2,6)
        | (MCEIL ML) => (2,7)
        | (MLEFT LM) => (4,1)
        | (MLEFT MH) => (5,1)
        | (MLEFT HM) => (2,1)
        | (MLEFT ML) => (3,1)
        | (MRIGHT LM) => (4,2)
        | (MRIGHT MH) => (5,2)
        | (MRIGHT HM) => (2,2)
        | (MRIGHT ML) => (3,2)
        | (MDIAG _) => raise Tile "MDIAG is basically dead to us")

  fun masktotile (tstart,MEMPTY) = 0wx0
    | masktotile (tstart,mask) = 
      let
        val (r,c) = tilecoordsrel mask
      in
        tstart + (Word32.fromInt ((r * TILESET_WIDTH) + c))
      end
  fun masktotilem ((tile,max,_,_,norm),MSOLID) = 
      masktotile (tile,MSOLID) + (randomtile (max,norm))
    | masktotilem ((tile,max,_,_,norm),m) = masktotile (tile,m)
  
  val tileset = Images.requireimage "tiles.png"
  val brushset = Images.requireimage "brush.png"
  val PROPW = 16
  val propset = Images.requireimage "prop.png"
  val solid = Images.requireimage "solid.png"
  val error = Images.requireimage "error_frame.png"

  val ramp_lm = Images.requireimage "ramp_lm.png"
  val ramp_mh = Images.requireimage "ramp_mh.png"
  val ramp_hm = Images.requireimage "ramp_hm.png"
  val ramp_ml = Images.requireimage "ramp_ml.png"

  val ceil_hm = Images.requireimage "ceil_hm.png"
  val ceil_ml = Images.requireimage "ceil_ml.png"
  val ceil_lm = Images.requireimage "ceil_lm.png"
  val ceil_mh = Images.requireimage "ceil_mh.png"

  val left_hm = Images.requireimage "left_hm.png"
  val left_ml = Images.requireimage "left_ml.png"
  val left_lm = Images.requireimage "left_lm.png"
  val left_mh = Images.requireimage "left_mh.png"

  val right_hm = Images.requireimage "right_hm.png"
  val right_ml = Images.requireimage "right_ml.png"
  val right_lm = Images.requireimage "right_lm.png"
  val right_mh = Images.requireimage "right_mh.png"

  val diag_ne = Images.requireimage "diag_ne.png"
  val diag_nw = Images.requireimage "diag_nw.png"
  val diag_sw = Images.requireimage "diag_sw.png"
  val diag_se = Images.requireimage "diag_se.png"

  fun actgraphic n =
    Vector.sub (Vector.fromList [ceil_hm, ceil_ml, ceil_lm, ceil_mh,
                  left_hm, left_ml, left_lm, left_mh], n)

  fun maskgraphic MSOLID = solid
    | maskgraphic (MRAMP LM) = ramp_lm
    | maskgraphic (MRAMP MH) = ramp_mh
    | maskgraphic (MRAMP HM) = ramp_hm
    | maskgraphic (MRAMP ML) = ramp_ml
    | maskgraphic (MDIAG NW) = diag_nw 
    | maskgraphic (MDIAG NE) = diag_ne
    | maskgraphic (MDIAG SW) = diag_sw
    | maskgraphic (MDIAG SE) = diag_se
    | maskgraphic (MCEIL HM) = ceil_hm
    | maskgraphic (MCEIL ML) = ceil_ml
    | maskgraphic (MCEIL LM) = ceil_lm
    | maskgraphic (MCEIL MH) = ceil_mh
    | maskgraphic (MLEFT HM) = left_hm
    | maskgraphic (MLEFT ML) = left_ml
    | maskgraphic (MLEFT LM) = left_lm
    | maskgraphic (MLEFT MH) = left_mh
    | maskgraphic (MRIGHT HM) = right_hm
    | maskgraphic (MRIGHT ML) = right_ml
    | maskgraphic (MRIGHT LM) = right_lm
    | maskgraphic (MRIGHT MH) = right_mh

    | maskgraphic _ = error

  fun masktostring MSOLID = "solid"
    | masktostring MEMPTY = "empty"
    | masktostring (MRAMP LM) = "ramp_lm"
    | masktostring (MRAMP MH) = "ramp_mh"
    | masktostring (MRAMP HM) = "ramp_hm"
    | masktostring (MRAMP ML) = "ramp_ml"
    | masktostring (MDIAG NW) = "diag_nw "
    | masktostring (MDIAG NE) = "diag_ne"
    | masktostring (MDIAG SW) = "diag_sw"
    | masktostring (MDIAG SE) = "diag_se"
    | masktostring (MCEIL HM) = "ceil_hm"
    | masktostring (MCEIL ML) = "ceil_ml"
    | masktostring (MCEIL LM) = "ceil_lm"
    | masktostring (MCEIL MH) = "ceil_mh"
    | masktostring (MLEFT HM) = "left_hm"
    | masktostring (MLEFT ML) = "left_ml"
    | masktostring (MLEFT LM) = "left_lm"
    | masktostring (MLEFT MH) = "left_mh"
    | masktostring (MRIGHT HM) = "right_hm"
    | masktostring (MRIGHT ML) = "right_ml"
    | masktostring (MRIGHT LM) = "right_lm"
    | masktostring (MRIGHT MH) = "right_mh"

  (* assumes power of two *)
  val NSEEDS = 16
  val seeds = 
      Vector.fromList 
      [0wxD76AA478 : Word32.word,
       0wxE8C7B756,
       0wx242070DB,
       0wxC1BDCEEE,
       0wxF57C0FAF,
       0wx4787C62A,
       0wxA8304613,
       0wxFD469501,
       0wx698098D8,
       0wx8B44F7AF,
       0wxFFFF5BB1,
       0wx895CD7BE,
       0wx6B901122,
       0wxFD987193,
       0wxA679438E,
       0wx49B40821]

  (* should do something not ad hoc here... *)
  fun animate (n, t, wx, wy) =
      (* star *)
      if t >= 32 andalso t <= 34
      then 32 + (((n div 8) + (t - 32)) mod 3)
      else 
          (* flower *)
          if t = 134 andalso (((n div 4) + wx) mod 3 = 1)
          then 135
          else 
              (* bubble *)
              if t = 128
              then (case ((n div 7) + wx + wy) mod 9 of
                        0 => 128
                      | 1 => 129
                      | 2 => 130
                      | _ => 0 (* blank *))
              else
                  (* water surface *)
                  if t = 162
                  then 162 + ((n div 5) mod 4)
                  else t

  fun drawmask (MEMPTY, surf, x, y) = ()
    | drawmask (m, surf, x, y) =
      SDL.blitall (maskgraphic m, surf, x, y)

  fun drawselection (surf,x,y) = SDL.blit (brushset,TILEW * 8,0,TILEW,TILEH,surf,x,y)

  (* standard order within tileset *)
  fun diroff Types.UP = 0
    | diroff Types.DOWN = 1
    | diroff Types.LEFT = 2
    | diroff Types.RIGHT = 3

  fun lightoff LWHITE = 0
    | lightoff LYELLOW = 1
    | lightoff LBLUE = 2
    | lightoff LRED = 3
    | lightoff LGREEN = 4

  fun drawprop (PPLAIN, _, _, _) = ()
    | drawprop (m, surf, x, y) =
      let
        val ACTW = 7
        val (px, py) =
          case m of 
            PPLAIN => raise Tile "impossible"
          | PCURRENT d => 
            let val d = diroff d in (d, 0) end
          | PGOTO (d, i) =>
            let val d = diroff d in (d + 5, i) end
          | PACTION x => (x mod ACTW + 9, x div ACTW)
          | PLIGHT l => (lightoff l, 4)
      in
          SDL.blit (propset, 
                    px * TILEW,
                    py * TILEH,
                    TILEW, TILEH,
                    surf,
                    x, y)
      end

  fun randomize (w1, w2, w3) = 
      let
          val xx = w1
          val yy = w2 + Word32.andb(xx, 0w3)
          val zz = Vector.sub(seeds, 
                              Word32.toInt(Word32.andb(Word32.fromInt (NSEEDS - 1),
                                                       yy)))
          (* val () = print (" - " ^ Word32.toString zz ^ "\n") *)
          val h = Word32.xorb(xx * yy, 0wxDEADBEEF) + (zz * 0w31337)
      in
          h
      end

  fun drawat (_, 0w0, surf, x, y, _, _) = ()
      (* meta-dirt *)

    | drawat (time, 0w65, surf, x, y, wx, wy) =
      let
          val n = randomize (Word32.fromInt wx,
                             Word32.fromInt wy,
                             0w33)
          val u = Word32.andb(Word32.>> (n, 0w7), 0w15)
          val t = 
              case u of
                  0w0 => #1 dirt_material
                | 0w1 => #1 dirt_material
                | 0w2 => #1 dirt_material + 0w1
                | 0w3 => #1 dirt_material + 0w2
                | 0w4 => #1 dirt_material + 0w3
                | 0w5 => #1 dirt_material + 0w5
                | 0w6 => #1 dirt_material + 0w5
                | _ =>   #1 dirt_material + 0w4
      in
          drawat (time, t, surf, x, y, wx, wy)
      end

    (* open problem: draw moving stars! *)
    (* stars (stationary) *)
    | drawat (time, 0w64, surf, x, y, wx, wy) =
      let
          (* stars are drawn at a fixed location
             on the screen, so figure out what
             patch we're in and draw a whole tile's worth *)
          val xx = Word32.fromInt wx
          val yy = Word32.fromInt wy
          val h = Word32.xorb(xx, 0wxDEADBEEF) + (yy * 0w31337)
          val n = Word32.toInt (Word32.andb(h, 0w7))

          fun star (0, _) = ()
            | star (n, h) =
              let
                  (* assumes tilebits * 3 < 32 *)
                  val px = Word32.andb(Word32.>>(h, TILEBITSW),
                                       MASKW)
                  val py = Word32.andb(Word32.>>(h, 0w1 + (0w2 * TILEBITSH)),
                                       MASKH)
                      
                  val lightc = Word8.fromInt time * Word8.fromInt n 
                  val c = SDL.color (lightc, lightc, 0wx22, 0wxFF)

                  val tx = x + Word32.toInt px
                  val ty = y + Word32.toInt py
              in
                  if tx >= 0 andalso ty >= 0
                     andalso tx < SDL.surface_width surf
                     andalso ty < SDL.surface_height surf
                  then SDL.drawpixel(surf, tx, ty, c)
                  else ();

                  star (n - 1, 
                        Word32.xorb(Word32.fromInt n * (h + 0w1234567),
                                    Word32.orb(Word32.<<(h, 0w11),
                                               Word32.>>(h, 0w32 - 0w11))))
              end
              
      in
          star (n, h)
      end

    (* PERF probably other solid colors can be done more efficiently *)
    | drawat (n, t, surf, x, y, wx, wy) =
      let 
          val t = Word32.toInt t
          val t = animate (n, t, wx, wy)
      in
          SDL.blit (tileset, 
                    (t mod TILESET_WIDTH) * TILEW,
                    (t div TILESET_WIDTH) * TILEH,
                    TILEW, TILEH,
                    surf,
                    x, y)
      end

  fun flipslope LM = ML
    | flipslope MH = HM
    | flipslope HM = MH
    | flipslope ML = LM
  fun ceilslope LM = HM
    | ceilslope MH = ML
    | ceilslope HM = LM
    | ceilslope ML = MH

  fun mirrordiag1 (m,x,y) = 
    ((case m of
         MRAMP sl => MRIGHT sl
       | MRIGHT sl => MRAMP sl
       | MLEFT sl => MCEIL (ceilslope sl)
       | MCEIL sl => MLEFT (ceilslope sl)
       | m => m),y,x)
  fun mirroryaxis1 w (m,x,y) = 
    ((case m of 
         MRAMP sl => MRAMP (flipslope sl)
       | MRIGHT sl => MLEFT sl
       | MLEFT sl => MRIGHT sl
       | MCEIL sl => MCEIL (flipslope sl)
       | m => m),w - x - 1,y)
  fun mirrorxaxis1 h (m,x,y) = 
    ((case m of 
         MRAMP sl => MCEIL (ceilslope sl)
       | MRIGHT sl => MRIGHT (flipslope sl)
       | MLEFT sl => MLEFT (flipslope sl)
       | MCEIL sl => MRAMP (ceilslope sl)
       | m => m),x,h - y - 1)


  fun mirrormasksdiag masks = List.map mirrordiag1 masks
  fun mirrormasksxaxis masks = List.map (mirrorxaxis1 (#2 (areasize (0,0) masks))) masks
  fun mirrormasksyaxis masks = List.map (mirroryaxis1 (#1 (areasize (0,0) masks))) masks

  fun cornersfromtop (masks,h) =
    foldl (fn (a,b) => (mirrorxaxis1 h) a::b) masks masks
  fun cornersfromtopleft (masks,w,h) = 
    cornersfromtop (foldl (fn (a,b) => (mirroryaxis1 w) a::b) masks masks,h)
  fun cornersfromtopleftpart (masks,w,h) = 
    cornersfromtopleft (foldl (fn (a,b) => mirrordiag1 a::b) masks masks,w,h)

  fun terrainlhslope (acc,n) (l,h) = 
    let
      val leftpart = List.tabulate (h-1,(fn x => (MSOLID,l-2,n + x + 1)))
      val rightpart = List.tabulate (h-1,(fn x => (MSOLID,l-1,n + x + 1)))
    in
      (MRAMP MH,l-1,n)::(MRAMP LM,l-2,n)::leftpart@(rightpart@acc)
    end
  fun terrainlhflat (acc,n) (l,h) = 
    List.tabulate (h,(fn x => (MSOLID,l-1,n+x)))@acc
  fun terrainlhbump (acc,n) (l,h) = 
    let
      val leftpart = List.tabulate (h,(fn x => (MSOLID,l-2,n + x + 1)))
      val rightpart = List.tabulate (h,(fn x => (MSOLID,l-1,n + x + 1)))
    in
      (MRAMP ML,l-1,n-1)::(MRAMP LM,l-2,n-1)::leftpart@(rightpart@acc)
    end
  fun terrainlhdip (acc,n) (l,h) = 
    let
      val leftpart = List.tabulate (h-1,(fn x => (MSOLID,l-2,n + x + 1)))
      val rightpart = List.tabulate (h-1,(fn x => (MSOLID,l-1,n + x + 1)))
    in
      (MRAMP MH,l-1,n)::(MRAMP HM,l-2,n)::leftpart@(rightpart@acc)
    end
  fun terrainlhfeature (acc,n) (l,h) =
    if MT.random_bool mt andalso h <> 0
    then terrainlhbump (acc,n) (l,h)
    else terrainlhdip (acc,n) (l,h)
  fun terrainlh (acc,n) (0,0) = acc 
    | terrainlh (acc,n) (l,h) = 
      if h <> 0 andalso (l = (2 * h) orelse (MT.random_nat mt (l - h * 2)) < (MT.random_nat mt h))
      then 
        let 
          val acc = terrainlhslope (acc,n) (l,h)
          val _ = print ("acc: "^Int.toString (List.length acc)^"\n")
        in
          terrainlh (acc,n+1) (l-2,h-1)
          (*terrainlh (terrainlhslope (acc,n) (l,h),n) (l-2,h-1)*)
        end 
      else if true orelse l = ((2 * h) + 1) orelse (MT.random_nat mt 3) <> 0
        then 
          let
            val acc = terrainlhflat (acc,n) (l,h)
            val _ = print ("acc: "^Int.toString (List.length acc)^"\n")
          in
            terrainlh (acc,n) (l-1,h)
          end 
        else 
          let 
            val acc = terrainlhfeature (acc,n) (l,h)(*terrainlh (terrainlhfeature (acc,n) (l,h),n) (l-2,h)*)
            val _ = print ("acc: "^Int.toString (List.length acc)^"\n")
          in
            terrainlhfeature (acc,n) (l-2,h)
          end
  fun terrainhl (acc,n) (l,h) = 
    let
      val masks = (terrainlh (acc,n) (l,h))
    in
      foldr (fn (a,b) => (mirroryaxis1 l) a::b) [] masks
    end

  fun brushtomasks (BRECT (w,h)) =
      let
        val xs = List.tabulate (w,(fn x => x))
        val ys = List.tabulate (h,(fn y => y))
      in
        foldr (fn (x,l) => foldr (fn (y,l) => (MSOLID,x,y)::l) l ys) [] xs
      end
    | brushtomasks (BSLOPEDRECT (w,h)) =
      let
        val w = if w < 2 then 2 else w
        val h = if h < 2 then 2 else h
        val xs = List.tabulate (w,(fn x=> x))
        val xs' = if w < 4 then List.tabulate (w-2,(fn x=> x+1)) else List.tabulate (w-4,(fn x=> x+2))
        val ys' = List.tabulate (h-2,(fn y=> y+1))
        val firstrow = 
          if w < 4 
          then [(MRAMP MH,0,0),(MRAMP HM,w-1,0)]
          else [(MRAMP LM,0,0),(MRAMP MH,1,0),(MRAMP HM,w-2,0),(MRAMP ML,w-1,0)] 
        val lastrow = 
          if w < 4 
          then [(MCEIL ML,0,h-1),(MCEIL LM,w-1,h-1)]
          else [(MCEIL HM,0,h-1),(MCEIL ML,1,h-1),(MCEIL LM,w-2,h-1),(MCEIL MH,w-1,h-1)]
        val firstrow = 
          foldr (fn (x,l) => (MSOLID,x,0)::l) 
            firstrow 
            xs'
        val midrows = 
          foldr (fn (x,l) => foldr (fn (y,l) => (MSOLID,x,y)::l) l ys') [] xs
        val lastrow = 
          foldr (fn (x,l) => (MSOLID,x,h-1)::l)
            lastrow
            xs'
      in
        firstrow@midrows@lastrow
      end
    | brushtomasks (BERASERECT (w,h)) =
      let
        val xs = List.tabulate (w,(fn x => x))
        val ys = List.tabulate (h,(fn y => y))
      in
        foldr (fn (x,l) => foldr (fn (y,l) => (MEMPTY,x,y)::l) l ys) [] xs
      end
    | brushtomasks (BHILLLH (l,h)) = 
      let
        val l = if l < 2 * h then 2 * h else l
      in
        terrainlh ([],0) (l,h)
      end
    | brushtomasks (BHILLHL (l,h)) =
      let
        val l = if l < 2 * h then 2 * h else l
      in
        terrainhl ([],0) (l,h)
      end
    | brushtomasks (BPASTEMASKS masks) = (!(!masks))
    
    | brushtomasks _ = []

  fun draw (t, surf, x, y, wx, wy) = drawat(Animate.now (), t, surf, x, y, wx, wy)


  fun toword x = x
  fun fromword x = x (* XXX check bounds *)
end

